/**
 * abaaso
 *
 * @author Jason Mulligan <jason.mulligan@avoidwork.com>
 * @copyright Jason Mulligan 2012
 * @license BSD-3 <http://opensource.org/licenses/BSD-3-Clause>
 * @link http://abaaso.com
 * @module abaaso
 * @version 3.2.2
 */

(function (global) {

var document  = global.document,
    location  = global.location,
    navigator = global.navigator,
    server    = typeof document === "undefined",
    abaaso, http, https, url;

if (server) {
	url   = require("url");
	http  = require("http");
	https = require("https");

	if (typeof Storage === "undefined")        localStorage   = require("localStorage");
	if (typeof XMLHttpRequest === "undefined") XMLHttpRequest = null;
}

abaaso = global.abaaso || (function () {
"use strict";

var $, bootstrap, error, external;

/**
 * Array methods
 *
 * @class array
 * @namespace abaaso
 */
var array = {
	/**
	 * Adds 'arg' to 'obj' if it is not found
	 * 
	 * @method add
	 * @param  {Array} obj Array to receive 'arg'
	 * @param  {Mixed} arg Argument to set in 'obj'
	 * @return {Array}     Array that was queried
	 */
	add : function (obj, arg) {
		if (!array.contains(obj, arg)) obj.push(arg);
		return obj;
	},

	/**
	 * Returns an Object (NodeList, etc.) as an Array
	 *
	 * @method cast
	 * @param  {Object}  obj Object to cast
	 * @param  {Boolean} key [Optional] Returns key or value, only applies to Objects without a length property
	 * @return {Array}       Object as an Array
	 */
	cast : function (obj, key) {
		key   = (key === true);
		var o = [];

		if (!isNaN(obj.length)) {
			if (!client.ie || client.version > 8) o = Array.prototype.slice.call(obj);
			else {
				try { o = Array.prototype.slice.call(obj); }
				catch (e) { utility.iterate(obj, function (i, idx) { if (idx !== "length") o.push(i); }); }
			}
		}
		else key ? o = array.keys(obj) : utility.iterate(obj, function (i) { o.push(i); });

		return o;
	},

	/**
	 * Transforms an Array to a 2D Array of chunks
	 * 
	 * @param  {Array}  obj  Array to parse
	 * @param  {Number} size Chunk size (integer)
	 * @return {Array}       Chunked Array
	 */
	chunk : function (obj, size) {
		var result = [],
		    nth    = number.round((obj.length / size), "up"),
		    start  = 0,
		    i      = -1;

		while (++i < nth) {
			start = i * size;
			result.push(array.limit(obj, start, size));
		}

		return result;
	},

	/**
	 * Clears an Array without destroying it
	 * 
	 * @param  {Array} obj Array to clear
	 * @return {Array}     Cleared Array
	 */
	clear : function (obj) {
		return obj.length > 0 ? array.remove(obj, 0, obj.length) : obj;
	},

	/**
	 * Clones an Array
	 * 
	 * @method clone
	 * @param  {Array} obj Array to clone
	 * @return {Array}     Clone of Array
	 */
	clone : function (obj) {
		return utility.clone(obj);
	},

	/**
	 * Determines if obj contains arg
	 * 
	 * @method contains
	 * @param  {Array} obj Array to search
	 * @param  {Mixed} arg Value to look for
	 * @return {Boolean}   True if found, false if not
	 */
	contains : function (obj, arg) {
		return (array.index(obj, arg) > -1);
	},

	/**
	 * Creates a new Array of the result of `fn` executed against every index of `obj`
	 * 
	 * @param  {Array}    obj Array to iterate
	 * @param  {Function} fn  Function to execute against indices
	 * @return {Array}        New Array
	 */
	collect : function (obj, fn) {
		var result = [];

		array.each(obj, function (i) {
			result.push(fn(i));
		});
		return result;
	},

	/**
	 * Compacts a Array by removing `null` or `undefined` indices
	 * 
	 * @param  {Array} obj    Array to compact
	 * @param  {Boolean} diff Indicates to return resulting Array only if there's a difference
	 * @return {Array}        Compacted copy of `obj`, or null (if `diff` is passed & no diff is found)
	 */
	compact : function (obj, diff) {
		var valid  = /null|undefined/,
		    result = [];

		result = obj.filter(function (i) {
			if (!valid.test(i)) return true;
		});

		return !diff ? result : (result.length < obj.length ? result : null);
	},

	/**
	 * Finds the difference between array1 and array2
	 *
	 * @method diff
	 * @param  {Array} array1 Source Array
	 * @param  {Array} array2 Comparison Array
	 * @return {Array}        Array of the differences
	 */
	diff : function (array1, array2) {
		var result = [];

		array.each(array1, function (i) { if (!array.contains(array2, i)) result.add(i); });
		array.each(array2, function (i) { if (!array.contains(array1, i)) result.add(i); });
		return result;
	},

	/**
	 * Iterates obj and executes fn
	 * Parameters for fn are 'value', 'key'
	 * 
	 * @param  {Array}    obj Array to iterate
	 * @param  {Function} fn  Function to execute on index values
	 * @return {Array}        Array
	 */
	each : function (obj, fn) {
		var nth = obj.length,
		    i   = -1,
		    r;

		while (++i < nth) {
			r = fn.call(obj, obj[i], i);
			if (r === false) break;
		}
		return obj;
	},

	/**
	 * Determines if an Array is empty
	 * 
	 * @param  {Array} obj Array to inspect
	 * @return {Boolean}   `true` if there's no indices
	 */
	empty : function (obj) {
		return (obj.length === 0);
	},

	/**
	 * Determines if `a` is equal to `b`
	 * 
	 * @param  {Array} a Array to compare
	 * @param  {Array} b Array to compare
	 * @return {Boolean} `true` if the Arrays match
	 */
	equal : function (a, b) {
		return (json.encode(a) === json.encode(b));
	},

	/**
	 * Fills `obj` with the evalution of `arg`, optionally from `start` to `offset`
	 * 
	 * @param  {Array}  obj   Array to fill
	 * @param  {Mixed}  arg   String, Number of Function to fill with
	 * @param  {Number} start [Optional] Index to begin filling at
	 * @param  {Number} end   [Optional] Offset from start to stop filling at
	 * @return {Array}        Filled Array
	 */
	fill : function (obj, arg, start, offset) {
		var fn  = typeof arg === "function",
		    l   = obj.length,
		    i   = !isNaN(start)  ? start      : 0,
		    nth = !isNaN(offset) ? i + offset : l - 1;

		if (nth > (l - 1)) nth = l - 1;
		for (; i <= nth; i++) {
			obj[i] = fn ? arg(obj[i]) : arg;
		}
		return obj;
	},

	/**
	 * Returns the first Array node
	 *
	 * @method first
	 * @param  {Array} obj The array
	 * @return {Mixed}     The first node of the array
	 */
	first : function (obj) {
		return obj[0];
	},

	/**
	 * Flattens a 2D Array
	 * 
	 * @param  {Array} obj 2D Array to flatten
	 * @return {Array}     Flatten Array
	 */
	flat : function (obj) {
		var result = [];

		result = obj.reduce(function (a, b) {
			return a.concat(b);
		}, result);

		return result;
	},

	/**
	 * Facade to indexOf for shorter syntax
	 *
	 * @method index
	 * @param  {Array} obj Array to search
	 * @param  {Mixed} arg Value to find index of
	 * @return {Number}    The position of arg in instance
	 */
	index : function (obj, arg) {
		return obj.indexOf(arg);
	},

	/**
	 * Returns an Associative Array as an Indexed Array
	 *
	 * @method indexed
	 * @param  {Array} obj Array to index
	 * @return {Array}     Indexed Array
	 */
	indexed : function (obj) {
		var indexed = [];

		utility.iterate(obj, function (v, k) {
			typeof v === "object" ? indexed = indexed.concat(array.indexed(v)) : indexed.push(v);
		});
		return indexed;
	},

	/**
	 * Finds the intersections between array1 and array2
	 *
	 * @method intersect
	 * @param  {Array} array1 Source Array
	 * @param  {Array} array2 Comparison Array
	 * @return {Array}        Array of the intersections
	 */
	intersect : function (array1, array2) {
		var a = array1.length > array2.length ? array1 : array2,
		    b = a === array1 ? array2 : array1;

		return a.filter(function (key) { return array.contains(b, key); });
	},

	/**
	 * Keeps every element of `obj` for which `fn` evaluates to true
	 * 
	 * @param  {Array}    obj Array to iterate
	 * @param  {Function} fn  Function to test indices against
	 * @return {Array}        Array
	 */
	keep_if : function (obj, fn) {
		if (typeof fn !== "function") throw Error(label.error.invalidArguments);
		var result = [],
		    remove = [];

		result = obj.filter(fn);
		remove = array.diff(obj, result);
		array.each(remove, function (i, idx) {
			array.remove(obj, array.index(obj, i));
		});
		return obj;
	},

	/**
	 * Returns the keys in an Associative Array
	 *
	 * @method keys
	 * @param  {Array} obj Array to extract keys from
	 * @return {Array}     Array of the keys
	 */
	keys : function (obj) {
		var keys = [];

		if (typeof Object.keys === "function") keys = Object.keys(obj)
		else utility.iterate(obj, function (v, k) {
			keys.push(k);
		});
		return keys;
	},

	/**
	 * Returns the last index of the Array
	 *
	 * @method last
	 * @param  {Array}  obj Array
	 * @param  {Number} arg [Optional] Negative offset from last index to return
	 * @return {Mixed}      Last index(s) of Array
	 */
	last : function (obj, arg) {
		var n = obj.length - 1;

		return isNaN(arg) || arg === 1 ? obj[n] : array.limit(obj, n - --arg, n);
	},

	/**
	 * Returns a limited range of indices from the Array
	 * 
	 * @param  {Array}  obj    Array to iterate
	 * @param  {Number} start  Starting index
	 * @param  {Number} offset Number of indices to return
	 * @return {Array}         Array of indices
	 */
	limit : function (obj, start, offset) {
		var result = [],
		    i      = start - 1,
		    nth    = start + offset;

		while (++i < nth) {
			if (typeof obj[i] !== "undefined") result.push(obj[i]);
		}

		return result;
	},

	/**
	 * Finds the maximum value in an Array
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Mixed}     Number, String, etc.
	 */
	max : function (obj) {
		return array.last(obj.sort(array.sort));
	},

	/**
	 * Finds the mean of an Array (of numbers)
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Number}    Mean of the Array (float or integer)
	 */
	mean : function (obj) {
		return (array.sum(obj) / obj.length);
	},

	/**
	 * Finds the median value of an Array (of numbers)
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Number}    Median number of the Array (float or integer)
	 */
	median : function (obj) {
		var nth    = obj.length,
		    mid    = number.round(nth / 2, "down"),
		    sorted = obj.sort(array.sort);

		return number.odd(nth) ? sorted[mid] : ((sorted[mid - 1] + sorted[mid]) / 2);
	},

	/**
	 * Finds the minimum value in an Array
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Mixed}     Number, String, etc.
	 */
	min : function (obj) {
		return obj.sort(array.sort)[0];
	},

	/**
	 * Finds the mode value of an Array
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Mixed}     Mode value of the Array
	 */
	mode : function (obj) {
		var values = {},
		    count  = 0,
		    nth    = 0,
		    mode   = [],
		    result;

		// Counting values
		array.each(obj, function (i) {
			!isNaN(values[i]) ? ++values[i] : values[i] = 1;
		});

		// Finding the highest occurring count
		count = array.max(array.cast(values));

		// Finding values that match the count
		utility.iterate(values, function (v, k) {
			if (v === count) mode.push(number.parse(k));
		});

		// Determining the result
		nth = mode.length;
		if (nth > 0) result = nth === 1 ? mode[0] : mode;

		return result;
	},

	/**
	 * Finds the range of the Array (of numbers) values
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Number}    Range of the array (float or integer)
	 */
	range : function (obj) {
		return array.max(obj) - array.min(obj);
	},

	/**
	 * Searches a 2D Array `obj` for the first match of `arg` as a second index
	 * 
	 * @param  {Array} obj 2D Array to search
	 * @param  {Mixed} arg Primitive to find
	 * @return {Mixed}     Array or undefined
	 */
	rassoc : function (obj, arg) {
		var result;

		array.each(obj, function (i, idx) {
			if (i[1] === arg) {
				result = obj[idx];
				return false;
			}
		});
		return result;
	},

	/**
	 * Returns Array containing the items in `obj` for which `fn()` is not true
	 * 
	 * @param  {Array}    obj Array to iterate
	 * @param  {Function} fn  Function to execute against `obj` indices
	 * @return {Array}        Array of indices which fn() is not true
	 */
	reject : function (obj, fn) {
		return array.diff(obj, obj.filter(fn));
	},
	
	/**
	 * Replaces the contents of `obj` with `arg`
	 * 
	 * @param  {Array} obj Array to modify
	 * @param  {Array} arg Array to become `obj`
	 * @return {Array}     New version of `obj`
	 */
	replace : function (obj, arg) {
		array.remove(obj, 0, obj.length);
		array.each(arg, function (i) {
			obj.push(i);
		});
		return obj;
	},

	/**
	 * Removes indices from an Array without recreating it
	 *
	 * @method remove
	 * @param  {Array}  obj   Array to remove from
	 * @param  {Mixed}  start Starting index, or value to find within obj
	 * @param  {Number} end   [Optional] Ending index
	 * @return {Array}        Modified Array
	 */
	remove : function (obj, start, end) {
		if (isNaN(start)) {
			start = obj.index(start);
			if (start === -1) return obj;
		}
		else start = start || 0;

		var length    = obj.length,
		    remaining = obj.slice((end || start) + 1 || length);

		obj.length = start < 0 ? (length + start) : start;
		obj.push.apply(obj, remaining);
		return obj;
	},

	/**
	 * Deletes every element of `obj` for which `fn` evaluates to true
	 * 
	 * @param  {Array}    obj Array to iterate
	 * @param  {Function} fn  Function to test indices against
	 * @return {Array}        Array
	 */
	remove_if : function (obj, fn) {
		if (typeof fn !== "function") throw Error(label.error.invalidArguments);
		var remove = [];

		remove = obj.filter(fn);
		array.each(remove, function (i, idx) {
			array.remove(obj, array.index(obj, i));
		});
		return obj;
	},

	/**
	 * Deletes elements of `obj` until `fn` evaluates to false
	 * 
	 * @param  {Array}    obj Array to iterate
	 * @param  {Function} fn  Function to test indices against
	 * @return {Array}        Array
	 */
	remove_while : function (obj, fn) {
		if (typeof fn !== "function") throw Error(label.error.invalidArguments);
		var remove = [];

		array.each(obj, function (i) {
			var result = fn(i);

			if (result !== false) remove.push(i);
			else return false;
		});

		array.each(remove, function (i, idx) {
			array.remove(obj, array.index(obj, i));
		});

		return obj;	
	},

	/**
	 * Returns the "rest" of `obj` from `arg`
	 * 
	 * @param  {Array}  obj Array to parse
	 * @param  {Number} arg [Optional] Start position of subset of `obj` (positive number only)
	 * @return {Array}      Array of a subset of `obj`
	 */
	rest : function (obj, arg) {
		arg = arg || 1;
		if (arg < 1) arg = 1;
		return array.limit(obj, arg, obj.length);
	},

	/**
	 * Finds the last index of `arg` in `obj`
	 * 
	 * @param  {Array} obj Array to search
	 * @param  {Mixed} arg Primitive to find
	 * @return {Mixed}     Index or undefined
	 */
	rindex : function (obj, arg) {
		var result = -1;

		array.each(obj, function (i, idx) {
			if (i === arg) result = idx;
		});
		return result;
	},

	/**
	 * Returns new Array with `arg` moved to the first index
	 * 
	 * @param  {Array}  obj Array to rotate
	 * @param  {Number} arg Index to become the first index, if negative the rotation is in the opposite direction
	 * @return {Array}      Newly rotated Array
	 */
	rotate : function (obj, arg) {
		var result = [],
		    nth    = obj.length;

		if (arg === 0) result = obj;
		else {
			arg < 0 ? arg += nth : arg--;
			result = array.limit(obj, arg, nth);
			result = result.concat(array.limit(obj, 0, arg));
		}
		return result;
	},

	/**
	 * Generates a series Array
	 * 
	 * @param  {Number} start  Start value the series
	 * @param  {Number} end    [Optional] The end of the series
	 * @param  {Number} offset [Optional] Offset for indices, default is 1
	 * @return {Array}         Array of new series
	 */
	series : function (start, end, offset) {
		start      = start  || 0;
		end        = end    || start;
		offset     = offset || 1;
		var result = [],
		    n      = -1,
		    nth    = Math.max(0, Math.ceil((end - start) / offset));

		while (++n < nth) {
			result[n]  = start;
			start     += offset;
		}

		return result;
	},

	/**
	 * Splits an Array by divisor
	 * 
	 * @param  {Array}  obj     Array to parse
	 * @param  {Number} divisor Integer to divide the Array by
	 * @return {Array}          Split Array
	 */
	split : function (obj, divisor) {
		var result  = [],
		    total   = obj.length,
		    nth     = Math.ceil(total / divisor),
		    low     = Math.floor(total / divisor),
		    lower   = Math.ceil(total / nth),
		    lowered = false,
		    start   = 0,
		    i       = -1;

		// Finding the fold
		if (number.diff(total, (divisor * nth)) > nth) {
			lower = total - (low * divisor) + low - 1;
		}

		while (++i < divisor) {
			if (!lowered && lower < divisor && i === lower) {
				--nth;
				lowered = true;
			}
			if (i > 0) start = start + nth;
			result.push(array.limit(obj, start, nth));
		}

		return result;
	},

	/**
	 * Sorts the Array by parsing values
	 * 
	 * @param  {Mixed} a Argument to compare
	 * @param  {Mixed} b Argument to compare
	 * @return {Boolean} Boolean indicating sort order
	 */
	sort : function (a, b) {
		var nums = false,
		    result;

		if (!isNaN(a) && !isNaN(b)) nums = true;

		a = nums ? number.parse(a) : String(a);
		b = nums ? number.parse(b) : String(b);

		switch (true) {
			case a < b:
				result = -1;
				break;
			case a > b:
				result = 1;
				break;
			default:
				result = 0;
		}
		return result;
	},

	/**
	 * Gets the summation of an Array of numbers
	 * 
	 * @param  {Array} obj Array to sum
	 * @return {Number}    Summation of Array
	 */
	sum : function (obj) {
		var result = 0;

		result = obj.reduce(function (prev, cur) {
			return prev + cur;
		});

		return result;
	},

	/**
	 * Takes the first `arg` indices from `obj`
	 * 
	 * @param  {Array}  obj Array to parse
	 * @param  {Number} arg Offset from 0 to return
	 * @return {Array}      Subset of `obj`
	 */
	take : function (obj, arg) {
		return array.limit(obj, 0, arg);
	},

	/**
	 * Gets the total keys in an Array
	 *
	 * @method total
	 * @param  {Array} obj Array to find the length of
	 * @return {Number}    Number of keys in Array
	 */
	total : function (obj) {
		return array.indexed(obj).length;
	},

	/**
	 * Casts an Array to Object
	 * 
	 * @param  {Array} ar Array to transform
	 * @return {Object}   New object
	 */
	toObject : function (ar) {
		var obj = {},
		    i   = ar.length;

		while (i--) obj[i.toString()] = ar[i];
		return obj;
	},

	/**
	 * Returns an Array of unique indices of `obj`
	 * 
	 * @param  {Array} obj Array to parse
	 * @return {Array}     Array of unique indices
	 */
	unique : function (obj, fn) {
		var result = [];

		array.each(obj, function (i) {
			array.add(result, i);
		});
		return result;
	},

	/**
	 * Converts any arguments to Arrays, then merges elements of `obj` with corresponding elements from each argument
	 * 
	 * @param  {Array} obj  Array to transform
	 * @param  {Mixed} args Argument instance or Array to merge
	 * @return {Array}      Array
	 */
	zip : function (obj, args) {
		var result = [];

		// Preparing args
		if (!(args instanceof Array)) args = typeof args === "object" ? array.cast(args) : [args];
		array.each(args, function (i, idx) {
			if (!(i instanceof Array)) this[idx] = [i];
		});

		// Building result Array
		array.each(obj, function (i, idx) {
			result[idx] = [i];
			array.each(args, function (x) {
				result[idx].push(x[idx] || null);
			});
		});

		return result;
	}
};

/**
 * Cache for RESTful behavior
 *
 * @class cache
 * @namespace abaaso
 * @private
 */
var cache = {
	// Collection URIs
	items : {},

	/**
	 * Garbage collector for the cached items
	 *
	 * @method clean
	 * @return {Undefined} undefined
	 */
	clean : function () {
		return utility.iterate(cache.items, function (v, k) { if (cache.expired(k)) cache.expire(k); });
	},

	/**
	 * Expires a URI from the local cache
	 * 
	 * Events: expire    Fires when the URI expires
	 *
	 * @method expire
	 * @param  {String}  uri    URI of the local representation
	 * @param  {Boolean} silent [Optional] If 'true', the event will not fire
	 * @return {Undefined}      undefined
	 */
	expire : function (uri, silent) {
		silent = (silent === true);
		if (typeof cache.items[uri] !== "undefined") {
			delete cache.items[uri];
			if (!silent) uri.fire("beforeExpire, expire, afterExpire");
			return true;
		}
		else return false;
	},

	/**
	 * Determines if a URI has expired
	 *
	 * @method expired
	 * @param  {Object} uri Cached URI object
	 * @return {Boolean}    True if the URI has expired
	 */
	expired : function (uri) {
		var item = cache.items[uri];
		return typeof item !== "undefined" && typeof item.expires !== "undefined" && item.expires < new Date();
	},

	/**
	 * Returns the cached object {headers, response} of the URI or false
	 *
	 * @method get
	 * @param  {String}  uri    URI/Identifier for the resource to retrieve from cache
	 * @param  {Boolean} expire [Optional] If 'false' the URI will not expire
	 * @param  {Boolean} silent [Optional] If 'true', the event will not fire
	 * @return {Mixed}          URI Object {headers, response} or False
	 */
	get : function (uri, expire) {
		expire = (expire !== false);
		if (typeof cache.items[uri] === "undefined") return false;
		if (expire && cache.expired(uri)) {
			cache.expire(uri);
			return false;
		}
		return utility.clone(cache.items[uri]);
	},

	/**
	 * Sets, or updates an item in cache.items
	 *
	 * @method set
	 * @param  {String} uri      URI to set or update
	 * @param  {String} property Property of the cached URI to set
	 * @param  {Mixed} value     Value to set
	 * @return {Mixed}           URI Object {headers, response} or undefined
	 */
	set : function (uri, property, value) {
		if (typeof cache.items[uri] === "undefined") {
			cache.items[uri] = {};
			cache.items[uri].permission = 0;
		}
		property === "permission" ? cache.items[uri].permission |= value
		                          : (property === "!permission" ? cache.items[uri].permission &= ~value
		                                                        : cache.items[uri][property]   =  value);
		return cache.items[uri];
	}
};

/**
 * Client properties and methods
 *
 * @class client
 * @namespace abaaso
 */
var client = {
	android : (function () { return !server && /android/i.test(navigator.userAgent); })(),
	blackberry : (function () { return !server && /blackberry/i.test(navigator.userAgent); })(),
	chrome  : (function () { return !server && /chrome/i.test(navigator.userAgent); })(),
	firefox : (function () { return !server && /firefox/i.test(navigator.userAgent); })(),
	ie      : (function () { return !server && /msie/i.test(navigator.userAgent); })(),
	ios     : (function () { return !server && /ipad|iphone/i.test(navigator.userAgent); })(),
	linux   : (function () { return !server && /linux|bsd|unix/i.test(navigator.userAgent); })(),
	mobile  : (function () { abaaso.client.mobile = this.mobile = !server && (/blackberry|iphone|webos/i.test(navigator.userAgent) || (/android/i.test(navigator.userAgent) && (abaaso.client.size.height < 720 || abaaso.client.size.width < 720))); }),
	playbook: (function () { return !server && /playbook/i.test(navigator.userAgent); })(),
	opera   : (function () { return !server && /opera/i.test(navigator.userAgent); })(),
	osx     : (function () { return !server && /macintosh/i.test(navigator.userAgent); })(),
	safari  : (function () { return !server && /safari/i.test(navigator.userAgent.replace(/chrome.*/i, "")); })(),
	tablet  : (function () { abaaso.client.tablet = this.tablet = !server && (/ipad|playbook|webos/i.test(navigator.userAgent) || (/android/i.test(navigator.userAgent) && (abaaso.client.size.width >= 720 || abaaso.client.size.width >= 720))); }),
	webos   : (function () { return !server && /webos/i.test(navigator.userAgent); })(),
	windows : (function () { return !server && /windows/i.test(navigator.userAgent); })(),
	version : (function () {
		var version = 0;
		switch (true) {
			case this.chrome:
				version = navigator.userAgent.replace(/(.*chrome\/|safari.*)/gi, "");
				break;
			case this.firefox:
				version = navigator.userAgent.replace(/(.*firefox\/)/gi, "");
				break;
			case this.ie:
				version = parseInt(navigator.userAgent.replace(/(.*msie|;.*)/gi, ""));
				if (document.documentMode < version) version = document.documentMode;
				break;
			case this.opera:
				version = navigator.userAgent.replace(/(.*opera\/|\(.*)/gi, "");
				break;
			case this.safari:
				version = navigator.userAgent.replace(/(.*version\/|safari.*)/gi, "");
				break;
			default:
				version = (typeof navigator !== "undefined") ? navigator.appVersion : 0;
		}
		version = !isNaN(parseInt(version)) ? parseInt(version) : 0;
		abaaso.client.version = this.version = version;
		return version;
	}),

	/**
	 * Quick way to see if a URI allows a specific command
	 *
	 * @method allows
	 * @param  {String} uri     URI to query
	 * @param  {String} command Command to query for
	 * @return {Boolean}        True if the command is allowed
	 */
	allows : function (uri, command) {
		if (uri.isEmpty() || command.isEmpty()) throw Error(label.error.invalidArguments);

		if (!cache.get(uri, false)) return undefined;

		command = command.toLowerCase();
		var result;

		switch (true) {
			case command === "delete":
				result = !((client.permissions(uri, command).bit & 1) === 0);
				break;
			case (/^(head|get|options)$/.test(command)):
				result = !((client.permissions(uri, command).bit & 4) === 0);
				break;
			case (/^(post|put)$/.test(command)):
				result = !((client.permissions(uri, command).bit & 2) === 0);
				break;
			default:
				result = false;
		}
		return result;
	},

	/**
	 * Gets bit value based on args
	 *
	 * 1 --d delete
	 * 2 -w- write
	 * 3 -wd write and delete
	 * 4 r-- read
	 * 5 r-d read and delete
	 * 6 rw- read and write
	 * 7 rwd read, write and delete
	 *
	 * @method bit
	 * @param  {Array} args Array of commands the URI accepts
	 * @return {Number} To be set as a bit
	 * @private
	 */
	bit : function (args) {
		var result = 0;

		args.each(function (a) {
			switch (a.toLowerCase()) {
				case "head":
				case "get":
				case "options":
					result |= 4;
					break;
				case "post":
				case "put":
					result |= 2;
					break;
				case "delete":
					result |= 1;
					break;
			}
		});
		return result;
	},

	/**
	 * Determines if a URI is a CORS end point
	 * 
	 * @method cors
	 * @param  {String} uri  URI to parse
	 * @return {Boolean}     True if CORS
	 */
	cors : function (uri) {
		return (!server && uri.indexOf("//") > -1 && uri.indexOf("//" + location.host) === -1);
	},

	/**
	 * Caches the headers from the XHR response
	 * 
	 * @method headers
	 * @param  {Object} xhr  XMLHttpRequest Object
	 * @param  {String} uri  URI to request
	 * @param  {String} type Type of request
	 * @return {Object}      Cached URI representation
	 * @private
	 */
	headers : function (xhr, uri, type) {
		var headers = String(xhr.getAllResponseHeaders()).trim().split("\n"),
		    items   = {},
		    o       = {},
		    allow   = null,
		    expires = new Date(),
		    cors    = client.cors(uri),
		    rvalue  = /.*:\s+/,
		    rheader = /:.*/,
		    rallow  = /^allow$/i,
		    rcallow = /^access-control-allow-methods$/i;

		headers.each(function (h) {
			var header, value;

			value         = h.replace(rvalue, "");
			header        = h.replace(rheader, "");
			header        = header.indexOf("-") === -1 ? header.capitalize() : (function () { var x = []; header.explode("-").each(function (i) { x.push(i.capitalize()) }); return x.join("-"); })();
			items[header] = value;
			if (allow === null) {
				if (cors && rcallow.test(header)) allow = value;
				else if (rallow.test(header)) allow = value;
			}
		});

		switch (true) {
			case typeof items["Cache-Control"] !== "undefined" && /no/.test(items["Cache-Control"]):
			case typeof items["Pragma"] !== "undefined" && /no/.test(items["Pragma"]):
				break;
			case typeof items["Cache-Control"] !== "undefined" && /\d/.test(items["Cache-Control"]):
				expires = expires.setSeconds(expires.getSeconds() + parseInt(/\d{1,}/.exec(items["Cache-Control"])[0]));
				break;
			case typeof items["Expires"] !== "undefined":
				expires = new Date(items["Expires"]);
				break;
			default:
				expires = expires.setSeconds(expires.getSeconds() + $.expires);
		}

		o.expires    = expires;
		o.headers    = items;
		o.permission = client.bit(allow !== null ? allow.explode() : [type]);

		if (type !== "head") {
			cache.set(uri, "expires",    o.expires);
			cache.set(uri, "headers",    o.headers);
			cache.set(uri, "permission", o.permission);
		}

		return o;
	},

	/**
	 * Parses an XHR response
	 * 
	 * @param  {Object} xhr  XHR Object
	 * @param  {String} type [Optional] Content-Type header value
	 * @return {Mixed}       Array, Boolean, Document, Number, Object or String
	 */
	parse : function (xhr, type) {
		type = type || "";
		var result, obj;

		switch (true) {
			case (/json|plain|javascript/.test(type) || type.isEmpty()) && Boolean(obj = json.decode(/[\{\[].*[\}\]]/.exec(xhr.responseText), true)):
				result = obj;
				break;
			case (/xml/.test(type) && String(xhr.responseText).isEmpty() && xhr.responseXML !== null):
				result = xml.decode(typeof xhr.responseXML.xml !== "undefined" ? xhr.responseXML.xml : xhr.responseXML);
				break;
			case (/<[^>]+>[^<]*]+>/.test(xhr.responseText)):
				result = xml.decode(xhr.responseText);
				break;
			default:
				result = xhr.responseText;
		}
		return result;
	},

	/**
	 * Returns the permission of the cached URI
	 *
	 * @method permissions
	 * @param  {String} uri URI to query
	 * @return {Object}     Contains an Array of available commands, the permission bit and a map
	 */
	permissions : function (uri) {
		var cached = cache.get(uri, false),
		    bit    = !cached ? 0 : cached.permission,
		    result = {allows: [], bit: bit, map: {read: 4, write: 2, "delete": 1}};

		if (bit & 1) result.allows.push("DELETE");
		if (bit & 2) (function () { result.allows.push("POST"); result.allows.push("PUT"); })();
		if (bit & 4) result.allows.push("GET");
		return result;
	},

	/**
	 * Creates a JSONP request
	 *
	 * Events: beforeJSONP     Fires before the SCRIPT is made
	 *         afterJSONP      Fires after the SCRIPT is received
	 *         failedJSONP     Fires on error
	 *         timeoutJSONP    Fires 30s after SCRIPT is made
	 *
	 * @method jsonp
	 * @param  {String}   uri     URI to request
	 * @param  {Function} success A handler function to execute when an appropriate response been received
	 * @param  {Function} failure [Optional] A handler function to execute on error
	 * @param  {Mixed}    args    Custom JSONP handler parameter name, default is "callback"; or custom headers for GET request (CORS)
	 * @return {String}           URI to query
	 */
	jsonp : function (uri, success, failure, args) {
		var curi = uri,
		    guid = utility.guid(true),
		    callback, cbid, s;

		// Utilizing the sugar if namespace is not global
		if (typeof external === "undefined") {
			if (typeof global.abaaso === "undefined") utility.define("abaaso.callback", {}, global);
			external = "abaaso";
		}

		switch (true) {
			case typeof args === "undefined":
			case args === null:
			case args instanceof Object && (args.callback === null || typeof args.callback === "undefined"):
			case typeof args === "string" && args.isEmpty():
				callback = "callback";
				break;
			case args instanceof Object && typeof args.callback !== "undefined":
				callback = args.callback;
				break;
			default:
				callback = "callback";
		}

		curi = curi.replace(callback + "=?", "");

		curi.once("afterJSONP", function (arg) {
			this.un("failedJSONP", guid);
			if (typeof success === "function") success(arg);
		}, guid).once("failedJSONP", function () {
			this.un("failedJSONP", guid);
			if (typeof failure === "function") failure();
		}, guid);

		do cbid = utility.genId().slice(0, 10);
		while (typeof global.abaaso.callback[cbid] !== "undefined");

		uri = uri.replace(callback + "=?", callback + "=" + external + ".callback." + cbid);

		global.abaaso.callback[cbid] = function (arg) {
			s.destroy();
			clearTimeout(utility.timer[cbid]);
			delete utility.timer[cbid];
			delete global.abaaso.callback[cbid];
			curi.fire("afterJSONP", arg);
		};

		s = $("head")[0].create("script", {src: uri, type: "text/javascript"});
		utility.defer(function () { curi.fire("failedJSONP"); }, 30000, cbid);
		return uri;
	},

	/**
	 * Creates an XmlHttpRequest to a URI (aliased to multiple methods)
	 *
	 * Events: beforeXHR             Fires before the XmlHttpRequest is made
	 *         before[type]          Fires before the XmlHttpRequest is made, type specific
	 *         failed[type]          Fires on error
	 *         progress[type]        Fires on progress
	 *         progressUpload[type]  Fires on upload progress
	 *         received[type]        Fires on XHR readystate 2
	 *         timeout[type]         Fires when XmlHttpRequest times out
	 *
	 * @method request
	 * @param  {String}   uri     URI to query
	 * @param  {String}   type    Type of request (DELETE/GET/POST/PUT/HEAD)
	 * @param  {Function} success A handler function to execute when an appropriate response been received
	 * @param  {Function} failure [Optional] A handler function to execute on error
	 * @param  {Mixed}    args    Data to send with the request
	 * @param  {Object}   headers Custom request headers (can be used to set withCredentials)
	 * @return {String}           URI to query
	 * @private
	 */
	request : function (uri, type, success, failure, args, headers) {
		var cors, xhr, payload, cached, typed, guid, contentType, doc, ab, blob;

		if (/^(post|put)$/i.test(type) && typeof args === "undefined") throw Error(label.error.invalidArguments);

		type         = type.toLowerCase();
		headers      = headers instanceof Object ? headers : null;
		cors         = client.cors(uri);
		xhr          = (client.ie && client.version < 10 && cors) ? new XDomainRequest() : new XMLHttpRequest();
		payload      = /^(post|put)$/i.test(type) && typeof args !== "undefined" ? args : null;
		cached       = type === "get" ? cache.get(uri) : false;
		typed        = type.capitalize();
		guid         = utility.guid(true);
		contentType  = null;
		doc          = (typeof Document !== "undefined");
		ab           = (typeof ArrayBuffer !== "undefined");
		blob         = (typeof Blob !== "undefined");

		if (type === "delete") uri.once("afterDelete", function () { cache.expire(this); });

		uri.once("after" + typed, function () {
			uri.un("failed" + typed, guid);
			if (typeof success === "function") success.apply(this, arguments);
		}, guid).once("failed" + typed, function () {
			uri.un("after" + typed, guid);
			if (typeof failure === "function") failure.apply(this, arguments);
		}, guid).fire("before" + typed);

		if (!/^(head|options)$/.test(type) && uri.allows(type) === false) return uri.fire("failed" + typed, null, {status: 405});

		if (type === "get" && Boolean(cached)) {
			if (server) {
				// Decorating XHR for proxy behavior
				xhr.readyState  = 4;
				xhr.status      = 200;
				xhr._resheaders = cached.headers;
			}
			uri.fire("afterGet", cached.response, (server ? xhr : undefined));
			xhr = null;
		}
		else {
			xhr[xhr instanceof XMLHttpRequest ? "onreadystatechange" : "onload"] = function (e) { client.response(xhr, uri, type); };

			// Setting events
			if (typeof xhr.onerror    !== "undefined") xhr.onerror    = function (e) { uri.fire("failed" + typed, e, xhr); };
			if (typeof xhr.ontimeout  !== "undefined") xhr.ontimeout  = function (e) { uri.fire("timeout" + typed, e, xhr); };
			if (typeof xhr.onprogress !== "undefined") xhr.onprogress = function (e) { uri.fire("progress" + typed, e, xhr); };
			if (typeof xhr.upload     !== "undefined" && typeof xhr.upload.onprogress !== "undefined") xhr.upload.onprogress = function (e) { uri.fire("progressUpload" + typed, e, xhr); };

			try {
				xhr.open(type.toUpperCase(), uri, true);

				// Setting Content-Type value
				if (headers !== null && headers.hasOwnProperty("Content-Type")) contentType = headers["Content-Type"];
				if (cors && contentType === null) contentType = "text/plain";

				// Transforming payload
				if (payload !== null) {
					if (payload.hasOwnProperty("xml")) payload = payload.xml;
					if (doc && payload instanceof Document) payload = xml.decode(payload);
					if (typeof payload === "string" && /<[^>]+>[^<]*]+>/.test(payload)) contentType = "application/xml";
					if (!(ab && payload instanceof ArrayBuffer) && !(blob && payload instanceof Blob) && payload instanceof Object) {
						contentType = "application/json";
						payload = json.encode(payload);
					}
					if (contentType === null && ((ab && payload instanceof ArrayBuffer) || (blob && payload instanceof Blob))) contentType = "application/octet-stream";
					if (contentType === null) contentType = "application/x-www-form-urlencoded; charset=UTF-8";
				}

				// Setting headers
				if (typeof xhr.setRequestHeader !== "undefined") {
					if (typeof cached === "object" && cached.headers.hasOwnProperty("ETag")) xhr.setRequestHeader("ETag", cached.headers.ETag);
					if (headers === null) headers = {};
					if (contentType !== null) headers["Content-Type"] = contentType;
					if (headers.hasOwnProperty("callback")) delete headers.callback;
					utility.iterate(headers, function (v, k) {
						if (v !== null && k !== "withCredentials") xhr.setRequestHeader(k, v);
					});
				}

				// Cross Origin Resource Sharing (CORS)
				if (typeof xhr.withCredentials === "boolean" && headers !== null && typeof headers.withCredentials === "boolean") xhr.withCredentials = headers.withCredentials;

				// Firing event & sending request
				uri.fire("beforeXHR", uri, xhr);
				payload !== null ? xhr.send(payload) : xhr.send();
			}
			catch (e) {
				error(e, arguments, this, true);
				uri.fire("failed" + typed, client.parse(xhr), xhr);
			}
		}
		return uri;
	},

	/**
	 * Caches the URI headers & response if received, and fires the relevant events
	 *
	 * If abaaso.state.header is set, an application state change is possible
	 *
	 * Permissions are handled if the ACCEPT header is received; a bit is set on the cached
	 * resource
	 *
	 * Events: afterXHR     Fires after the XmlHttpRequest response is received
	 *         after[type]  Fires after the XmlHttpRequest response is received, type specific
	 *         reset        Fires if a 206 response is received
	 *         moved        Fires if a 301 response is received
	 *         failure      Fires if an exception is thrown
	 *         headers      Fires after a possible state change, with the headers from the response
	 *
	 * @method response
	 * @param  {Object} xhr  XMLHttpRequest Object
	 * @param  {String} uri  URI to query
	 * @param  {String} type Type of request
	 * @return {String} uri  URI to query
	 * @private
	 */
	response : function (xhr, uri, type) {
		var typed = type.toLowerCase().capitalize(),
		    l     = location,
		    state = null,
		    xdr   = client.ie && typeof xhr.readyState === "undefined",
		    exception, s, o, r, t, x;

		// server-side exception handling
		exception = function (e, xhr) {
			error(e, arguments, this, true);
			uri.fire("failed" + typed, client.parse(xhr), xhr);
		}

		switch (true) {
			case !xdr && xhr.readyState === 2:
				uri.fire("received" + typed, null, xhr);
				break;
			case !xdr && xhr.readyState === 4:
				uri.fire("afterXHR", null, xhr);
				switch (xhr.status) {
					case 200:
					case 201:
					case 202:
					case 203:
					case 204:
					case 205:
					case 206:
					case 301:
						s = abaaso.state;
						o = client.headers(xhr, uri, type);

						switch (true) {
							case type === "head":
								return uri.fire("afterHead", o.headers);
							case type === "options":
								return uri.fire("afterOptions", o.headers);
							case type !== "delete" && /200|201/.test(xhr.status):
								t = typeof o.headers["Content-Type"] !== "undefined" ? o.headers["Content-Type"] : "";
								r = client.parse(xhr, t);
								if (typeof r === "undefined") throw Error(label.error.serverError);
								cache.set(uri, "response", (o.response = utility.clone(r)));
								break;
						}

						// Application state change triggered by hypermedia (HATEOAS)
						if (s.header !== null && Boolean(state = o.headers[s.header]) && s.current !== state) typeof s.change === "function" ? s.change(state) : s.current = state;

						uri.fire("headers", o.headers, xhr);

						switch (xhr.status) {
							case 200:
							case 201:
								uri.fire("after" + typed, r, xhr);
								break;
							case 202:
							case 203:
							case 204:
							case 206:
								uri.fire("after" + typed, null, xhr);
								break;
							case 205:
								uri.fire("reset", null, xhr);
								break;
							case 301:
								uri.fire("moved", r, xhr);
								break;
						}
						break;
					case 401:
						exception(!server ? Error(label.error.serverUnauthorized) : label.error.serverUnauthorized, xhr);
						break;
					case 403:
						cache.set(uri, "!permission", client.bit([type]));
						exception(!server ? Error(label.error.serverForbidden) : label.error.serverForbidden, xhr);
						break;
					case 405:
						cache.set(uri, "!permission", client.bit([type]));
						exception(!server ? Error(label.error.serverInvalidMethod) : label.error.serverInvalidMethod, xhr);
						break
					default:
						exception(!server ? Error(label.error.serverError) : label.error.serverError, xhr);
				}
				xhr.onreadystatechange = null;
				xhr = null;
				break;
			case xdr: // XDomainRequest
				switch (true) {
					case Boolean(x = json.decode(/[\{\[].*[\}\]]/.exec(xhr.responseText))):
						r = x;
						break;
					case (/<[^>]+>[^<]*]+>/.test(xhr.responseText)):
						r = xml.decode(xhr.responseText);
						break;
					default:
						r = xhr.responseText;
				}
				cache.set(uri, "permission", client.bit(["get"]));
				cache.set(uri, "response", r);
				uri.fire("afterGet", r, xhr);
				xhr.onload = null;
				xhr = null;
				break;
		}
		return uri;
	},


	/**
	 * Returns the visible area of the View
	 *
	 * @method size
	 * @return {Object} Describes the View {x: ?, y: ?}
	 */
	size : function () {
		var view = !server ? (typeof document.documentElement !== "undefined" ? document.documentElement : document.body) : {clientHeight: 0, clientWidth: 0};

		return {height: view.clientHeight, width: view.clientWidth};
	}
};

/**
 * Cookie methods
 *
 * @class cookie
 * @namespace abaaso
 */
var cookie = {
	/**
	 * Expires a cookie if it exists
	 *
	 * @method expire
	 * @param  {String} name   Name of the cookie to expire
	 * @param  {String} domain [Optional] Domain to set the cookie for
	 * @param  {Boolea} secure [Optional] Make the cookie only accessible via SSL
	 * @return {String}        Name of the expired cookie
	 */
	expire : function (name, domain, secure) {
		if (typeof cookie.get(name) !== "undefined") cookie.set(name, "", "-1s", domain, secure);
		return name;
	},

	/**
	 * Gets a cookie
	 *
	 * @method get
	 * @param  {String} name Name of the cookie to get
	 * @return {Mixed}       Cookie or undefined
	 */
	get : function (name) {
		return utility.coerce(cookie.list()[name]);
	},

	/**
	 * Gets the cookies for the domain
	 *
	 * @method list
	 * @return {Object} Collection of cookies
	 */
	list : function () {
		var result = {},
		    item, items;

		if (typeof document.cookie !== "undefined" && !document.cookie.isEmpty()) {
			items = document.cookie.explode(";");
			items.each(function (i) {
				item = i.explode("=");
				result[decodeURIComponent(item[0].toString().trim())] = decodeURIComponent(item[1].toString().trim());
			});
		}
		return result;
	},

	/**
	 * Creates a cookie
	 *
	 * The offset specifies a positive or negative span of time as day, hour, minute or second
	 *
	 * @method set
	 * @param  {String} name   Name of the cookie to create
	 * @param  {String} value  Value to set
	 * @param  {String} offset A positive or negative integer followed by "d", "h", "m" or "s"
	 * @param  {String} domain [Optional] Domain to set the cookie for
	 * @param  {Boolea} secure [Optional] Make the cookie only accessible via SSL
	 * @return {Object}        The new cookie
	 */
	set : function (name, value, offset, domain, secure) {
		if (typeof value === "undefined") value = "";
		value     += ";";
		if (typeof offset === "undefined") offset = "";
		domain     = (typeof domain === "string") ? (" domain=" + domain + ";") : "";
		secure     = (secure === true) ? "; secure" : "";
		var expire = "",
		    span   = null,
		    type   = null,
		    types  = ["d", "h", "m", "s"],
		    regex  = new RegExp(),
		    i      = types.length;

		if (!offset.isEmpty()) {
			while (i--) {
				utility.compile(regex, types[i]);
				if (regex.test(offset)) {
					type = types[i];
					span = parseInt(offset);
					break;
				}
			}

			if (isNaN(span)) throw Error(label.error.invalidArguments);

			expire = new Date();
			switch (type) {
				case "d":
					expire.setDate(expire.getDate() + span);
					break;
				case "h":
					expire.setHours(expire.getHours() + span);
					break;
				case "m":
					expire.setMinutes(expire.getMinutes() + span);
					break;
				case "s":
					expire.setSeconds(expire.getSeconds() + span);
					break;
			}
		}
		if (expire instanceof Date) expire = " expires=" + expire.toUTCString() + ";";
		document.cookie = (name.toString().trim() + "=" + value + expire + domain + " path=/" + secure);
		return cookie.get(name);
	}
};

/**
 * DataStore
 *
 * RESTful behavior is supported, by setting the 'key' & 'uri' properties
 *
 * @class data
 * @namespace abaaso
 */
var data = {
	// Inherited by data stores
	methods : {
		/**
		 * Batch sets or deletes data in the store
		 *
		 * Events: beforeDataBatch  Fires before the batch is queued
		 *         afterDataBatch   Fires after the batch is queued
		 *         failedDataBatch  Fires when an exception occurs
		 *
		 * @method batch
		 * @param  {String}  type  Type of action to perform (set/del/delete)
		 * @param  {Mixed}   data  Array of keys or indices to delete, or Object containing multiple records to set
		 * @param  {Boolean} sync  [Optional] Syncs store with data, if true everything is erased
		 * @param  {Number}  chunk Size to chunk Array to batch set or delete
		 * @return {Object}        Data store
		 */
		batch : function (type, data, sync, chunk) {
			type  = type.toString().toLowerCase();
			sync  = (sync === true);
			chunk = chunk || 1000;

			if (!/^(set|del|delete)$/.test(type) || typeof data !== "object") throw Error(label.error.invalidArguments);

			var obj  = this.parentNode,
			    self = this,
			    r    = 0,
			    nth  = data.length,
			    f    = false,
			    guid = utility.genId(true),
			    root = /^\/[^\/]/,
			    completed, failure, key, set, success, parsed;

			completed = function () {
				if (type === "delete") self.reindex();
				self.loaded = true;
				obj.fire("afterDataBatch");
			};

			failure = function (arg) {
				obj.fire("failedDataSet, failedDataBatch", arg);
			};

			set = function (data, key) {
				var guid = utility.genId(),
				    rec  = {};

				if (typeof rec.batch !== "function") rec = utility.clone(data)
				else utility.iterate(data, function (v, k) {
					if (!self.collections.contains(k)) rec[k] = utility.clone(v);
				});

				if (self.key !== null && typeof rec[self.key] !== "undefined") {
					key = rec[self.key];
					delete rec[self.key];
				}

				obj.once("afterDataSet", function () {
					this.un("failedDataSet", guid);
					if (++r && r === nth) completed();
				}, guid).once("failedDataSet", function () {
					this.un("afterDataSet", guid)
					if (!f) {
						f = true;
						this.fire("failedDataBatch");
					}
				}, guid);

				if (rec instanceof Array && self.uri !== null) self.generate(key);
				else self.set(key, rec, sync);
			};

			obj.fire("beforeDataBatch", data);

			if (type === "delete") {
				obj.on("afterDataDelete", function () {
					if (r++ && r === nth) {
						obj.un("afterDataDelete, failedDataDelete", guid);
						completed();
					}
				}, guid).once("failedDataDelete", function () {
					obj.un("afterDataDelete", guid);
					if (!f) {
						f = true;
						obj.fire("failedDataBatch");
					}
				}, guid);
			}

			if (type === "set") data.chunk(chunk).each(function (a, adx) {
					var offset = adx * chunk;

					$.defer(function () {
						a.each(function (i, idx) {
							idx = (offset + idx).toString();
							switch (true) {
								case typeof i === "object":
									set(i, idx);
									break;
								case i.indexOf("//") === -1:
									// Relative path to store, i.e. a child
									if (i.charAt(0) !== "/") i = self.uri + "/" + i;
									// Root path, relative to store, i.e. a domain
									else if (self.uri !== null && root.test(i)) {
										parsed = utility.parse(self.uri);
										i      = parsed.protocol + "//" + parsed.host + i;
									};
								default:
									idx = i.replace(/.*\//, "");
									if (idx.isEmpty()) break;
									utility.defer(function () {
										i.get(function (arg) { set(self.source === null ? arg : utility.walk(arg, self.source), idx); }, failure, utility.merge({withCredentials: self.credentials}, self.headers));
									});
									break;
							}
						});
					});
				});
			else data.sort(array.sort).reverse().each(function (i) {
				self.del(i, false, sync);
			});

			return this;
		},

		/**
		 * Clears the data object, unsets the uri property
		 *
		 * Events: beforeDataClear  Fires before the data is cleared
		 *         afterDataClear   Fires after the data is cleared
		 *
		 * @method clear
		 * @param {Boolean} sync [Optional] Boolean to limit clearing of properties
		 * @return {Object}      Data store
		 */
		clear : function (sync) {
			sync    = (sync === true);
			var obj = this.parentNode;

			if (!sync) {
				obj.fire("beforeDataClear");
				this.callback    = null;
				this.collections = [];
				this.crawled     = false;
				this.credentials = null;
				this.expires     = null;
				this._expires    = null;
				this.headers     = {Accept: "application/json"};
				this.ignore      = [];
				this.key         = null;
				this.keys        = {};
				this.leafs       = [];
				this.loaded      = false;
				this.pointer     = null;
				this.records     = [];
				this.recursive   = false;
				this.retrieve    = false;
				this.source      = null;
				this.total       = 0;
				this.views       = {};
				this.uri         = null;
				this._uri        = null;
				obj.fire("afterDataClear");
			}
			else {
				this.collections = [];
				this.crawled     = false;
				this.keys        = {};
				this.loaded      = false;
				this.records     = [];
				this.total       = 0;
				this.views       = {};
			}
			return this;
		},

		/**
		 * Crawls a record's properties and creates data stores when URIs are detected
		 *
		 * Events: afterDataRetrieve  Fires after the store has retrieved all data from crawling
		 * 
		 * @method crawl
		 * @param  {Mixed}  arg    Record key or index
		 * @param  {String} ignore [Optional] Comma delimited fields to ignore
		 * @param  {String} key    [Optional] data.key property to set on new stores, defaults to record.key
		 * @return {Object}        Record
		 */
		crawl : function (arg, ignore, key) {
			var ignored = false,
			    self    = this,
			    record;

			if (typeof arg !== "number" && typeof arg !== "string") throw Error(label.error.invalidArguments);

			this.crawled = true;
			this.loaded  = false;
			record       = this.get(arg);
			record       = this.records[this.keys[record.key]];
			key          = key || this.key;

			if (typeof ignore === "string") {
				ignored = true;
				ignore  = ignore.explode();
			}

			utility.iterate(record.data, function (v, k) {
				switch (true) {
					case ignored && ignore.contains(k):
					case !(v instanceof Array) && typeof v !== "string":
						return;
				}

				if (v instanceof Array) {
					// Possibly a subset of the collection, so it relies on valid URI paths
					if (!self.collections.contains(k)) self.collections.push(k);
					record.data[k] = data.factory({id: record.key + "-" + k}, null, {key: key, pointer: self.pointer, source: self.source});
					record.data[k].data.headers = utility.merge(record.data[k].data.headers, self.headers);
					if (ignored) ignore.each(function (i) { record.data[k].data.ignore.add(i); });
					self.leafs.each(function (i) { record.data[k].data.leafs.add(i); });
					if (!self.leafs.contains(k) && self.recursive && self.retrieve) {
						record.data[k].data.recursive = true;
						record.data[k].data.retrieve  = true;
					}

					if (v.length > 0) {
						record.data[k].once("afterDataBatch", function () { this.fire("afterDataRetrieve"); }, "dataRetrieve");
						record.data[k].data.batch("set", v, true);
					}
				}
				else {
					// If either condition is satisified it's assumed that "v" is a URI because it's not ignored
					if (v.charAt(0) === "/" || v.indexOf("//") > -1) {
						if (!self.collections.contains(k)) self.collections.push(k);
						record.data[k] = data.factory({id: record.key + "-" + k}, null, {key: key, pointer: self.pointer, source: self.source});
						record.data[k].once("afterDataSync", function () { this.fire("afterDataRetrieve"); }, "dataRetrieve");
						record.data[k].data.headers = utility.merge(record.data[k].data.headers, self.headers);
						if (ignored) ignore.each(function (i) { record.data[k].data.ignore.add(i); });
						self.leafs.each(function (i) { record.data[k].data.leafs.add(i); });
						if (!self.leafs.contains(k) && self.recursive && self.retrieve) {
							record.data[k].data.recursive = true;
							record.data[k].data.retrieve  = true;
						}
						typeof record.data[k].data.setUri === "function" ? record.data[k].data.setUri(v) : record.data[k].data.uri = v;
					}
				}
			});
			return this.get(arg);
		},

		/**
		 * Deletes a record based on key or index
		 *
		 * Events: beforeDataDelete  Fires before the record is deleted
		 *         afterDataDelete   Fires after the record is deleted
		 *         syncDataDelete    Fires when the local store is updated
		 *         failedDataDelete  Fires if the store is RESTful and the action is denied
		 *
		 * @method del
		 * @param  {Mixed}   record  Record key or index
		 * @param  {Boolean} reindex Default is true, will re-index the data object after deletion
		 * @param  {Boolean} sync    [Optional] True if called by data.sync
		 * @return {Object}          Data store
		 */
		del : function (record, reindex, sync) {
			if (typeof record === "undefined" || !/number|string/.test(typeof record)) throw Error(label.error.invalidArguments);

			reindex  = (reindex !== false);
			sync     = (sync === true);
			var obj  = this.parentNode,
			    r    = /true|undefined/,
			    key, args, uri, p;

			switch (typeof record) {
				case "string":
					key    = record;
					if (typeof key === "undefined") throw Error(label.error.invalidArguments);
					record = this.keys[key];
					break;
				default:
					key = this.records[record];
					if (typeof key === "undefined") throw Error(label.error.invalidArguments);
					key = key.key;
			}

			args   = {key: key, record: record, reindex: reindex};

			if (!sync && this.callback === null && this.uri !== null) {
				uri = this.uri + "/" + key;
				p   = uri.allows("delete");
			}

			obj.fire("beforeDataDelete", args);
			switch (true) {
				case sync:
				case this.callback !== null:
				case this.uri === null:
					obj.fire("syncDataDelete", args);
					break;
				case r.test(p):
					uri.del(function () { obj.fire("syncDataDelete", args); }, function () { obj.fire("failedDataDelete", args); }, utility.merge({withCredentials: this.credentials}, this.headers));
					break;
				default:
					obj.fire("failedDataDelete", args);
			}
			return this;
		},

		/**
		 * Finds needle in the haystack
		 *
		 * @method find
		 * @param  {Mixed}  needle    String, Number or Pattern to test for
		 * @param  {String} haystack  [Optional] Commma delimited string of the field(s) to search
		 * @param  {String} modifiers [Optional] Regex modifiers, defaults to "gi" unless value is null
		 * @return {Array}            Array of results
		 */
		find : function (needle, haystack, modifiers) {
			if (typeof needle === "undefined") throw Error(label.error.invalidArguments);

			needle     = typeof needle   === "string" ? needle.explode() : [needle];
			haystack   = typeof haystack === "string" ? haystack.explode() : null;

			var result = [],
			    obj    = this.parentNode,
			    keys   = [],
			    regex  = new RegExp();

			if (this.total === 0) return result;

			switch (true) {
				case typeof modifiers === "undefined":
				case String(modifiers).isEmpty():
					modifiers = "gi";
					break;
				case modifiers === null:
					modifiers = "";
					break;
			}

			// No haystack, testing everything
			if (haystack === null) {
				this.records.each(function (r) {
					utility.iterate(r.data, function (v, k) {
						if (keys.contains(r.key)) return false;
						if (typeof v.data === "object") return;

						needle.each(function (n) {
							utility.compile(regex, n, modifiers);
							if (regex.test(v)) {
								keys.push(r.key);
								result.add(r);
								return false;
							}
						});
					});
				});
			}
			// Looking through the haystack
			else this.records.each(function (r) {
				haystack.each(function (h) {
					if (keys.contains(r.key)) return false;
					if (typeof r.data[h] === "undefined" || typeof r.data[h].data === "object") return;

					needle.each(function (n) {
						utility.compile(regex, n, modifiers);
						if (regex.test(r.data[h])) {
							keys.push(r.key);
							result.add(r);
							return false;
						}
					});
				});
			});

			return result;
		},

		/**
		 * Generates a micro-format form from a record
		 * 
		 * If record is null, an empty form based on the first record is generated.
		 * The submit action is data.set() which triggers a POST or PUT
		 * from the data store.
		 * 
		 * @method form
		 * @param  {Mixed}   record null, record, key or index
		 * @param  {Object}  target Target HTML Element
		 * @param  {Boolean} test   [Optional] Test form before setting values
		 * @return {Object}         Generated HTML form
		 */
		form : function (record, target, test) {
			test      = (test !== false);
			var empty = (record === null),
			    self  = this,
			    entity, obj, handler, structure, key, data;

			switch (true) {
				case empty:
					record = this.get(0);
					break;
				case !(record instanceof Object):
					record = this.get(record);
					break;
			}

			switch (true) {
				case typeof record === "undefined":
					throw Error(label.error.invalidArguments);
				case this.uri !== null && !this.uri.allows("post"): // POST & PUT are interchangable for this bit
					throw Error(label.error.serverInvalidMethod);
			}

			key  = record.key;
			data = record.data;

			if (typeof target !== "undefined") target = utility.object(target);
			if (this.uri !== null) {
				entity = this.uri.replace(/.*\//, "").replace(/\?.*/, "")
				if (entity.isDomain()) entity = entity.replace(/\..*/g, "");
			}
			else entity = "record";

			/**
			 * Button handler
			 * 
			 * @method handler
			 * @param  {Object} event Window event
			 * @return {Undefined}    undefined
			 */
			handler = function (event) {
				var form    = event.srcElement.parentNode,
				    nodes   = $("#" + form.id + " input"),
				    entity  = nodes[0].name.match(/(.*)\[/)[1],
				    result  = true,
				    newData = {},
				    guid;

				self.parentNode.fire("beforeDataFormSubmit");

				if (test) result = form.validate();

				switch (result) {
					case false:
						self.parentNode.fire("failedDataFormSubmit");
						break;
					case true:
						nodes.each(function (i) {
							if (typeof i.type !== "undefined" && /button|submit|reset/.test(i.type)) return;
							utility.define(i.name.replace("[", ".").replace("]", ""), i.value, newData);
						});
						guid = utility.genId(true);
						self.parentNode.once("afterDataSet", function () { form.destroy(); });
						self.set(key, newData[entity]);
						break;
				}

				self.parentNode.fire("afterDataFormSubmit", key);
			};

			/**
			 * Data structure in micro-format
			 * 
			 * @method structure
			 * @param  {Object} record Data store record
			 * @param  {Object} obj    [description]
			 * @param  {String} name   [description]
			 * @return {Undefined}     undefined
			 */
			structure = function (record, obj, name) {
				var x, id;
				utility.iterate(record, function (v, k) {
					switch (true) {
						case v instanceof Array:
							x = 0;
							v.each(function (o) { structure(o, obj, name + "[" + k + "][" + (x++) + "]"); });
							break;
						case v instanceof Object:
							structure(v, obj, name + "[" + k + "]");
							break;
						default:
							id = (name + "[" + k + "]").replace(/\[|\]/g, "");
							obj.create("label", {"for": id}).html(k.capitalize());
							obj.create("input", {id: id, name: name + "[" + k + "]", type: "text", value: empty ? "" : v});
					}
				});
			};

			this.parentNode.fire("beforeDataForm");
			obj = element.create("form", {style: "display:none;"}, target);
			structure(data, obj, entity);
			obj.create("input", {type: "button", value: label.common.submit}).on("click", function(e) { handler(e); });
			obj.create("input", {type: "reset", value: label.common.reset});
			obj.css("display", "inherit");
			this.parentNode.fire("afterDataForm", obj);
			return obj;
		},

		/**
		 * Generates a RESTful store (replacing a record) when consuming an API end point
		 *
		 * @param  {Object} key  Record key
		 * @param  {String} uri  [Optional] Related URI
		 * @return {Object}      Data store
		 */
		generate : function (key, uri) {
			var params, idx;
			
			params = {
				headers   : this.headers,
				ignore    : array.clone(this.ignore),
				leafs     : array.clone(this.leafs),
				key       : this.key,
				pointer   : this.pointer,
				recursive : this.recursive,
				retrieve  : this.retrieve,
				source    : this.source
			};

			// Create stub or teardown existing data store
			if (typeof this.keys[key] !== "undefined") {
				idx = this.keys[key];
				if (typeof this.records[idx].data.teardown === "function") this.records[idx].data.teardown();
			}
			else {
				this.set(key, {}, true);
				idx = this.keys[key];
				this.collections.add(key);
			}

			// Creating new child data store
			this.records[idx] = data.factory({id: this.parentNode.id + "-" + key}, null, params);

			// Conditionally making the store RESTful
			if (this.uri !== null && typeof uri === "undefined") {
				uri = this.uri + "/" + key;
				typeof this.records[idx].data.setUri === "function" ? this.records[idx].data.setUri(uri) : this.records[idx].data.uri = uri;
			}

			return this.records[idx].data;
		},

		/**
		 * Retrieves a record based on key or index
		 *
		 * If the key is an integer, cast to a string before sending as an argument!
		 *
		 * @method get
		 * @param  {Mixed}  record Key, index or Array of pagination start & end; or comma delimited String of keys or indices
		 * @param  {Number} offset [Optional] Offset from `record` for pagination
		 * @return {Mixed}         Individual record, or Array of records
		 */
		get : function (record, offset) {
			var records = this.records,
			    obj     = this.parentNode,
			    type    = typeof record,
			    self    = this,
			    r;

			switch (true) {
				case type === "undefined":
				case String(record).length === 0:
					r = records;
					break;
				case type === "string" && record.indexOf(",") > -1:
					r = [];
					record.explode().each(function (i) {
						if (!isNaN(i)) i = parseInt(i);
						r.push(self.get(i));
					});
					break;
				case type === "string" && typeof this.keys[record] !== "undefined":
					r = records[this.keys[record]];
					break;
				case type === "number" && typeof offset === "undefined":
					r = records[parseInt(record)];
					break;
				case type === "number" && typeof offset === "number":
					r = records.limit(parseInt(record), parseInt(offset));
					break;
				default:
					r = undefined;
			}

			return r;
		},

		/**
		 * Reindexes the data store
		 *
		 * @method reindex
		 * @return {Object} Data store
		 */
		reindex : function () {
			var nth = this.total,
			    obj = this.parentNode,
			    key = (this.key !== null),
			    i   = -1;

			if (nth > 0 && nth < this.records.length) {
				this.views = {};
				while (++i < nth) {
					if (!key && this.records[i].key.isNumber()) {
						delete this.keys[this.records[i].key];
						this.records[i].key = i.toString();
					}
					this.keys[this.records[i].key] = i;
				}
			}
			return this;
		},

		/**
		 * Creates or updates an existing record
		 *
		 * If a POST is issued and the data.key property is not set, the
		 * URI is parsed for the key
		 *
		 * Events: beforeDataSet  Fires before the record is set
		 *         afterDataSet   Fires after the record is set, the record is the argument for listeners
		 *         syncDataSet    Fires when the local store is updated
		 *         failedDataSet  Fires if the store is RESTful and the action is denied
		 *
		 * @method set
		 * @param  {Mixed}   key  Integer or String to use as a Primary Key
		 * @param  {Object}  data Key:Value pairs to set as field values
		 * @param  {Boolean} sync [Optional] True if called by data.sync
		 * @return {Object}       The data store
		 */
		set : function (key, data, sync) {
			if (key === null) key = undefined;
			sync = (sync === true);

			switch (true) {
				case (typeof key === "undefined" || String(key).isEmpty()) && this.uri === null:
				case typeof data === "undefined":
					throw Error(label.error.invalidArguments);
				case data instanceof Array:
					return this.generate(key).batch("set", data, true);
				case data instanceof Number:
				case data instanceof String:
				case typeof data !== "object":
					throw Error(label.error.invalidArguments);
			}

			var record = typeof key === "undefined" ? undefined : this.get(key),
			    obj    = this.parentNode,
			    method = typeof key === "undefined" ? "post" : "put",
			    self   = this,
			    args   = {data: {}, key: key, record: undefined},
			    uri    = this.uri,
			    r      = /true|undefined/,
			    p, success, failure;

			if (typeof record !== "undefined") {
				args.record = this.records[this.keys[record.key]];
				utility.iterate(args.record.data, function (v, k) {
					if (!self.collections.contains(k) && !self.ignore.contains(k)) args.data[k] = v;
				});
				utility.merge(args.data, data);
			}
			else args.data = data;

			success = function (arg) {
				args["result"] = arg;
				obj.fire("syncDataSet", args);
			};

			failure = function (e) {
				obj.fire("failedDataSet");
			};

			if (!sync && this.callback === null && uri !== null) {
				if (typeof record !== "undefined") uri += "/" + record.key;
				p = uri.allows(method);
			}

			obj.fire("beforeDataSet", {key: key, data: data});
			switch (true) {
				case sync:
				case this.callback !== null:
				case this.uri === null:
					obj.fire("syncDataSet", args);
					break;
				case r.test(p):
					uri[method](success, failure, data, utility.merge({withCredentials: this.credentials}, this.headers));
					break;
				default:
					obj.fire("failedDataSet", args);
			}
			return this;
		},

		/**
		 * Returns a view, or creates a view and returns it
		 *
		 * @method sort
		 * @param  {String} query       SQL (style) order by
		 * @param  {String} create      [Optional, default behavior is true, value is false] Boolean determines whether to recreate a view if it exists
		 * @param  {String} sensitivity [Optional] Sort sensitivity, defaults to "ci" (insensitive = "ci", sensitive = "cs", mixed = "ms")
		 * @return {Array}               View of data
		 */
		sort : function (query, create, sensitivity) {
			if (typeof query === "undefined" || String(query).isEmpty()) throw Error(label.error.invalidArguments);
			if (!/ci|cs|ms/.test(sensitivity)) sensitivity = "ci";

			create       = (create === true);
			var view     = (query.replace(/\s*asc/g, "").replace(/,/g, " ").toCamelCase()) + sensitivity.toUpperCase(),
			    queries  = query.explode(),
			    needle   = /:::(.*)$/,
			    asc      = /\s*asc$/i,
			    desc     = /\s*desc$/i,
			    nil      = /^null/,
			    key      = this.key,
			    result   = [],
			    bucket, sort, crawl;

			queries.each(function (query) { if (String(query).isEmpty()) throw Error(label.error.invalidArguments); });

			if (!create && this.views[view] instanceof Array) return this.views[view];
			if (this.total === 0) return [];

			crawl = function (q, data) {
				var queries = q.clone(),
				    query   = q.first(),
				    sorted  = {},
				    result  = [];

				queries.remove(0);
				sorted = bucket(query, data, desc.test(query));
				sorted.order.each(function (i) {
					if (sorted.registry[i].length < 2) return;
					if (queries.length > 0) sorted.registry[i] = crawl(queries, sorted.registry[i]);
				});
				sorted.order.each(function (i) { result = result.concat(sorted.registry[i]); });
				return result;
			}

			bucket = function (query, records, reverse) {
				query        = query.replace(asc, "");
				var prop     = query.replace(desc, ""),
				    pk       = (key === prop),
				    order    = [],
				    registry = {};

				records.each(function (r) {
					var val = pk ? r.key : r.data[prop],
					    k   = val === null ? "null" : String(val);

					switch (sensitivity) {
						case "ci":
							k = k.toCamelCase();
							break;
						case "cs":
							k = k.trim();
							break;
						case "ms":
							k = k.trim().slice(0, 1).toLowerCase();
							break;
					}

					if (!(registry[k] instanceof Array)) {
						registry[k] = [];
						order.push(k);
					}
					registry[k].push(r);
				});

				order.sort(array.sort);
				if (reverse) order.reverse();
				
				order.each(function (k) {
					if (registry[k].length === 1) return;
					registry[k] = sort(registry[k], query, prop, reverse, pk);
				});

				return {order: order, registry: registry};
			};

			sort = function (data, query, prop, reverse, pk) {
				var tmp    = [],
				    sorted = [];

				data.each(function (i, idx) {
					var v  = pk ? i.key : i.data[prop];

					v = String(v).trim() + ":::" + idx;
					tmp.push(v.replace(nil, "\"\""));
				});

				if (tmp.length > 1) {
					tmp.sort(array.sort);
					if (reverse) tmp.reverse();
				}

				tmp.each(function (v) { sorted.push(data[needle.exec(v)[1]]); });
				return sorted;
			};

			result           = crawl(queries, this.records);
			this.views[view] = result;
			return result;
		},

		/**
		 * Storage interface
		 * 
		 * @param  {Mixed}  obj  Record (Object, key or index) or store
		 * @param  {Object} op   Operation to perform (get, remove or set)
		 * @param  {String} type [Optional] Type of Storage to use (local or session, default is local)
		 * @return {Object}      Record or store
		 */
		storage : function (obj, op, type) {
			var record  = false,
			    session = (type === "session" && typeof sessionStorage !== "undefined"),
			    ns      = /number|string/,
			    result, key, data;

			if (!/number|object|string/.test(typeof obj) || !/get|remove|set/.test(op)) throw Error(label.error.invalidArguments);

			record = (ns.test(obj) || (obj.hasOwnProperty("key") && !obj.hasOwnProperty("parentNode")));
			if (record && !(obj instanceof Object)) obj = this.get(obj);
			key    = record ? obj.key : obj.parentNode.id;

			switch (op) {
				case "get":
					result = session ? sessionStorage.getItem(key) : localStorage.getItem(key);
					if (result === null) throw Error(label.error.invalidArguments);
					result = json.decode(result);
					record ? this.set(key, result, true) : utility.merge(this, result);
					result = record ? obj : this;
					break;
				case "remove":
					session ? sessionStorage.removeItem(key) : localStorage.removeItem(key);
					result = this;
					break;
				case "set":
					data = json.encode(record ? obj.data : {total: this.total, keys: this.keys, records: this.records});
					session ? sessionStorage.setItem(key, data) : localStorage.setItem(key, data);
					result = this;
					break;
			}

			return result;
		},

		/**
		 * Syncs the data store with a URI representation
		 *
		 * Events: beforeDataSync  Fires before syncing the data store
		 *         afterDataSync   Fires after syncing the data store
		 *         failedDataSync  Fires when an exception occurs
		 *
		 * @method sync
		 * @param {Boolean} reindex [Optional] True will reindex the data store
		 * @return {Object}         Data store
		 */
		sync : function (reindex) {
			if (this.uri === null || this.uri.isEmpty()) throw Error(label.error.invalidArguments);

			reindex  = (reindex === true);
			var self = this,
			    obj  = self.parentNode,
			    guid = utility.guid(true),
			    success, failure;

			success = function (arg) {
				try {
					if (typeof arg !== "object") throw Error(label.error.expectedObject);

					var data, found = false, guid = utility.genId(true);

					if (self.source !== null) arg = utility.walk(arg, self.source);

					if (arg instanceof Array) data = arg;
					else utility.iterate(arg, function (i) {
						if (!found && i instanceof Array) {
							found = true;
							data  = i;
						}
					});

					if (typeof data === "undefined") data = [arg];

					obj.once("afterDataBatch", function (arg) {
						if (reindex) self.reindex();
						this.un("failedDataBatch", guid).fire("afterDataSync", self.get());
					}, guid);

					obj.once("failedDataBatch", function (arg) {
						self.clear(true);
						this.un("afterDataBatch", guid).fire("failedDataSync");
					}, guid);

					self.batch("set", data, true);
				}
				catch (e) {
					error(e, arguments, this);
					obj.fire("failedDataSync", arg);
				}
			};

			failure = function (e) { obj.fire("failedDataSync", e); };

			obj.fire("beforeDataSync");
			this.callback !== null ? this.uri.jsonp(success, failure, {callback: this.callback}) : this.uri.get(success, failure, utility.merge({withCredentials: this.credentials}, this.headers));
			return this;
		},

		/**
		 * Tears down a store & expires all records associated to an API
		 * 
		 * @return {Undefined} undefined
		 */
		teardown : function () {
			var uri = this.uri,
			    records, id;

			if (uri !== null) {
				cache.expire(uri, true);
				observer.remove(uri);
				id = this.parentNode.id + "DataExpire";

				utility.clearTimers(id);

				records = this.get();
				records.each(function (i) {
					cache.expire((uri + "/" + i.key), true);
					observer.remove(uri + "/" + i.key);
					utility.iterate(i.data, function (v, k) {
						if (v === null) return;
						if (v.hasOwnProperty("data") && typeof v.data.teardown === "function") {
							observer.remove(v.id);
							v.data.teardown();
						}
					});
				});
			}
			this.clear(true);
			this.parentNode.fire("afterDataTeardown");
			return this;
		}
	},

	/**
	 * Registers a data store on an Object
	 *
	 * Events: beforeDataStore  Fires before registering the data store
	 *         afterDataStore   Fires after registering the data store
	 *
	 * @method register
	 * @param  {Object} obj  Object to register with
	 * @param  {Mixed}  data [Optional] Data to set with this.batch
	 * @param  {Object} args [Optional] Arguments to set on the store
	 * @return {Object}      Object registered with
	 */
	factory : function (obj, recs, args) {
		var methods = {
			expires : {
				getter : function () { return this._expires; },
				setter : function (arg) {
					// Expiry cannot be less than a second, and must be a valid scenario for consumption; null will disable repetitive expiration
					if ((arg !== null && this.uri === null) || (arg !== null && (isNaN(arg) || arg < 1000))) throw Error(label.error.invalidArguments);

					if (this._expires === arg) return;
					this._expires = arg;

					var id      = this.parentNode.id + "DataExpire",
					    expires = arg,
					    self    = this;

					utility.clearTimers(id);

					if (arg === null) return;

					utility.repeat(function () {
						if (self.uri === null) {
							typeof self.setExpires === "function" ? self.setExpires(null) : self.expires = null;
							return false;
						}
						if (!cache.expire(self.uri)) self.uri.fire("beforeExpire, expire, afterExpire");
					}, expires, id);
				}
			},
			uri : {
				getter : function () { return this._uri; },
				setter : function (arg) {
					if (arg !== null && arg.isEmpty()) throw Error(label.error.invalidArguments);

					switch (true) {
						case this._uri === arg:
							return;
						case this._uri !== null:
							this._uri.un();
						default:
							this._uri = arg;
					}

					if (arg !== null) {
						arg.on("expire", function () { this.sync(true); }, "dataSync", this);
						cache.expire(arg, true);
						this.sync();
					}
				}
			}
		};

		obj = utility.object(obj);
		utility.genId(obj);

		// Hooking observer if not present in prototype chain
		if (typeof obj.fire === "undefined") observer.hook(obj);

		obj.fire("beforeDataStore");

		obj.data = utility.extend(data.methods);
		obj.data.parentNode = obj; // Recursion, useful
		obj.data.clear();          // Setting properties

		if (args instanceof Object) utility.merge(obj.data, args);

		obj.on("syncDataDelete", function (data) {
			var record = this.get(data.record);

			this.records.remove(this.keys[data.key]);
			delete this.keys[data.key];
			this.total--;
			this.views = {};
			if (data.reindex) this.reindex();
			utility.iterate(record.data, function (v, k) {
				if (v === null) return;
				if (typeof v.data !== "undefined" && typeof v.data.teardown === "function") v.data.teardown();
			});
			this.parentNode.fire("afterDataDelete", record);
			return this.parentNode;
		}, "recordDelete", obj.data);

		obj.on("syncDataSet", function (arg) {
			var data = typeof arg.record === "undefined" ? utility.clone(arg) : arg,
			    fire = true,
			    self = this,
			    record, uri;

			switch (true) {
				case typeof data.record === "undefined":
					var index = this.total;
					this.total++;
					if (typeof data.key === "undefined") {
						if (typeof data.result === "undefined") {
							this.fire("failedDataSet");
							throw Error(label.error.expectedObject);
						}
						if (this.source !== null) data.result = utility.walk(data.result, this.source);
						if (this.key === null) data.key = array.cast(data.result).first();
						else {
							data.key = data.result[this.key];
							delete data.result[this.key];
						}
						if (typeof data.key !== "string") data.key = data.key.toString();
						data.data = data.result;
					}
					this.keys[data.key] = index;
					this.records[index] = {};
					record              = this.records[index];
					record.key          = data.key;
					if (this.pointer === null || typeof data.data[this.pointer] === "undefined") {
						record.data = data.data;
						if (this.key !== null && this.records[index].data.hasOwnProperty(this.key)) delete this.records[index].data[this.key];
					}
					else {
						fire = false;
						uri  = data.data[this.pointer];
						if (typeof uri === "undefined" || uri === null) {
							delete this.records[index];
							delete this.keys[data.key];
							this.fire("failedDataSet");
							throw Error(label.error.expectedObject);
						}
						record.data = {};
						uri.get(function (args) {
							if (self.source !== null) args = utility.walk(args, self.source);
							if (typeof args[self.key] !== "undefined") delete args[self.key];
							utility.merge(record.data, args);
							if (self.retrieve) {
								self.crawl(record.key, self.ignore.length > 0 ? self.ignore.join(",") : undefined, self.key);
								self.loaded = true;
							}
							self.parentNode.fire("afterDataSet", record);
						}, function () { self.parentNode.fire("failedDataSet"); }, self.headers);
					}
					break;
				default:
					record = this.records[this.keys[data.record.key]];
					utility.merge(record.data, data.data);
			}
			this.views = {};
			if (this.retrieve) this.crawl(record.key, this.ignore.length > 0 ? this.ignore.join(",") : undefined, this.key);
			if (fire) this.parentNode.fire("afterDataSet", record);
		}, "recordSet", obj.data);

		// Getters & setters
		switch (true) {
			case (!client.ie || client.version > 8):
				utility.property(obj.data, "uri",     {enumerable: true, get: methods.uri.getter,     set: methods.uri.setter});
				utility.property(obj.data, "expires", {enumerable: true, get: methods.expires.getter, set: methods.expires.setter});
				break;
			default: // Only exists when no getters/setters (IE8)
				obj.data.setExpires = function (arg) {
					obj.data.expires = arg;
					methods.expires.setter.call(obj.data, arg);
				};
				obj.data.setUri = function (arg) {
					obj.data.uri = arg;
					methods.uri.setter.call(obj.data, arg);
				};
		}

		if (typeof recs === "object" && recs !== null) obj.data.batch("set", recs);
		obj.fire("afterDataStore");
		return obj;
	}
};

/**
 * DataList
 *
 * Provides a reactive View of a DataStore
 * Requires a CSS class named "hidden" to toggle "display:none" of list items
 *
 * @class datalist
 * @namespace abaaso
 */
var datalist = {
	position : /top|bottom/,

	// Inherited by datalists
	methods : {
		/**
		 * Changes the page index of the DataList
		 * 
		 * @return {Object}  DataList instance
		 */
		page : function (arg) {
			if (isNaN(arg)) throw Error(label.error.invalidArguments);

			this.pageIndex = arg;
			this.refresh();
			return this;
		},

		/**
		 * Adds pagination Elements to the View
		 * 
		 * @return {Object}  DataList instance
		 */
		pages : function () {
			var obj   = this.element,
			    list  = $("#" + obj.id + "-pages-top, #" + obj.id + "-pages-bottom"),
			    page  = this.pageIndex,
			    pos   = this.pagination,
			    range = this.pageRange,
			    mid   = range.half().roundDown(),
			    start = page - mid,
			    end   = page + mid,
			    self  = this,
			    total = datalist.pages.call(this),
			    i     = 0,
			    diff, li, anchor;

			if (!datalist.position.test(pos)) throw Error(label.error.invalidArguments);

			// Removing the existing controls
			list.each(function (i) { if (typeof i !== "undefined") i.destroy(); });
			
			// Halting because there's 1 page, or nothing
			if (this.total === 0 || total === 1) return this;

			// Getting the range to display
			if (start < 1) {
				diff  = start.diff(1);
				start = start + diff;
				end   = end   + diff;
			}
			if (end > total) {
				end   = total;
				start = (end - range) + 1;
				if (start < 1) start = 1;
			}

			pos.explode().each(function (i) {
				// Setting up the list
				list = obj[i === "bottom" ? "after" : "before"]("ul", {"class": "list pages " + i, id: obj.id + "-pages-" + i});

				// First page
				page > 1 ? list.create("li").create("a", {"class": "first page", "data-page": 1}).html("&lt;&lt;").on("click", function () { self.page(this.data("page")); }, "click")
				         : list.create("li").create("span", {"class": "first page"}).html("&lt;&lt;");

				// Previous page
				page > 1 ? list.create("li").create("a", {"class": "prev page", "data-page": (page - 1)}).html("&lt;").on("click", function () { self.page(this.data("page")); }, "click")
				         : list.create("li").create("span", {"class": "prev page"}).html("&lt;");

				// Rendering the page range
				for (i = start; i <= end; i++) {
					i !== page ? list.create("li").create("a", {"class": i === page ? "current page" : "page", "data-page": i}).html(i).on("click", function (e) { self.page(this.data("page")); }, "click")
					           : list.create("li").create("span", {"class": "page"}).html(i);
				}

				// Next page
				((page + 1) <= total) ? list.create("li").create("a", {"class": "next page", "data-page": (page + 1)}).html("&gt;").on("click", function () { self.page(this.data("page")); }, "click")
				                      : list.create("li").create("span", {"class": "next page"}).html("&gt;");

				// Last page
				page < total ? list.create("li").create("a", {"class": "last page", "data-page": total}).html("&gt;&gt;").on("click", function () { self.page(this.data("page")); }, "click")
				             : list.create("li").create("span", {"class": "last page"}).html("&gt;&gt;");

				// Scroll to top the top
				list.find("a").on("click", function (e) { window.scrollTo(0, 0); });
			});

			return this;
		},

		/**
		 * Refreshes element
		 * 
		 * Events: beforeDataListRefresh  Fires from the element containing the DataList
		 *         afterDataListRefresh   Fires from the element containing the DataList
		 * 
		 * @param {Boolean} redraw [Optional] Boolean to force clearing the DataList (default), false toggles "hidden" class of items
		 * @return {Object}        DataList instance
		 */
		refresh : function (redraw) {
			redraw       = (redraw !== false);
			var element  = this.element,
			    template = (typeof this.template === "object"),
			    key      = (!template && String(this.template).replace(/\{\{|\}\}/g, "") === this.store.key),
			    consumed = [],
			    items    = [],
			    self     = this,
			    callback = (typeof this.callback === "function"),
			    cleanup  = /\{\{.*\}\}/g,
			    regex    = new RegExp(),
			    registry = [], // keeps track of records in the list (for filtering)
			    limit    = [],
			    fn, obj;

			if (!this.ready) {
				this.ready = true;
				this.store.parentNode.on("afterDataSet", function (r)  {
					if (datalist.garbage(this.store.parentNode, element.id, "afterDataSet", "set-" + element.id)) return;
					if (!this.refreshing) this.refresh();
				}, "set-" + element.id, this);
			}

			this.element.fire("beforeDataListRefresh");
			this.refreshing = true;

			// Creating templates for the html rep
			if (!template) fn = function (i) {
				var html = String(self.template);

				html = html.replace("{{" + self.store.key + "}}", i.key)
				utility.iterate(i.data, function (v, k) {
					regex.compile("{{" + k + "}}", "g");
					html = html.replace(regex, v);
				});
				return {li: html.replace(cleanup, self.placeholder)};
			}
			else fn = function (i) {
				var obj = json.encode(self.template);

				obj = obj.replace("{{" + self.store.key + "}}", i.key)
				json.iterate(i.data, function (v, k) {
					regex.compile("{{" + k + "}}", "g");
					obj = obj.replace(regex, json.encode(v).replace(/(^")|("$)/g, "")); // stripping first and last " to concat to valid JSON
				});
				obj = json.decode(obj.replace(cleanup, self.placeholder));
				return {li: obj};
			};

			// Consuming records based on sort
			consumed = this.order.isEmpty() ? this.store.get() : this.store.sort(this.order, false, this.sensitivity);

			// Processing (filtering) records & generating templates
			consumed.each(function (i) {
				if (self.filter === null || !(self.filter instanceof Object)) items.push({key: i.key, template: fn(i)});
				else {
					utility.iterate(self.filter, function (v, k) {
						if (registry.index(i.key) > -1) return;

						var x     = 0,
						    regex = new RegExp(),
						    nth;

						v   = String(v).explode();
						nth = v.length;

						for (x = 0; x < nth; x++) {
							regex.compile(v[x], "i");
							if ((k === self.store.key && regex.test(i.key)) || (typeof i.data[k] !== "undefined" && regex.test(i.data[k]))) {
								registry.push(i.key);
								items.push({key: i.key, template: fn(i)});
								return;
							}
						}
					});
				}
			});

			// Total count of items in the list
			this.total = items.length;

			// Pagination supporting filtering
			if (typeof this.pageIndex === "number" && typeof this.pageSize === "number") {
				limit = datalist.range.call(this);
				items = items.limit(limit[0], limit[1]);
			}

			// Preparing the target element
			if (redraw) {
				element.clear();
				items.each(function (i) {
					var obj = element.tpl(i.template);
					obj.data("key", i.key);
					if (callback) self.callback(obj);
				});
			}
			else {
				element.find("> li").addClass("hidden");
				items.each(function (i) { element.find("> li[data-key='" + i.key + "']").removeClass("hidden"); });
			}

			// Rendering pagination elements
			if (datalist.position.test(this.pagination) && typeof this.pageIndex === "number" && typeof this.pageSize === "number") this.pages();
			else {
				$("#" + this.element.id + "-pages-top, #" + this.element.id + "-pages-bottom");
				if (typeof obj !== "undefined") obj.destroy();
			}

			this.refreshing = false;
			this.element.fire("afterDataListRefresh", element);
			return this;
		},

		/**
		 * Sorts data list & refreshes element
		 * 
		 * Events: beforeDataListRefresh, afterDataListRefresh
		 * 
		 * @param  {String} order       SQL "order by" statement
		 * @param  {String} sensitivity [Optional] Defaults to "ci" ("ci" = insensitive, "cs" = sensitive, "ms" = mixed sensitive)
		 * @return {Object}              DataList instance
		 */
		sort : function (order, sensitivity) {
			if (typeof order !== "string") throw Error(label.error.invalidArguments);
			this.element.fire("beforeDataListSort");
			this.order       = order;
			this.sensitivity = sensitivity || "ci";
			this.refresh();
			this.element.fire("afterDataListSort");
			return this;
		},

		/**
		 * Tears down references to the DataList in the Observer
		 * 
		 * @return {Object}  DataList instance
		 */
		teardown : function () {
			var id = this.element.id;

			this.store.parentNode.un("afterDataDelete", "delete-" + id).un("afterDataRetrieve, afterDataSync", "refresh-" + id);
			return this;
		}
	},

	/**
	 * Creates an instance of datalist
	 *
	 * Events: beforeDataList  Fires before target receives the DataList
	 *         afterDataList   Fires after DataList is setup in target
	 *           
	 * @param  {Object} target   Element to receive the DataList
	 * @param  {Object} store    Data store to feed the DataList
	 * @param  {Mixed}  template Record field, template ($.tpl), or String, e.g. "<p>this is a {{field}} sample.</p>", fields are marked with {{ }}
	 * @param  {Object} options  Optional parameters to set on the DataList
	 * @return {Object}          DataList instance
	 */
	factory : function (target, store, template, options) {
		var ref      = [store],
		    instance = {},
		    params   = {},
		    element, fn;

		if (!(target instanceof Element) || typeof store !== "object" || !/string|object/.test(typeof template)) throw Error(label.error.invalidArguments);

		// Preparing
		element  = target.fire("beforeDataList").create("ul", {"class": "list", id: store.parentNode.id + "-datalist"});
		params   = {
			callback    : null,
			filter      : null,
			pageIndex   : 1,
			pageSize    : null,
			pageRange   : 5,
			pagination  : "bottom", // "top" or "bottom|top" are also valid
			placeholder : "",
			order       : "",
			ready       : false,
			refreshing  : false,
			template    : template,
			total       : 0,
			sensitivity : "ci"
		};

		// Creating instance
		instance         = utility.extend(datalist.methods, params);
		instance.element = element;
		instance.store   = ref[0];

		// Applying customization
		if (options instanceof Object) utility.iterate(options, function (v, k) { instance[k] = v; });

		// Cleaning up orphaned element(s)
		instance.store.parentNode.on("afterDataDelete", function (r) {
			this.element.fire("beforeDataListRefresh");
			if (datalist.garbage(this.store.parentNode, element.id, "afterDataDelete", "delete-" + element.id)) return;
			if (typeof this.pageIndex === "number" && typeof this.pageSize === "number") this.refresh();
			else this.element.find("> li[data-key='" + r.key + "']").destroy();
			this.element.fire("afterDataListRefresh");
		}, "delete-" + element.id, instance);

		fn = function () {
			var ev = this.store.retrieve ? "afterDataRetrieve" : "afterDataSync";

			this.store.parentNode.once(ev, function () {
				if (datalist.garbage(this.store.parentNode, element.id, ev, "refresh-" + element.id)) return;
				if (!this.refreshing && (ev === "afterDataRetrieve" || this.store.loaded)) this.refresh();
				fn.call(this);
			}, "refresh-" + element.id, this);

			if (datalist.now.call(instance)) instance.refresh();
		}

		datalist.now.call(instance) ? fn.call(instance) : instance.store.parentNode.once(instance.store.parentNode.retrieve ? "afterDataRetrieve" : "afterDataSync", fn, "initialize-" + element.id, instance);

		target.fire("afterDataList", element);

		return instance;
	},

	/**
	 * Garbage collector which removes invalid events from the observer
	 * 
	 * @param  {Object} obj     Data store parentNode
	 * @param  {Object} element DataList Element id
	 * @param  {String} event   Event
	 * @param  {String} id      Observerable event id
	 * @return {Boolean}        True if invalid & removed from the observer
	 */
	garbage : function (obj, element, event, id) {
		var result = false;

		if (typeof $("#" + element) === "undefined") {
			obj.un(event, id);
			result = true;
		}

		return result;
	},

	/**
	 * Determines if a store is ready to refresh
	 * 
	 * @return {Boolean} True to refresh
	 */
	now : function () {
		return (this.store.uri === null || this.store.loaded);
	},

	/**
	 * Calculates the total pages
	 * 
	 * @return {Number} Total pages
	 */
	pages : function () {
		if (isNaN(this.pageSize)) throw Error(label.error.invalidArguments);
		return (this.total / this.pageSize).roundUp();
	},

	/**
	 * Calculates the page size as an Array of start & finish
	 * 
	 * @return {Array}  Array of start & end numbers
	 */
	range : function () {
		var start = (this.pageIndex * this.pageSize) - this.pageSize,
		    end   = this.pageSize;

		return [start, end];
	}
};

/**
 * Element methods
 *
 * @class element
 * @namespace abaaso
 */
var element = {
	/**
	 * Gets or sets attributes of Element
	 * 
	 * @param  {Mixed}  obj   Element or $ query
	 * @param  {String} name  Attribute name
	 * @param  {Mixed}  value Attribute value
	 * @return {Object}       Element
	 */
	attr : function (obj, key, value) {
		if (typeof value === "string") value = value.trim();

		var regex = /checked|disabled/,
		    target;

		obj = utility.object(obj);

		if (!(obj instanceof Element) || typeof key == "undefined" || String(key).isEmpty()) throw Error(label.error.invalidArguments);

		switch (true) {
			case regex.test(key) && typeof value === "undefined":
				return obj[key];
			case regex.test(key) && typeof value !== "undefined":
				obj[key] = value;
				return obj;
			case obj.nodeName === "SELECT" && key === "selected" && typeof value === "undefined":
				return $("#" + obj.id + " option[selected=\"selected\"]").first() || $("#" + obj.id + " option").first();
			case obj.nodeName === "SELECT" && key === "selected" && typeof value !== "undefined":
				target = $("#" + obj.id + " option[selected=\"selected\"]").first();
				if (typeof target !== "undefined") {
					target.selected = false;
					target.removeAttribute("selected");
				}
				target = $("#" + obj.id + " option[value=\"" + value + "\"]").first();
				target.selected = true;
				target.setAttribute("selected", "selected");
				return obj;
			case typeof value === "undefined":
				return obj.getAttribute(key);
			case value === null:
				obj.removeAttribute(key);
				return obj;
			default:
				obj.setAttribute(key, value);
				return obj;
		}
	},

	/**
	 * Clears an object's innerHTML, or resets it's state
	 *
	 * Events: beforeClear  Fires before the Object is cleared
	 *         afterClear   Fires after the Object is cleared
	 *
	 * @method clear
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Element
	 */
	clear : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		obj.fire("beforeClear");
		switch (true) {
			case typeof obj.reset === "function":
				obj.reset();
				break;
			case typeof obj.value !== "undefined":
				obj.update({innerHTML: "", value: ""});
				break;
			default:
				obj.update({innerHTML: ""});
		}
		obj.fire("afterClear");
		return obj;
	},

	/**
	 * Creates an Element in document.body or a target Element
	 *
	 * An id is generated if not specified with args
	 *
	 * Events: beforeCreate  Fires before the Element has been created, but not set
	 *         afterCreate   Fires after the Element has been appended to it's parent
	 *
	 * @method create
	 * @param  {String} type   Type of Element to create
	 * @param  {Object} args   [Optional] Collection of properties to apply to the new element
	 * @param  {Mixed}  target [Optional] Target object or element.id value to append to
	 * @param  {Mixed}  pos    [Optional] "first", "last" or Object describing how to add the new Element, e.g. {before: referenceElement}
	 * @return {Object}        Element that was created or undefined
	 */
	create : function (type, args, target, pos) {
		if (typeof type === "undefined" || String(type).isEmpty()) throw Error(label.error.invalidArguments);

		var obj, uid, frag;

		switch (true) {
			case typeof target !== "undefined":
				target = utility.object(target);
				break;
			case typeof args !== "undefined" && (typeof args === "string" || typeof args.childNodes !== "undefined"):
				target = utility.object(args);
				break;
			default:
				target = document.body;
		}

		if (typeof target === "undefined") throw Error(label.error.invalidArguments);
		
		frag = !(target instanceof Element);
		uid  = typeof args !== "undefined"
		        && typeof args !== "string"
		        && typeof args.childNodes === "undefined"
		        && typeof args.id !== "undefined"
		        && typeof $("#" + args.id) === "undefined" ? args.id : utility.genId();

		if (typeof args !== "undefined" && typeof args.id !== "undefined") delete args.id;

		$.fire("beforeCreate", uid);
		if (!frag) target.fire("beforeCreate", uid);
		else if (frag && target.parentNode !== null) target.parentNode.fire("beforeCreate", uid);

		obj = !/svg/i.test(type) ? document.createElement(type) : document.createElementNS("http://www.w3.org/2000/svg", "svg");
		obj.id = uid;
		if (typeof args === "object" && typeof args.childNodes === "undefined") obj.update(args);
		switch (true) {
			case typeof pos === "undefined":
			case pos === "last":
				target.appendChild(obj);
				break;
			case pos === "first":
				target.prependChild(obj);
				break;
			case pos === "after":
				pos = {};
				pos.after = target;
				target    = target.parentNode;
			case typeof pos.after !== "undefined":
				target.insertBefore(obj, pos.after.nextSibling);
				break;
			case pos === "before":
				pos = {};
				pos.before = target;
				target     = target.parentNode;
			case typeof pos.before !== "undefined":
				target.insertBefore(obj, pos.before);
				break;
			default:
				target.appendChild(obj);
		}

		if (!frag) target.fire("afterCreate", obj);
		else if (frag && target.parentNode !== null) target.parentNode.fire("afterCreate", obj);
		$.fire("afterCreate", obj);
		
		return obj;
	},

	/**
	 * Gets or sets a CSS style attribute on an Element
	 *
	 * @method css
	 * @param  {Mixed}  obj   Element or $ query
	 * @param  {String} key   CSS to put in a style tag
	 * @param  {String} value [Optional] Value to set
	 * @return {Object}       Element
	 */
	css : function (obj, key, value) {
		obj = utility.object(obj);
		key = key.indexOf("-") > -1 ? string.toCamelCase(key, true) : key;

		var i, result;

		if (typeof value !== "undefined") {
			obj.style[key] = value;
			result = obj;
		}
		else result = obj.style[key];

		return result;
	},

	/**
	 * Data attribute facade acting as a getter (with coercion) & setter
	 *
	 * @method data
	 * @param  {Mixed}  obj   Element or $ query
	 * @param  {String} key   Data key
	 * @param  {Mixed}  value Boolean, Number or String to set
	 * @return {Mixed}        undefined, Element or value
	 */
	data : function (obj, key, value) {
		var result;

		obj = utility.object(obj);

		if (typeof value !== "undefined") {
			typeof obj.dataset === "object" ? obj.dataset[key] = value : element.attr(obj, "data-" + key, value);
			result = obj;
		}
		else result = utility.coerce(typeof obj.dataset === "object" ? obj.dataset[key] : element.attr(obj, "data-" + key));
		return result;
	},

	/**
	 * Destroys an Element
	 *
	 * Events: beforeDestroy  Fires before the destroy starts
	 *         afterDestroy   Fires after the destroy ends
	 *
	 * @method destroy
	 * @param  {Mixed} obj Element or $ query
	 * @return {Undefined} undefined
	 */
	destroy : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		$.fire("beforeDestroy", obj);
		obj.fire("beforeDestroy");
		observer.remove(obj.id);
		if (obj.parentNode !== null) obj.parentNode.removeChild(obj);
		obj.fire("afterDestroy");
		$.fire("afterDestroy", obj.id);
		return undefined;
	},

	/**
	 * Disables an Element
	 *
	 * Events: beforeDisable  Fires before the disable starts
	 *         afterDisable   Fires after the disable ends
	 *
	 * @method disable
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Element
	 */
	disable : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		if (typeof obj.disabled === "boolean" && !obj.disabled) {
			obj.fire("beforeDisable");
			obj.disabled = true;
			obj.fire("afterDisable");
		}
		return obj;
	},

	/**
	 * Enables an Element
	 *
	 * Events: beforeEnable  Fires before the enable starts
	 *         afterEnable   Fires after the enable ends
	 *
	 * @method enable
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Element
	 */
	enable : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		if (typeof obj.disabled === "boolean" && obj.disabled) {
			obj.fire("beforeEnable");
			obj.disabled = false;
			obj.fire("afterEnable");
		}
		return obj;
	},

	/**
	 * Finds descendant childNodes of Element matched by arg
	 *
	 * @method find
	 * @param  {Mixed}  obj Element to search
	 * @param  {String} arg Comma delimited string of descendant selectors
	 * @return {Mixed}      Array of Elements or undefined
	 */
	find : function (obj, arg) {
		var result = [];

		obj = utility.object(obj);
		if (!(obj instanceof Element) || typeof arg !== "string") throw Error(label.error.invalidArguments);
		utility.genId(obj);
		arg.explode().each(function (i) {
			 $("#" + obj.id + " " + i).each(function (o) {
			 	result.add(o);
			 });
		});
		return result;
	},

	/**
	 * Determines if Element has descendants matching arg
	 *
	 * @method has
	 * @param  {Mixed}   obj Element or Array of Elements or $ queries
	 * @param  {String}  arg Type of Element to find
	 * @return {Boolean}     True if 1 or more Elements are found
	 */
	has : function (obj, arg) {
		var result = element.find(obj, arg);
		return (!isNaN(result.length) && result.length > 0);
	},

	/**
	 * Determines if obj has a specific CSS class
	 * 
	 * @method hasClass
	 * @param  {Mixed} obj Element or $ query
	 * @return {Mixed}     Element, Array of Elements or undefined
	 */
	hasClass : function (obj, klass) {
		obj = utility.object(obj);
		return obj.classList.contains(klass);
	},

	/**
	 * Hides an Element if it's visible
	 *
	 * Events: beforeHide  Fires before the object is hidden
	 *         afterHide   Fires after the object is hidden
	 *
	 * @method hide
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Element
	 */
	hide : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		obj.fire("beforeHide");
		switch (true) {
			case typeof obj.hidden === "boolean":
				obj.hidden = true;
				break;
			default:
				obj["data-display"] = obj.style.display;
				obj.style.display = "none";
		}
		obj.fire("afterHide");
		return obj;
	},

	/**
	 * Returns a Boolean indidcating if the Object is hidden
	 *
	 * @method hidden
	 * @param  {Mixed} obj Element or $ query
	 * @return {Boolean}   True if hidden
	 */
	hidden : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		return obj.style.display === "none" || (typeof obj.hidden === "boolean" && obj.hidden);
	},

	/**
	 * Determines if Element is equal to arg, supports nodeNames & CSS2+ selectors
	 *
	 * @method is
	 * @param  {Mixed}   obj Element or $ query
	 * @param  {String}  arg Property to query
	 * @return {Boolean}     True if a match
	 */
	is : function (obj, arg) {
		obj = utility.object(obj);

		if (!(obj instanceof Element) || typeof arg !== "string") throw Error(label.error.invalidArguments);

		utility.genId(obj);
		return /^:/.test(arg) ? (element.find(obj.parentNode, obj.nodeName.toLowerCase() + arg).contains(obj)) : new RegExp(arg, "i").test(obj.nodeName);
	},

	/**
	 * Adds or removes a CSS class
	 *
	 * Events: beforeClassChange  Fires before the Object's class is changed
	 *         afterClassChange   Fires after the Object's class is changed
	 *
	 * @method clear
	 * @param  {Mixed}   obj Element or $ query
	 * @param  {String}  arg Class to add or remove (can be a wildcard)
	 * @param  {Boolean} add Boolean to add or remove, defaults to true
	 * @return {Object}      Element
	 */
	klass : function (obj, arg, add) {
		var classes;

		obj = utility.object(obj);

		if (!(obj instanceof Element) || String(arg).isEmpty()) throw Error(label.error.invalidArguments);

		obj.fire("beforeClassChange");
		add = (add !== false);
		arg = arg.explode(" ");
		if (add) arg.each(function (i) { obj.classList.add(i); });
		else arg.each(function (i) {
			if (i !== "*") obj.classList.remove(i);
			else {
				obj.classList.each(function (x) { this.remove(x); });
				return false;
			}
		});
		obj.fire("afterClassChange");
		return obj;
	},

	/**
	 * Finds the position of an element
	 *
	 * @method position
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Object {top: n, right: n, bottom: n, left: n}
	 */
	position : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		var left, top, height, width;

		left   = top = 0;
		width  = obj.offsetWidth;
		height = obj.offsetHeight;

		if (obj.offsetParent) {
			top    = obj.offsetTop;
			left   = obj.offsetLeft;

			while (obj = obj.offsetParent) {
				left += obj.offsetLeft;
				top  += obj.offsetTop;
			}
		}

		return {
			top    : top,
			right  : document.documentElement.clientWidth  - (left + width),
			bottom : document.documentElement.clientHeight + global.scrollY - (top + height),
			left   : left
		};
	},

	/**
	 * Prepends an Element to an Element
	 * 
	 * @method prependChild
	 * @param  {Object} obj   Element or $ query
	 * @param  {Object} child Child Element
	 * @return {Object}       Element
	 */
	prependChild : function (obj, child) {
		obj = utility.object(obj);

		if (!(obj instanceof Element) || !(child instanceof Element)) throw Error(label.error.invalidArguments);
		
		return obj.childNodes.length === 0 ? obj.appendChild(child) : obj.insertBefore(child, obj.childNodes[0]);
	},

	/**
	 * Shows an Element if it's not visible
	 *
	 * Events: beforeEnable  Fires before the object is visible
	 *         afterEnable   Fires after the object is visible
	 *
	 * @method show
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Element
	 */
	show : function (obj) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		obj.fire("beforeShow");
		switch (true) {
			case typeof obj.hidden === "boolean":
				obj.hidden = false;
				break;
			default:
				obj.style.display = obj.getAttribute("data-display") !== null ? obj.getAttribute("data-display") : "inherit";
		}
		obj.fire("afterShow");
		return obj;
	},

	/**
	 * Returns the size of the Object
	 *
	 * @method size
	 * @param  {Mixed} obj Element or $ query
	 * @return {Object}    Size {height: n, width:n}
	 */
	size : function (obj) {
		obj = utility.object(obj);

		var num, height, width;

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		/**
		 * Casts n to a number or returns zero
		 *
		 * @param  {Mixed} n The value to cast
		 * @return {Number}  The casted value or zero
		 */
		num = function (n) {
			return !isNaN(n) ? parseInt(n) : 0;
		};

		height = obj.offsetHeight + num(obj.style.paddingTop)  + num(obj.style.paddingBottom) + num(obj.style.borderTop)  + num(obj.style.borderBottom);
		width  = obj.offsetWidth  + num(obj.style.paddingLeft) + num(obj.style.paddingRight)  + num(obj.style.borderLeft) + num(obj.style.borderRight);

		return {height: height, width: width};
	},

	/**
	 * Getter / setter for an Element's text
	 * 
	 * @param  {Object} obj Element or $ query
	 * @param  {String} arg [Optional] Value to set
	 * @return {Object}     Element
	 */
	text : function (obj, arg) {
		obj = utility.object(obj);

		var key     = typeof obj.textContent !== "undefined" ? "textContent" : "innerText",
		    payload = {},
		    set     = false;

		if (typeof arg !== "undefined") {
			set          = true;
			payload[key] = arg;
		}

		return set ? element.update(obj, payload) : obj[key];
	},

	/**
	 * Toggles a CSS class
	 * 
	 * @param  {Object} obj Element, or $ query
	 * @param  {String} arg CSS class to toggle
	 * @return {Object}     Element
	 */
	toggleClass : function (obj, arg) {
		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		obj.classList.toggle(arg);
		return obj;
	},

	/**
	 * Updates an Element
	 *
	 * Events: beforeUpdate  Fires before the update starts
	 *         afterUpdate   Fires after the update ends
	 *
	 * @method update
	 * @param  {Mixed}  obj  Element or $ query
	 * @param  {Object} args Collection of properties
	 * @return {Object}      Element
	 */
	update : function (obj, args) {
		var regex;

		obj  = utility.object(obj);
		args = args || {};

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		obj.fire("beforeUpdate");
		regex = /innerHTML|innerText|textContent|type|src/;

		utility.iterate(args, function (v, k) {
			switch (true) {
				case regex.test(k):
					obj[k] = v;
					break;
				case k === "class":
					!v.isEmpty() ? obj.addClass(v) : obj.removeClass("*");
					break;
				case k.indexOf("data-") === 0:
					element.data(obj, k.replace("data-", ""), v);
					break;
				case "id":
					var o = observer.listeners;
					if (typeof o[obj.id] !== "undefined") {
						o[k] = utility.clone(o[obj.id]);
						delete o[obj.id];
					}
				default:
					obj.attr(k, v);
			}
		});

		obj.fire("afterUpdate");
		return obj;
	},

	/**
	 * Gets or sets the value of Element
	 *
	 * Events: beforeValue  Fires before the object receives a new value
	 *         afterValue   Fires after the object receives a new value
	 * 
	 * @param  {Mixed}  obj   Element or $ query
	 * @param  {Mixed}  value [Optional] Value to set
	 * @return {Object}       Element
	 */
	val : function (obj, value) {
		var output = null,
		    select = /^select$/i,
		    check  = /^(radio|checkbox)$/i,
		    items;

		obj = utility.object(obj);

		if (!(obj instanceof Element)) throw Error(label.error.invalidArguments);

		switch (true) {
			case typeof value === "undefined":
				switch (true) {
					case (check.test(obj.type)):
						if (obj.name.isEmpty()) throw Error(label.error.expectedProperty);
						items = $("input[name='" + obj.name + "']");
						items.each(function (i) {
							if (output !== null) return;
							if (i.checked) output = i.value;
						});
						break;
					case (select.test(obj.type)):
						output = obj.options[obj.selectedIndex].value;
						break;
					default:
						output = typeof obj.value !== "undefined" ? obj.value : element.text(obj);
				}
				if (typeof output === "string") output = output.trim();
				break;
			default:
				value = String(value);
				obj.fire("beforeValue");
				switch (true) {
					case (check.test(obj.type)):
						items = $("input[name='" + obj.name + "']");
						items.each(function (i) {
							if (i.value === value) {
								i.checked = true;
								output    = i;
								return false;
							}
						});
						break;
					case (select.test(obj.type)):
						obj.find("> *").each(function (i) {
							if (i.value === value) {
								i.selected = true;
								output     = i;
								return false;
							}
						});
						break;
					default:
						typeof obj.value !== "undefined" ? obj.value = value : element.text(obj, value);
				}
				obj.fire("afterValue");
				output = obj;
		}
		return output;
	}
};

/**
 * DataList Filter
 * 
 * @class filter
 * @namespace abaaso
 */
var filter = {
	methods : {
		/**
		 * Initiate all event listeners
		 *
		 * @returns {Undefined} undefined
		 */
		init : function () {
			this.element.on("keyup", this.update, "filter", this);
			return this;
		},

		/**
		 * Set the filters
		 * 
		 * Create an object based on comma separated key string
		 * 
		 * @param {String} fields Comma separated filters
		 * @returns {Undefined} undefined
		 */
		set : function (fields) {
			var obj = {};

			if (typeof fields !== "string" || String(fields).isEmpty()) throw Error($.label.error.invalidArguments);

			fields.explode().each(function (v) { obj[v] = ""; });
			this.filters = obj;
			return this;
		},

		/**
		 * Cancel all event listeners
		 *
		 * @returns {Undefined} undefined
		 */
		teardown : function () {
			this.element.un("keyup", "filter");
			return this;
		},

		/**
		 * Update the results list
		 *
		 * @returns {Undefined} undefined
		 */
		update : function (e) {
			var val = this.element.val();
			
			if (!val.isEmpty()) {
				utility.iterate(this.filters, function (v, k) { this[k] = "^" + val.escape().replace("\\*", ".*"); });
				this.datalist.filter = this.filters;
			}
			else this.datalist.filter = null;

			this.datalist.pageIndex = 1;
			this.datalist.refresh();
			return this;
		}
	},

	/**
	 * DataList filter factory
	 * 
	 * @param  {Object} obj      Element to receive the filter
	 * @param  {Object} datalist List Module linked to the data store required
	 * @param  {String} filters  Comma delimited string of fields to filter by
	 * @return {Object}          Filter instance
	 */
	factory : function (obj, datalist, filters) {
		var instance = {},
		    ref = [datalist];

		if (!(obj instanceof Element) || (typeof datalist !== "undefined" && typeof datalist.store === "undefined") || (typeof filters !== "string" || String(filters).isEmpty())) throw Error($.label.error.invalidArguments);

		instance = utility.extend(filter.methods, {filters: {}});
		instance.datalist = ref[0];
		instance.element  = obj;
		instance.set(filters);
		instance.element.on("afterValue", instance.update, "value", instance);
		return instance.init();
	}
};

/**
 * JSON methods
 *
 * @class json
 * @namespace abaaso
 */
var json = {
	/**
	 * Decodes the argument
	 *
	 * @method decode
	 * @param  {String}  arg    String to parse
	 * @param  {Boolean} silent [Optional] Silently fail
	 * @return {Mixed}          Entity resulting from parsing JSON, or undefined
	 */
	decode : function (arg, silent) {
		try {
			return JSON.parse(arg);
		}
		catch (e) {
			if (silent !== true) error(e, arguments, this);
			return undefined;
		}
	},

	/**
	 * Encodes the argument as JSON
	 *
	 * @method encode
	 * @param  {Mixed}   arg    Entity to encode
	 * @param  {Boolean} silent [Optional] Silently fail
	 * @return {String}         JSON, or undefined
	 */
	encode : function (arg, silent) {
		try {
			return JSON.stringify(arg);
		}
		catch (e) {
			if (silent !== true) error(e, arguments, this);
			return undefined;
		}
	}
};

/**
 * Labels for localization
 *
 * Override this with another language pack
 *
 * @class label
 * @namespace abaaso
 */
var label = {
	// Common labels
	common : {
		back    : "Back",
		cancel  : "Cancel",
		clear   : "Clear",
		close   : "Close",
		cont    : "Continue",
		create	: "Create",
		del     : "Delete",
		edit    : "Edit",
		find    : "Find",
		gen     : "Generate",
		go      : "Go",
		loading : "Loading",
		next    : "Next",
		login   : "Login",
		ran     : "Random",
		reset   : "Reset",
		save    : "Save",
		search  : "Search",
		submit  : "Submit"
	},

	// Days of the week
	day : {
		0 : "Sunday",
		1 : "Monday",
		2 : "Tuesday",
		3 : "Wednesday",
		4 : "Thursday",
		5 : "Friday",
		6 : "Saturday"
	},

	// Error messages
	error : {
		databaseNotOpen       : "Failed to open the Database, possibly exceeded Domain quota",
		databaseNotSupported  : "Client does not support local database storage",
		databaseWarnInjection : "Possible SQL injection in database transaction, use the &#63; placeholder",
		elementNotCreated     : "Could not create the Element",
		elementNotFound       : "Could not find the Element",
		expectedArray         : "Expected an Array",
		expectedArrayObject   : "Expected an Array or Object",
		expectedBoolean       : "Expected a Boolean value",
		expectedNumber        : "Expected a Number",
		expectedProperty      : "Expected a property, and it was not set",
		expectedObject        : "Expected an Object",
		invalidArguments      : "One or more arguments is invalid",
		invalidDate           : "Invalid Date",
		invalidFields         : "The following required fields are invalid: ",
		invalidRoute          : "The route could not be found",
		notAvailable          : "Requested method is not available",
		notSupported          : "This feature is not supported by this platform",
		propertyNotFound      : "Could not find the requested property",
		promiseResolved       : "The promise has been resolved: {{outcome}}",
		serverError           : "Server error has occurred",
		serverForbidden       : "Forbidden to access URI",
		serverInvalidMethod   : "Method not allowed",
		serverUnauthorized    : "Authorization required to access URI"
	},

	// Months of the Year
	month : {
		0  : "January",
		1  : "February",
		2  : "March",
		3  : "April",
		4  : "May",
		5  : "June",
		6  : "July",
		7  : "August",
		8  : "September",
		9  : "October",
		10 : "November",
		11 : "December"
	}
};

/**
 * Messaging between iframes
 *
 * @class abaaso
 * @namespace abaaso
 */
var message = {
	/**
	 * Clears the message listener
	 *
	 * @method clear
	 * @return {Object} abaaso
	 */
	clear : function (state) {
		if (typeof state === "undefined") state = "all";
		return $.un(global, "message", "message", state);
	},

	/**
	 * Posts a message to the target
	 *
	 * @method send
	 * @param  {Object} target Object to receive message
	 * @param  {Mixed}  arg    Entity to send as message
	 * @return {Object}        target
	 */
	send : function (target, arg) {
		try {
			target.postMessage(arg, "*");
		}
		catch (e) {
			error(e, arguments, this);
		}
		return target;
	},

	/**
	 * Sets a handler for recieving a message
	 *
	 * @method recv
	 * @param  {Function} fn Callback function
	 * @return {Object}      abaaso
	 */
	recv : function (fn, state) {
		if (typeof state === "undefined") state = "all";
		return $.on(global, "message", fn, "message", global, state);
	}
};

/**
 * Mouse tracking
 *
 * @class mouse
 * @namespace abaaso
 */
var mouse = {
	//Indicates whether mouse tracking is enabled
	enabled : false,

	// Indicates whether to try logging co-ordinates to the console
	log     : false,

	// Mouse coordinates
	diff    : {x: null, y: null},
	pos     : {x: null, y: null},
	prev    : {x: null, y: null},

	/**
	 * Enables or disables mouse co-ordinate tracking
	 *
	 * @method track
	 * @param  {Mixed} n Boolean to enable/disable tracking, or Mouse Event
	 * @return {Object}  abaaso.mouse
	 */
	track : function (e) {
		var m = abaaso.mouse,
		    e = "mousemove",
		    n = "tracking";

		if (server) return m;

		switch (true) {
			case typeof e === "object":
				var view = document[client.ie && client.version < 9 ? "documentElement" : "body"],
				    x    = e.pageX ? e.pageX : (view.scrollLeft + e.clientX),
				    y    = e.pageY ? e.pageY : (view.scrollTop  + e.clientY),
				    c    = false;

				if (m.pos.x !== x) c = true;
				$.mouse.prev.x = m.prev.x = Number(m.pos.x);
				$.mouse.pos.x  = m.pos.x  = x;
				$.mouse.diff.x = m.diff.x = m.pos.x - m.prev.x;

				if (m.pos.y !== y) c = true;
				$.mouse.prev.y = m.prev.y = Number(m.pos.y);
				$.mouse.pos.y  = m.pos.y  = y;
				$.mouse.diff.y = m.diff.y = m.pos.y - m.prev.y;

				if (c && m.log) utility.log(m.pos.x + " [" + m.diff.x + "], " + m.pos.y + " [" + m.diff.y + "]");
				break;
			case typeof e === "boolean":
				e ? observer.add(document, e, abaaso.mouse.track, n) : observer.remove(document, e, n);
				$.mouse.enabled = m.enabled = e;
				break;
		}
		return m;
	}
};

/**
 * Number methods
 *
 * @class number
 * @namespace abaaso
 */
var number = {
	/**
	 * Returns the difference of arg
	 *
	 * @method odd
	 * @param {Number} arg Number to compare
	 * @return {Number}    The absolute difference
	 */
	diff : function (num1, num2) {
		if (isNaN(num1) || isNaN(num2)) throw Error(label.error.expectedNumber);

		return Math.abs(num1 - num2);
	},

	/**
	 * Tests if an number is even
	 *
	 * @method even
	 * @param {Number} arg Number to test
	 * @return {Boolean}   True if even, or undefined
	 */
	even : function (arg) {
		return arg % 2 === 0;
	},

	/**
	 * Formats a Number to a delimited String
	 * 
	 * @method format
	 * @param  {Number} arg       Number to format
	 * @param  {String} delimiter [Optional] String to delimit the Number with
	 * @param  {String} every     [Optional] Position to insert the delimiter, default is 3
	 * @return {String}           Number represented as a comma delimited String
	 */
	format : function (arg, delimiter, every) {
		if (isNaN(arg)) throw Error(label.error.expectedNumber);

		arg       = arg.toString();
		delimiter = delimiter || ",";
		every     = every     || 3;

		var d = arg.indexOf(".") > -1 ? "." + arg.replace(/.*\./, "") : "",
		    a = arg.replace(/\..*/, "").split("").reverse(),
		    p = Math.floor(a.length / every),
		    i = 1, n, b;

		for (b = 0; b < p; b++) {
			n = i === 1 ? every : (every * i) + (i === 2 ? 1 : (i - 1));
			a.splice(n, 0, delimiter);
			i++;
		}

		a = a.reverse().join("");
		if (a.charAt(0) === delimiter) a = a.substring(1);
		return a + d;
	},

	/**
	 * Returns half of a, or true if a is half of b
	 * 
	 * @param  {Number} a Number to divide
	 * @param  {Number} b [Optional] Number to test a against
	 * @return {Mixed}    Boolean if b is passed, Number if b is undefined
	 */
	half : function (a, b) {
		return typeof b !== "undefined" ? ((a / b) === .5) : (a / 2);
	},

	/**
	 * Tests if a number is odd
	 *
	 * @method odd
	 * @param {Number} arg Number to test
	 * @return {Boolean}   True if odd, or undefined
	 */
	odd : function (arg) {
		return !(arg % 2 === 0);
	},

	/**
	 * Parses the number
	 * 
	 * @param  {Mixed}  arg  Number to parse
	 * @param  {Number} base Integer representing the base or radix
	 * @return {Number}      Integer or float
	 */
	parse : function (arg, base) {
		if (base === null) base = undefined;
		return isNaN(base) ? Number(arg) : parseInt(arg, base);
	},

	/**
	 * Rounds a number up or down
	 * 
	 * @param  {Number} arg       Float to round
	 * @param  {String} direction [Optional] "up" or "down", defaults to "down"
	 * @return {Number}           Rounded interger
	 */
	round : function (arg, direction) {
		if (String(arg).indexOf(".") < 0) return arg;
		if (!/down|up/.test(direction)) direction = "down";
		return Math[direction === "down" ? "floor" : "ceil"](arg);
	}
};

/**
 * Global Observer wired to a State Machine
 *
 * @class observer
 * @namespace abaaso
 */
var observer = {
	// Collection of listeners
	listeners : {},

	// Boolean indicating if events are logged to the console
	log : false,

	/**
	 * Adds a handler to an event
	 *
	 * @method add
	 * @param  {Mixed}    obj   Entity or Array of Entities or $ queries
	 * @param  {String}   event Event, or Events being fired (comma delimited supported)
	 * @param  {Function} fn    Event handler
	 * @param  {String}   id    [Optional / Recommended] The id for the listener
	 * @param  {String}   scope [Optional / Recommended] The id of the object or element to be set as 'this'
	 * @param  {String}   state [Optional] The state the listener is for
	 * @return {Mixed}          Entity, Array of Entities or undefined
	 */
	add : function (obj, event, fn, id, scope, state) {
		obj   = utility.object(obj);
		scope = scope || abaaso;
		state = state || abaaso.state.current;

		if (obj instanceof Array) return obj.each(function (i) { observer.add(i, event, fn, id, scope, state); });

		if (typeof event !== "undefined") event = event.explode();
		if (typeof id === "undefined" || !/\w/.test(id)) id = utility.guid(true);

		var instance = null,
		    l        = observer.listeners,
		    o        = observer.id(obj),
		    n        = false,
		    c        = abaaso.state.current,
		    globals  = /body|document|window/i,
		    allowed  = /click|error|key|mousedown|mouseup|submit/i,
		    item, add, reg, handler;

		if (typeof o === "undefined" || event === null || typeof event === "undefined" || typeof fn !== "function") throw Error(label.error.invalidArguments);

		handler = function (e, i) {
			if (!allowed.test(e.type)) utility.stop(e);
			observer.fire(obj, i, e);
		};

		event.each(function (i) {
			n = false;
			if (typeof l[o] === "undefined")                  l[o]           = {};
			if (typeof l[o][i] === "undefined" && (n = true)) l[o][i]        = {};
			if (typeof l[o][i][state] === "undefined")        l[o][i][state] = {};

			if (n) {
				switch (true) {
					case globals.test(o):
					case !/\//g.test(o) && o !== "abaaso":
						instance = obj;
						break;
					default:
						instance = null;
				}

				if (instance !== null && typeof instance !== "undefined" && i.toLowerCase() !== "afterjsonp" && (globals.test(o) || typeof instance.listeners === "function")) {
					add = (typeof instance.addEventListener === "function");
					reg = (typeof instance.attachEvent === "object" || add);
					if (reg) instance[add ? "addEventListener" : "attachEvent"]((add ? "" : "on") + i, function (e) { handler(e, i); }, false);
				}
			}

			item = {fn: fn, scope: scope};
			l[o][i][state][id] = item;
		});

		return obj;
	},

	/**
	 * Gets the Observer id of arg
	 *
	 * @method id
	 * @param  {Mixed}  Object or String
	 * @return {String} Observer id
	 * @private
	 */
	id : function (arg) {
		var id;

		switch (true) {
			case arg === abaaso:
				id = "abaaso";
				break;
			case arg === global:
				id = "window";
				break;
			case arg === !server && document:
				id = "document";
				break;
			case arg === !server && document.body:
				id = "body";
				break;
			default:
				id = typeof arg.id !== "undefined" ? arg.id : (typeof arg.toString === "function" ? arg.toString() : arg);
		}
		return id;
	},

	/**
	 * Fires an event
	 *
	 * @method fire
	 * @param  {Mixed}  obj   Entity or Array of Entities or $ queries
	 * @param  {String} event Event, or Events being fired (comma delimited supported)
	 * @return {Mixed}        Entity, Array of Entities or undefined
	 */
	fire : function (obj, event) {
		obj      = utility.object(obj);
		if (obj instanceof Array) return obj.each(function (i) { observer.fire(obj[i], event, array.cast(arguments).remove(1).remove(0)); });

		var o    = observer.id(obj),
		    a    = array.cast(arguments).remove(1).remove(0),
		    s    = abaaso.state.current,
		    log  = ($.observer.log || abaaso.observer.log),
		    c, l, list;

		if (typeof o === "undefined" || String(o).isEmpty() || typeof obj === "undefined" || typeof event === "undefined") throw Error(label.error.invalidArguments);

		event.explode().each(function (e) {
			if (log) utility.log(o + " firing " + e);
			list = observer.list(obj, e);
			l = list.all;
			if (typeof l !== "undefined") utility.iterate(l, function (i, k) {
				i.fn.apply(i.scope, a);
			});
			if (s !== "all") {
				l = list[s];
				if (typeof l !== "undefined") utility.iterate(l, function (i, k) {
					i.fn.apply(i.scope, a);
				});
			}
		});
		return obj;
	},

	/**
	 * Provides observer hooks on obj
	 * 
	 * @param  {Object} obj Object to receive hooks
	 * @return {Object}     Object that received hooks
	 */
	hook : function (obj) {
		obj.fire      = function () { observer.fire.apply(observer, [this].concat(array.cast(arguments))); return this; };
		obj.listeners = function (event) { return $.listeners.call(this, event); };
		obj.on        = function (event, listener, id, scope, standby) { return $.on.call(this, event, listener, id, scope, standby); };
		obj.once      = function (event, listener, id, scope, standby) { return $.once.call(this, event, listener, id, scope, standby); };
		obj.un        = function (event, id) { return $.un.call(this, event, id); };
		return obj;
	},

	/**
	 * Gets the listeners for an event
	 *
	 * @method list
	 * @param  {Mixed}  obj   Entity or Array of Entities or $ queries
	 * @param  {String} event Event being queried
	 * @return {Mixed}        Object or Array of listeners for the event
	 */
	list : function (obj, event) {
		obj   = utility.object(obj);
		var l = observer.listeners,
		    o = observer.id(obj),
		    r;

		switch (true) {
			case typeof l[o] === "undefined" && typeof event === "undefined":
				r = {};
				break;
			case typeof l[o] !== "undefined" && (typeof event === "undefined" || String(event).isEmpty()):
				r = l[o];
				break;
			case typeof l[o] !== "undefined" && typeof l[o][event] !== "undefined":
				r = l[o][event];
				break;
			default:
				r = {};
		}
		return r;
	},

	/**
	 * Adds a listener for a single execution
	 * 
	 * @method once
	 * @param  {Mixed}    obj   Entity or Array of Entities or $ queries
	 * @param  {String}   event Event being fired
	 * @param  {Function} fn    Event handler
	 * @param  {String}   id    [Optional / Recommended] The id for the listener
	 * @param  {String}   scope [Optional / Recommended] The id of the object or element to be set as 'this'
	 * @param  {String}   state [Optional] The state the listener is for
	 * @return {Mixed}          Entity, Array of Entities or undefined
	 */
	once : function (obj, event, fn, id, scope, state) {
		var guid = id || utility.genId();

		obj   = utility.object(obj);
		scope = scope || abaaso;
		state = state || abaaso.state.current;

		if (typeof obj === "undefined" || event === null || typeof event === "undefined" || typeof fn !== "function") throw Error(label.error.invalidArguments);

		if (obj instanceof Array) return obj.each(function (i) { observer.once(i, event, fn, id, scope, state); });

		observer.add(obj, event, function () {
			observer.remove(obj, event, guid, state);
			fn.apply(scope, arguments);
		}, guid, scope, state);

		return obj;
	},

	/**
	 * Removes listeners
	 *
	 * @method remove
	 * @param  {Mixed}  obj   Entity or Array of Entities or $ queries
	 * @param  {String} event [Optional] Event, or Events being fired (comma delimited supported)
	 * @param  {String} id    [Optional] Listener id
	 * @param  {String} state [Optional] The state the listener is for
	 * @return {Mixed}        Entity, Array of Entities or undefined
	 */
	remove : function (obj, event, id, state) {
		obj   = utility.object(obj);
		state = state || abaaso.state.current;

		if (obj instanceof Array) return obj.each(function (i) { observer.remove(i, event, id, state); });

		var instance = null,
		    l = observer.listeners,
		    o = observer.id(obj);

		switch (true) {
			case typeof o === "undefined":
			case typeof l[o] === "undefined":
				return obj;
		}

		if (typeof event === "undefined" || event === null) delete l[o];
		else event.explode().each(function (e) {
			if (typeof l[o][e] === "undefined") return obj;
			typeof id === "undefined" ? l[o][e][state] = {} : delete l[o][e][state][id];
		});

		return obj;
	}
};

/**
 * Promises/A
 *
 * @class promise
 * @namespace abaaso
 */
var promise = {
	// States of a promise
	state : {
		broken   : "failed",
		initial  : "unfulfilled",
		resolved : "fulfilled",
	},

	// Inherited by promises
	methods : {
		/**
		 * Registers a failure handler for a Promise
		 * 
		 * @param  {Function} fn Executed when/if promise is broken
		 * @return {Object}      Promise instance
		 */
		fail : function (fn) {
			return promise.vouch.call(this, promise.state.broken, fn);
		},

		/**
		 * Breaks a Promise
		 * 
		 * @return {String} arg Status of promise
		 */
		reject : function (arg) {
			return promise.resolve.call(this, promise.state.broken, arg);
		},

		/**
		 * Promise is resolved
		 * 
		 * @param  {Mixed} arg Promise outcome
		 * @return {Object}    Promise instance
		 */
		resolve : function (arg) {
			return promise.resolve.call(this, promise.state.resolved, arg);
		},

		/**
		 * Returns a boolean indicating state of the Promise
		 * 
		 * @return {Boolean} `true` if resolved
		 */
		resolved : function () {
			return (this.state === promise.state.broken || this.state === promise.state.resolved);
		},

		/**
		 * Returns status of a Promise
		 * 
		 * @return {String} Status of promise
		 */
		status : function () {
			return this.state;
		},

		/**
		 * Registers handler(s) for a Promise
		 * 
		 * @param  {Function} success Executed when/if promise is resolved
		 * @param  {Function} failure [Optional] Executed when/if promise is broken
		 * @return {Object}           New Promise instance
		 */
		then : function (success, failure) {
			var self     = this,
			    instance = promise.factory(),
			    fn;

			fn = function (yay) {
				var handler = yay ? success : failure,
				    result;

				try {
					if (typeof handler !== "function") throw self.outcome;
					result = handler(self.outcome);
					if (!yay && typeof instance.fulfilled === "function") self.state = promise.state.resolved
				}
				catch (e) {
					result     = e.message || e;
					self.state = promise.state.broken;
				}
				finally {
					if (typeof result !== "undefined") self.outcome = result;
					switch (true) {
						case instance.fulfilled === null && instance.error === null:
							result = self;
							break;
						case self.state === promise.state.resolved && instance.fulfilled !== null:
							result = instance.resolve(self.outcome);
							break;
						case self.state === promise.state.broken && instance.error !== null:
							result = instance.reject(self.outcome);
							break;
					}
				}

				self.state   = result.state;
				self.outcome = result.outcome;

				return self.outcome;
			};

			promise.vouch.call(this, promise.state.resolved, function () { fn(true); });
			if (typeof failure === "function") this.fail(function () { fn(false); });
			return instance;
		}
	},

	/**
	 * Promise factory
	 * 
	 * @return {Object} Instance of promise
	 */
	factory : function () {
		var instance = {},
		    params   = {};

		// Promise structure
		params = {
			error     : null,
			fulfilled : null,
			outcome   : null,
			state     : promise.state.initial
		};

		instance = utility.extend(promise.methods, params);
		return instance;
	},

	/**
	 * Resolves a Promise (fulfilled or failed)
	 * 
	 * @param  {String} state State to resolve
	 * @param  {String} val   Value to set
	 * @return {Object}       Promise instance
	 */
	resolve : function (state, val) {
		var handler = state === promise.state.resolved ? "fulfilled" : "error";

		if (this.state !== promise.state.initial) throw Error(label.error.promiseResolved.replace("{{outcome}}", this.outcome));

		this.state     = state;
		this.outcome   = val;

		// The state & outcome can mutate here
		this[handler](val);

		this.error     = null;
		this.fulfilled = null;

		if (typeof Object.freeze === "function") Object.freeze(this);

		return this;
	},

	/**
	 * Vouches for a state
	 * 
	 * @param  {String}   state Promise descriptor
	 * @param  {Function} fn    Function to execute
	 * @return {Object}         Promise instance
	 */
	vouch : function (state, fn) {
		if (String(state).isEmpty()) throw Error(label.error.invalidArguments);

		switch (this.state) {
			case promise.state.initial:
				this[state === promise.state.resolved ? "fulfilled" : "error"] = fn;
				break;
			case state:
				fn(this.outcome);
		}

		return this;
	}
};

/**
 * URI routing via hashtag
 *
 * Client side routes will be in routes.all
 * 
 * @class route
 * @namespace abaaso
 */
var route = {
	bang    : /\#|\!\//g,
	nget    : /^(head|options)$/i,
	methods : /^(all|delete|get|put|post|head|options)$/i,
	regex   : new RegExp(),

	// Routing listeners
	routes : {
		all : {
			all : {
				error : function () {
					if (!server) {
						if (route.hash() === "") return history.go(-1);
						else {
							utility.error(label.error.invalidRoute);
							if (abaaso.route.initial !== null) route.hash(abaaso.route.initial);
						}
					}
					else throw Error(label.error.invalidRoute);
				}
			},
			"delete" : {},
			get      : {},
			put      : {},
			post     : {}
		}
	},

	/**
	 * Determines which HTTP method to use
	 * 
	 * @param  {String} arg HTTP method
	 * @return {[type]}     HTTP method to utilize
	 */
	method : function (arg) {
		return route.methods.test(arg) ? arg.toLowerCase() : "all";
	},

	/**
	 * Deletes a route
	 * 
	 * @method del
	 * @param  {String} name  Route name
	 * @param  {String} verb  HTTP method
	 * @return {Mixed}        True or undefined
	 */
	del : function (name, verb, host) {
		if (typeof host === "undefined") host = "all";
		verb      = route.method(verb);
		var error = (name === "error");

		if ((error && verb !== "all") || (!error && route.routes[host][verb].hasOwnProperty(name))) {
			if (abaaso.route.initial === name) abaaso.route.initial = null;
			return (delete route.routes[host][verb][name]);
		}
		else throw Error(label.error.invalidArguments);
	},

	/**
	 * Getter / setter for the hashbang
	 * 
	 * @method hash
	 * @param  {String} arg Route to set
	 * @return {String}     Current route
	 */
	hash : function (arg) {
		var output = "";

		switch (true) {
			case server:
				break;
			case typeof arg === "undefined":
				output = document.location.hash.replace(route.bang, "");
				break;
			default:
				output = arg.replace(route.bang, "");
				document.location.hash = "!/" + output;
		}
		return output;
	},

	/**
	 * Creates a hostname entry in the routes table
	 * 
	 * @param  {String} arg Hostname to route
	 * @return {Object}     Routes for hostname
	 */
	hostname : function (arg) {
		if (!route.routes.hasOwnProperty(arg)) {
			route.routes[arg] = {
				all      : {},
				"delete" : {},
				get      : {},
				post     : {},
				put      : {}
			};
		}

		return route.routes[arg];
	},

	/**
	 * Initializes the routing by loading the initial OR the first route registered
	 * 
	 * @method init
	 * @return {String} Route being loaded
	 */
	init : function () {
		var val = document.location.hash;

		val.isEmpty() ? route.hash(abaaso.route.initial !== null ? abaaso.route.initial : array.cast(route.routes.all.all, true).remove("error").first()) : route.load(val);
		return val.replace(route.bang, "");
	},

	/**
	 * Lists all routes
	 * 
	 * @set list
	 * @param {String} verb  HTTP method
	 * @return {Mixed}       Hash of routes if not specified, else an Array of routes for a method
	 */
	list : function (verb, host) {
		if (typeof host === "undefined") host = "all";

		var result;

		switch (true) {
			case !server:
				result = array.cast(route.routes.all.all, true);
				break;
			case typeof verb !== "undefined" && route.routes.hasOwnProperty(host):
				result = array.cast(route.routes[host][route.method(verb)], true);
				break;
			default:
				result = [];
				if (route.routes.hasOwnProperty(host)) {
					utility.iterate(route.routes[host], function (v, k) {
						result[k] = [];
						utility.iterate(v, function (fn, r) { result[k].push(r); });
					});
				}
		}

		if (!server && host !== "all") {
			utility.iterate(route.routes.all, function (v, k) {
				if (typeof result[k] === "undefined") result[k] = [];
				utility.iterate(v, function (fn, r) { result[k].push(r); });
			});
		}

		return result;
	},

	/**
	 * Loads the hash into the view
	 * 
	 * @method load
	 * @param  {String} name  Route to load
	 * @param  {Object} arg   [Optional] HTTP response (node)
	 * @param  {String} req   [Optional] HTTP request (node)
	 * @param  {String} host  [Optional] Hostname to query
	 * @return {Mixed}        True or undefined
	 */
	load : function (name, res, req, host) {
		if (typeof req === "undefined")  req  = "all";
		if (typeof host === "undefined") host = "all";

		var active = "",
		    path   = "",
		    result = true,
		    verb   = route.method(req.method || req),
		    crawl, find;

		// Not a GET, but assuming the route is smart enough to strip the entity body
		if (route.nget.test(verb)) verb = "get";

		name = name.replace(route.bang, "");

		// Crawls the hostnames
		crawl = function (host, verb, name) {
			switch (true) {
				case typeof route.routes[host][verb][name] !== "undefined":
					active = name;
					path   = verb;
					break;
				case verb !== "all" && typeof route.routes[host].all[name] !== "undefined":
					active = name;
					path   = "all";
					break;
				default:
					utility.iterate(route.routes[host][verb], function (v, k) { return find(k, verb, name); });
					if (active.isEmpty() && verb !== "all") utility.iterate(route.routes[host].all, function (v, k) { return find(k, "all", name); });
			}
		};

		// Finds a match
		find = function (pattern, method, arg) {
			if (utility.compile(route.regex, "^" + pattern + "$", "i") && route.regex.test(arg)) {
				active = pattern;
				path   = method;
				return false;
			}
		};

		if (host !== "all" && !route.routes.hasOwnProperty(host)) host = "all";

		crawl(host, verb, name);

		if (active.isEmpty()) {
			if (host !== "all") crawl("all", verb, name);
			if (active.isEmpty()) {
				host   = "all";
				active = "error";
				path   = "all";
				result = false;
			}
		}

		route.routes[host][path][active](res || active, req);
		return result;
	},

	/**
	 * Creates a Server with URI routing
	 * 
	 * @method server
	 * @param  {Object}   arg  Server options
	 * @param  {Function} fn   Error handler
	 * @param  {Boolean}  ssl  Determines if HTTPS server is created
	 * @return {Object}        Server
	 */
	server : function (args, fn, ssl) {
		var handler, err, obj;

		if (!server) return null;

		args = args || {};
		ssl  = (ssl === true || args.port === 443);

		// Request handler
		handler = function (req, res) {
			var parsed   = url.parse(req.url),
			    hostname = req.headers.host.replace(/:.*/, "");

			route.load(parsed.pathname, res, req, hostname);
		};

		// Error handler
		err = function (e) {
			error(e, this, arguments);
			if (typeof fn === "function") fn(e);
		};

		if (!server) throw Error(label.error.notSupported);

		// Enabling routing, in case it's not explicitly enabled prior to route.server()
		$.route.enabled = abaaso.route.enabled = true;

		// Server parameters
		args.host = args.host           || undefined;
		args.port = parseInt(args.port) || 8000;

		// Creating server
		obj = !ssl ? http.createServer(handler).on("error", err).listen(args.port, args.host)
		           : https.createServer(args, handler).on("error", err).listen(args.port);

		return obj;
	},

	/**
	 * Sets a route for a URI
	 * 
	 * @method set
	 * @param  {String}   name  Regex pattern for the route
	 * @param  {Function} fn    Route listener
	 * @param  {String}   verb  HTTP method the route is for (default is GET)
	 * @return {Mixed}          True or undefined
	 */
	set : function (name, fn, verb, host) {
		host = server ? (host || "all")    : "all";
		verb = server ? route.method(verb) : "all";
		if (typeof name !== "string" || name.isEmpty() || typeof fn !== "function") throw Error(label.error.invalidArguments);
		route.hostname(host)[verb][name] = fn;
		return true;
	}
};

/**
 * String methods
 * 
 * @class string
 * @namespace abaaso
 */
var string = {
	/**
	 * Capitalizes the String
	 * 
	 * @param  {String} obj String to capitalize
	 * @return {String}     Capitalized String
	 */
	capitalize : function (obj) {
		obj = string.trim(obj);
		return obj.charAt(0).toUpperCase() + obj.slice(1);
	},

	/**
	 * Escapes meta characters within a string
	 * 
	 * @param  {String} obj String to escape
	 * @return {String}     Escaped string
	 */
	escape : function (obj) {
		return obj.replace(/[-[\]{}()*+?.,\\^$|#\s]/g, "\\$&");
	},

	/**
	 * Splits a string on comma, or a parameter, and trims each value in the resulting Array
	 * 
	 * @param  {String} obj String to capitalize
	 * @param  {String} arg String to split on
	 * @return {Array}      Array of the exploded String
	 */
	explode : function (obj, arg) {
		if (typeof arg === "undefined" || arg.toString() === "") arg = ",";
		return string.trim(obj).split(new RegExp("\\s*" + arg + "\\s*"));
	},

	/**
	 * Replaces all spaces in a string with dashes
	 * 
	 * @param  {String} obj   String to hyphenate
	 * @param {Boolean} camel [Optional] Hyphenate camelCase
	 * @return {String}       String with dashes instead of spaces
	 */
	hyphenate : function (obj, camel) {
		var result;

		camel = (camel === true);

		result = string.trim(obj).replace(/\s+/g, "-");
		if (camel) result = result.replace(/([A-Z])/g, "-\$1").toLowerCase();
		return result;
	},

	/**
	 * Returns singular form of the string
	 * 
	 * @param  {String} obj String to transform
	 * @return {String}     Transformed string
	 */
	singular : function (obj) {
		return /s$/.test(obj) ? obj.slice(0, -1) : obj;
	},

	/**
	 * Transforms the case of a String into CamelCase
	 * 
	 * @param  {String} obj String to capitalize
	 * @return {String}     Camel case String
	 */
	toCamelCase : function (obj) {
		var s = string.trim(obj).toLowerCase().split(/\s|-/),
		    r = [];

		s.each(function (i, idx) {
			i = string.trim(i);
			if (i.isEmpty()) return;
			r.push(idx === 0 ? i : string.capitalize(i));
		});
		return r.join("");
	},

	/**
	 * Trims the whitespace around a String
	 * 
	 * @param  {String} obj String to capitalize
	 * @return {String}     Trimmed String
	 */
	trim : function (obj) {
		return obj.replace(/^\s+|\s+$/g, "");
	},

	/**
	 * Uncapitalizes the String
	 * 
	 * @param  {String} obj String to capitalize
	 * @return {String}     Uncapitalized String
	 */
	uncapitalize : function (obj) {
		obj = string.trim(obj);
		return obj.charAt(0).toLowerCase() + obj.slice(1);
	}
};

/**
 * Utilities
 *
 * @class utility
 * @namespace abaaso
 */
var utility = {
	// Collection of timers
	timer : {},

	/**
	 * Queries the DOM using CSS selectors and returns an Element or Array of Elements
	 * 
	 * Accepts comma delimited queries
	 *
	 * @method $
	 * @param  {String}  arg      Comma delimited string of target #id, .class, tag or selector
	 * @param  {Boolean} nodelist [Optional] True will return a NodeList (by reference) for tags & classes
	 * @return {Mixed}            Element or Array of Elements
	 */
	$ : function (arg, nodelist) {
		var result = [],
		    tmp    = [],
		    obj, sel;

		// Blocking node or DOM query of unique URIs via $.on()
		if (server || String(arg).indexOf("?") > -1) return undefined;

		arg      = arg.trim();
		nodelist = (nodelist === true);

		// Recursive processing, ends up below
		if (arg.indexOf(",") > -1) arg = arg.explode();
		if (arg instanceof Array) {
			arg.each(function (i) { tmp.push($(i, nodelist)); });
			tmp.each(function (i) { result = result.concat(i); });
			return result;
		}

		// Getting Element(s)
		switch (true) {
			case (/\s|>/.test(arg)):
				sel = arg.split(" ").filter(function (i) { if (i.trim() !== "" && i !== ">") return true; }).last();
				obj = document[sel.indexOf("#") > -1 && sel.indexOf(":") === -1 ? "querySelector" : "querySelectorAll"](arg);
				break;
			case arg.indexOf("#") === 0 && arg.indexOf(":") === -1:
				obj = isNaN(arg.charAt(1)) ? document.querySelector(arg) : document.getElementById(arg.substring(1));
				break;
			case arg.indexOf("#") > -1 && arg.indexOf(":") === -1:
				obj = document.querySelector(arg);
				break;
			default:
				obj = document.querySelectorAll(arg);
		}

		// Transforming obj if required
		if (typeof obj !== "undefined" && obj !== null && !(obj instanceof Element) && !nodelist) obj = array.cast(obj);
		if (obj === null) obj = undefined;

		return obj;
	},

	/**
	 * Aliases origin onto obj
	 *
	 * @method alias
	 * @param  {Object} obj    Object receiving aliasing
	 * @param  {Object} origin Object providing structure to obj
	 * @return {Object}        Object receiving aliasing
	 */
	alias : function (obj, origin) {
		var o = obj,
		    s = origin;

		utility.iterate(s, function (v, k) {
			var getter, setter;

			switch (true) {
				case !(v instanceof RegExp) && typeof v === "function":
					o[k] = v.bind(o[k]);
					break;
				case !(v instanceof RegExp) && !(v instanceof Array) && v instanceof Object:
					if (typeof o[k] === "undefined") o[k] = {};
					utility.alias(o[k], s[k]);
					break;
				default:
					getter = function () { return s[k]; },
					setter = function (arg) { s[k] = arg; };
					utility.property(o, k, {enumerable: true, get: getter, set: setter, value: s[k]});
			}
		});
		return obj;
	},

	/**
	 * Clears deferred & repeating functions
	 * 
	 * @param  {String} id ID of timer(s)
	 * @return {Undefined} undefined
	 */
	clearTimers : function (id) {
		if (typeof id === "undefined" || String(id).isEmpty()) throw Error(label.error.invalidArguments);

		// deferred
		if (typeof utility.timer[id] !== "undefined") {
			clearTimeout(utility.timer[id]);
			delete utility.timer[id];
		}

		// repeating
		if (typeof $.repeating[id] !== "undefined") {
			clearTimeout($.repeating[id]);
			delete $.repeating[id];
		}
	},

	/**
	 * Clones an Object
	 *
	 * @method clone
	 * @param {Object}  obj Object to clone
	 * @return {Object}     Clone of obj
	 */
	clone : function (obj) {
		var clone;

		switch (true) {
			case obj instanceof Array:
				return [].concat(obj);
			case typeof obj === "boolean":
				return Boolean(obj);
			case typeof obj === "function":
				return obj;
			case typeof obj === "number":
				return Number(obj);
			case typeof obj === "string":
				return String(obj);
			case !client.ie && !server && obj instanceof Document:
				return xml.decode(xml.encode(obj));
			case obj instanceof Object:
				clone = json.decode(json.encode(obj));
				if (typeof clone !== "undefined") {
					if (obj.hasOwnProperty("constructor")) clone.constructor = obj.constructor;
					if (obj.hasOwnProperty("prototype"))   clone.prototype   = obj.prototype;
				}
				return clone;
			default:
				return obj;
		}
	},

	/**
	 * Coerces a String to a Type
	 * 
	 * @param  {String} value String to coerce
	 * @return {Mixed}        Typed version of the String
	 */
	coerce : function (value) {
		var result = utility.clone(value),
		    tmp;

		switch (true) {
			case (/^\d$/.test(result)):
				result = number.parse(result);
				break;
			case (/^(true|false)$/i.test(result)):
				result = /true/i.test(result);
				break;
			case result === "undefined":
				result = undefined;
				break;
			case result === "null":
				result = null;
				break;
			case (tmp = json.decode(result, true)) && typeof tmp !== "undefined":
				result = tmp;
				break;
		}
		return result;
	},

	/**
	 * Recompiles a RegExp by reference
	 *
	 * This is ideal when you need to recompile a regex for use within a conditional statement
	 * 
	 * @param  {Object} regex     RegExp
	 * @param  {String} pattern   Regular expression pattern
	 * @param  {String} modifiers Modifiers to apply to the pattern
	 * @return {Boolean}          true
	 */
	compile : function (regex, pattern, modifiers) {
		regex.compile(pattern, modifiers);
		return true;
	},

	/**
	 * Creates a CSS stylesheet in the View
	 *
	 * @method css
	 * @param  {String} content CSS to put in a style tag
	 * @param  {String} media   [Optional] Medias the stylesheet applies to
	 * @return {Object}         Element created or undefined
	 */
	css : function (content, media) {
		var ss, css;

		ss = $("head")[0].create("style", {type: "text/css", media: media || "print, screen"});
		if (ss.styleSheet) ss.styleSheet.cssText = content;
		else {
			css = document.createTextNode(content);
			ss.appendChild(css);
		}
		return ss;
	},

	/**
	 * Allows deep setting of properties without knowing
	 * if the structure is valid
	 *
	 * @method define
	 * @param  {String} args  Dot delimited string of the structure
	 * @param  {Mixed}  value Value to set
	 * @param  {Object} obj   Object receiving value
	 * @return {Object}       Object receiving value
	 */
	define : function (args, value, obj) {
		var p   = obj,
		    nth = args.length;

		args  = args.split(".");
		if (typeof obj   === "undefined") obj   = this;
		if (typeof value === "undefined") value = null;
		if (obj === $) obj = abaaso;

		args.each(function (i, idx) {
			var num = idx + 1 < nth && !isNaN(parseInt(args[idx + 1])),
			    val = value;

			if (!isNaN(parseInt(i))) i = parseInt(i);
			
			// Creating or casting
			switch (true) {
				case typeof p[i] === "undefined":
					p[i] = num ? [] : {};
					break;
				case p[i] instanceof Object && num:
					p[i] = array.cast(p[i]);
					break;
				case p[i] instanceof Object:
					break;
				case p[i] instanceof Array && !num:
					p[i] = p[i].toObject();
					break;
				default:
					p[i] = {};
			}

			// Setting reference or value
			idx + 1 === nth ? p[i] = val : p = p[i];
		});

		return obj;
	},

	/**
	 * Defers the execution of Function by at least the supplied milliseconds
	 * Timing may vary under "heavy load" relative to the CPU & client JavaScript engine
	 *
	 * @method defer
	 * @param  {Function} fn Function to defer execution of
	 * @param  {Number}   ms Milliseconds to defer execution
	 * @param  {Number}   id [Optional] ID of the deferred function
	 * @return {Object}      undefined
	 */
	defer : function (fn, ms, id) {
		var op;

		ms = ms || 0;
		id = id || utility.guid(true);

		op = function () {
			delete utility.timer[id];
			fn();
		};

		utility.timer[id] = setTimeout(op, ms);
		return undefined;
	},

	/**
	 * Encodes a GUID to a DOM friendly ID
	 *
	 * @method domId
	 * @param  {String} GUID
	 * @return {String} DOM friendly ID
	 * @private
	 */
	domId : function (arg) {
		return "a" + arg.replace(/-/g, "").slice(1);
	},

	/**
	 * Error handling, with history in .log
	 *
	 * @method error
	 * @param  {Mixed}   e       Error object or message to display
	 * @param  {Array}   args    Array of arguments from the callstack
	 * @param  {Mixed}   scope   Entity that was "this"
	 * @param  {Boolean} warning [Optional] Will display as console warning if true
	 * @return {Undefined}       undefined
	 */
	error : function (e, args, scope, warning) {
		var o;

		if (typeof e !== "undefined") {
			warning = (warning === true);
			o = {
				arguments : args,
				message   : typeof e.message !== "undefined" ? e.message : e,
				number    : typeof e.number !== "undefined" ? (e.number & 0xFFFF) : undefined,
				scope     : scope,
				stack     : typeof e.stack === "string" ? e.stack : undefined,
				timestamp : new Date().toUTCString(),
				type      : typeof e.type !== "undefined" ? e.type : "TypeError"
			};

			if (typeof console !== "undefined") {
				console[!warning ? "error" : "warn"](o.message);
				if (typeof o.stack !== "undefined") console[!warning ? "error" : "warn"](o.stack);
			}
			$.error.log.push(o);
			$.fire("error", o);
		}

		return undefined;
	},

	/**
	 * Creates a class extending obj, with optional decoration
	 *
	 * @method extend
	 * @param  {Object} obj Object to extend
	 * @param  {Object} arg [Optional] Object for decoration
	 * @return {Object}     Decorated obj
	 */
	extend : function (obj, arg) {
		var o, f;

		if (typeof obj === "undefined") throw Error(label.error.invalidArguments);
		if (typeof arg === "undefined") arg = {};

		switch (true) {
			case typeof Object.create === "function":
				o = Object.create(obj);
				break;
			default:
				f = function () {};
				f.prototype = obj;
				o = new f();
		}

		utility.merge(o, arg);
		return o;
	},

	/**
	 * Generates an ID value
	 *
	 * @method genId
	 * @param  {Mixed} obj [Optional] Object to receive id
	 * @return {Mixed}     Object or id
	 */
	genId : function (obj) {
		var id;

		if (typeof obj !== "undefined") {
			switch (true) {
				case typeof obj.id !== "undefined" && obj.id !== "":
				case obj instanceof Array:
				case obj instanceof String || typeof obj === "string":
					return obj;
			}
		}

		do id = utility.domId(utility.guid(true));
		while (typeof $("#" + id) !== "undefined");

		if (typeof obj === "object") {
			obj.id = id;
			return obj;
		}
		else return id;
	},

	/**
	 * Generates a GUID
	 *
	 * @method guid
	 * @param {Boolean} safe [Optional] Strips - from GUID
	 * @return {String}      GUID
	 */
	guid : function (safe) {
		var s = function () { return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1); },
		    r = [8, 9, "a", "b"],
		    o;

		safe  = (safe === true);

		o = (s() + s() + "-" + s() + "-4" + s().substr(0, 3) + "-" + r[Math.floor(Math.random() * r.length)] + s().substr(0, 3) + "-" + s() + s() + s()).toLowerCase();
		if (safe) o = o.replace(/-/gi, "");
		return o;
	},

	/**
	 * Iterates an Object and executes a function against the properties
	 *
	 * Iteration can be stopped by returning false from fn
	 * 
	 * @param  {Object}   obj Object to iterate
	 * @param  {Function} fn  Function to execute against properties
	 * @return {Object}       Object
	 */
	iterate : function (obj, fn) {
		var has = Object.prototype.hasOwnProperty,
		    i, result;

		if (typeof fn !== "function") throw Error(label.error.invalidArguments);

		for (i in obj) {
			if (has.call(obj, i)) {
				result = fn.call(obj, obj[i], i);
				if (result === false) break;
			}
		}
		return obj;
	},

	/**
	 * Renders a loading icon in a target element,
	 * with a class of "loading"
	 *
	 * @method loading
	 * @param  {Mixed} obj Entity or Array of Entities or $ queries
	 * @return {Mixed}     Entity, Array of Entities or undefined
	 */
	loading : function (obj) {
		var l = abaaso.loading;

		obj = utility.object(obj);
		if (obj instanceof Array) return obj.each(function (i) { utility.loading(i); });

		if (l.url === null) throw Error(label.error.elementNotFound);

		if (typeof obj === "undefined") throw Error(label.error.invalidArguments);

		// Setting loading image
		if (typeof l.image === "undefined") {
			l.image     = new Image();
			l.image.src = l.url;
		}

		// Clearing target element
		obj.clear();

		// Creating loading image in target element
		obj.create("div", {"class": "loading"}).create("img", {alt: label.common.loading, src: l.image.src});

		return obj;
	},

	/**
	 * Writes argument to the console
	 *
	 * @method log
	 * @param  {String} arg String to write to the console
	 * @return undefined;
	 * @private
	 */
	log : function (arg) {
		var ts = !server || typeof arg !== "object";

		try {
			console.log((ts ? "[" + new Date().toLocaleTimeString() + "] " : "") + arg);
		}
		catch (e) {
			error(e, arguments, this);
		}
		return undefined;
	},

	/**
	 * Merges obj with arg
	 * 
	 * @param  {Object} obj Object to decorate
	 * @param  {Object} arg Object to decorate with
	 * @return {Object}     Object to decorate
	 */
	merge : function (obj, arg) {
		utility.iterate(arg, function (v, k) {
			obj[k] = utility.clone(v);
		});
		return obj;
	},
	
	/**
	 * Registers a module in the abaaso namespace
	 *
	 * IE8 will have factories (functions) duplicated onto $ because it will not respect the binding
	 * 
	 * @method module
	 * @param  {String} arg Module name
	 * @param  {Object} obj Module structure
	 * @return {Object}     Module registered
	 */
	module : function (arg, obj) {
		if (typeof $[arg] !== "undefined" || typeof abaaso[arg] !== "undefined" || !obj instanceof Object) throw Error(label.error.invalidArguments);
		
		abaaso[arg] = obj;
		if (typeof obj === "function") $[arg] = !client.ie || client.version > 8 ? abaaso[arg].bind($[arg]) : abaaso[arg];
		else {
			$[arg] = {};
			utility.alias($[arg], abaaso[arg]);
		}
		return $[arg];
	},

	/**
	 * Returns Object, or reference to Element
	 *
	 * @method object
	 * @param  {Mixed} obj Entity or $ query
	 * @return {Mixed}     Entity
	 * @private
	 */
	object : function (obj) {
		return typeof obj === "object" ? obj : (obj.toString().charAt(0) === "#" ? $(obj) : obj);
	},

	/**
	 * Parses a URI into an Object
	 * 
	 * @param  {String} uri URI to parse
	 * @return {Object}     Parsed URI
	 */
	parse : function (uri) {
		var obj    = {},
		    parsed = {};

		if (!server) {
			obj = document.createElement("a");
			obj.href = uri;
		}
		else obj = url.parse(uri);

		parsed = {
			protocol : obj.protocol,
			hostname : obj.hostname,
			port     : !String(obj.port).isEmpty() ? parseInt(obj.port) : "",
			pathname : obj.pathname,
			search   : obj.search,
			hash     : obj.hash,
			host     : obj.host,
		};

		// 'cause IE is ... IE; required for data.batch()
		if (client.ie) {
			if (parsed.protocol === ":")           parsed.protocol = location.protocol;
			if (parsed.hostname.isEmpty())         parsed.hostname = location.hostname;
			if (parsed.host.isEmpty())             parsed.host     = location.host;
			if (parsed.pathname.charAt(0) !== "/") parsed.pathname = "/" + parsed.pathname;
		}

		return parsed;
	},

	/**
	 * Sets a property on an Object, if defineProperty cannot be used the value will be set classically
	 * 
	 * @method property
	 * @param  {Object} obj        Object to decorate
	 * @param  {String} prop       Name of property to set
	 * @param  {Object} descriptor Descriptor of the property
	 * @return {Object}            Object receiving the property
	 */
	property : function (obj, prop, descriptor) {
		var define;

		if (!(descriptor instanceof Object)) throw Error(label.error.invalidArguments);

		define = (!client.ie || client.version > 8) && typeof Object.defineProperty === "function";
		if (define && typeof descriptor.value !== "undefined" && typeof descriptor.get !== "undefined") delete descriptor.value;
		define ? Object.defineProperty(obj, prop, descriptor) : obj[prop] = descriptor.value;
		return obj;
	},

	/**
	 * Sets methods on a prototype object
	 *
	 * @method proto
	 * @param  {Object} obj  Object receiving prototype extension
	 * @param  {String} type Identifier of obj, determines what Arrays to apply
	 * @return {Object}      obj or undefined
	 * @private
	 */
	proto : function (obj, type) {
		// Collection of methods to add to prototypes
		var i,
		    methods = {
			array   : {add      : function (arg) { return array.add(this, arg); },
			           addClass : function (arg) { return this.each(function (i) { i.addClass(arg); }); },
			           after    : function (type, args) { var a = []; this.each(function (i) { a.push(i.after(type, args)); }); return a; },
			           append   : function (type, args) { var a = []; this.each(function (i) { a.push(i.append(type, args)); }); return a; },
			           attr     : function (key, value) { var a = []; this.each(function (i) { a.push(i.attr(key, value)); }); return a; },
			           before   : function (type, args) { var a = []; this.each(function (i) { a.push(i.before(type, args)); }); return a; },
			           chunk    : function (size) { return array.chunk(this, size); },
			           clear    : function () { return !server && this[0] instanceof Element ? this.each(function (i) { i.clear(); }) : array.clear(this); },
			           clone    : function () { return utility.clone(this); },
			           collect  : function (arg) { return array.collect(this, arg); },
			           compact  : function () { return array.compact(this); },
			           contains : function (arg) { return array.contains(this, arg); },
			           create   : function (type, args, position) { var a = []; this.each(function (i) { a.push(i.create(type, args, position)); }); return a; },
			           css      : function (key, value) { return this.each(function (i) { i.css(key, value); }); },
			           data     : function (key, value) { var a = []; this.each(function (i) { a.push(i.data(key, value)); }); return a; },
			           diff     : function (arg) { return array.diff(this, arg); },
			           disable  : function () { return this.each(function (i) { i.disable(); }); },
			           destroy  : function () { this.each(function (i) { i.destroy(); }); return []; },
			           each     : function (arg) { return array.each(this, arg); },
			           empty    : function () { return array.empty(this); },
			           enable   : function () { return this.each(function (i) { i.enable(); }); },
			           equal    : function (arg) { return array.equal(this, arg); },
			           fill     : function (arg, start, offset) { return array.fill(this, arg, start, offset); },
			           find     : function (arg) { var a = []; this.each(function (i) { i.find(arg).each(function (r) { if (!a.contains(r)) a.push(r); }); }); return a; },
			           fire     : function () {
			           		var args = arguments;
			           		return this.each(function (i) { i.fire.apply(i, args); });
			           	},
			           first    : function () { return array.first(this); },
			           flat     : function () { return array.flat(this); },
			           get      : function (uri, headers) { this.each(function (i) { i.get(uri, headers); }); return []; },
			           has      : function (arg) { var a = []; this.each(function (i) { a.push(i.has(arg)); }); return a; },
			           hasClass : function (arg) { var a = []; this.each(function (i) { a.push(i.hasClass(arg)); }); return a; },
			           hide     : function () { return this.each(function (i){ i.hide(); }); },
			           html     : function (arg) {
			           		if (typeof arg !== "undefined") return this.each(function (i){ i.html(arg); });
			           		else {
			           			var a = []; this.each(function (i) { a.push(i.html()); }); return a;
			           		}
			           },
			           index    : function (arg) { return array.index(this, arg); },
			           indexed  : function () { return array.indexed(this); },
			           intersect: function (arg) { return array.intersect(this, arg); },
			           is       : function (arg) { var a = []; this.each(function (i) { a.push(i.is(arg)); }); return a; },
			           isAlphaNum: function () { var a = []; this.each(function (i) { a.push(i.isAlphaNum()); }); return a; },
			           isBoolean: function () { var a = []; this.each(function (i) { a.push(i.isBoolean()); }); return a; },
			           isChecked: function () { var a = []; this.each(function (i) { a.push(i.isChecked()); }); return a; },
			           isDate   : function () { var a = []; this.each(function (i) { a.push(i.isDate()); }); return a; },
			           isDisabled: function () { var a = []; this.each(function (i) { a.push(i.isDisabled()); }); return a; },
			           isDomain : function () { var a = []; this.each(function (i) { a.push(i.isDomain()); }); return a; },
			           isEmail  : function () { var a = []; this.each(function (i) { a.push(i.isEmail()); }); return a; },
			           isEmpty  : function () { var a = []; this.each(function (i) { a.push(i.isEmpty()); }); return a; },
			           isHidden : function () { var a = []; this.each(function (i) { a.push(i.isHidden()); }); return a; },
			           isIP     : function () { var a = []; this.each(function (i) { a.push(i.isIP()); }); return a; },
			           isInt    : function () { var a = []; this.each(function (i) { a.push(i.isInt()); }); return a; },
			           isNumber : function () { var a = []; this.each(function (i) { a.push(i.isNumber()); }); return a; },
			           isPhone  : function () { var a = []; this.each(function (i) { a.push(i.isPhone()); }); return a; },
			           isUrl    : function () { var a = []; this.each(function (i) { a.push(i.isUrl()); }); return a; },
			           keep_if  : function (fn) { return array.keep_if(this, fn); },
			           keys     : function () { return array.keys(this); },
			           last     : function (arg) { return array.last(this, arg); },
			           limit    : function (start, offset) { return array.limit(this, start, offset); },
			           listeners: function (event) { var a = []; this.each(function (i) { a = a.concat(i.listeners(event)); }); return a; },
			           loading  : function () { return this.each(function (i) { i.loading(); }); },
			           max      : function () { return array.max(this); },
			           mean     : function () { return array.mean(this); },
			           median   : function () { return array.median(this); },
			           min      : function () { return array.min(this); },
			           mode     : function () { return array.mode(this); },
			           on       : function (event, listener, id, scope, state) { return this.each(function (i) { i.on(event, listener, id, typeof scope !== "undefined" ? scope : i, state); }); },
			           once     : function (event, listener, id, scope, state) { return this.each(function (i) { i.once(event, listener, id, typeof scope !== "undefined" ? scope : i, state); }); },
			           position : function () { var a = []; this.each(function (i) { a.push(i.position()); }); return a; },
			           prepend  : function (type, args) { var a = []; this.each(function (i) { a.push(i.prepend(type, args)); }); return a; },
			           range    : function () { return array.range(this); },
			           rassoc   : function (arg) { return array.rassoc(this, arg); },
			           reject   : function (fn) { return array.reject(this, fn); },
			           remove   : function (start, end) { return array.remove(this, start, end); },
			           remove_if: function (fn) { return array.remove_if(this, fn); },
			           remove_while: function (fn) { return array.remove_while(this, fn); },
			           removeClass: function (arg) { return this.each(function (i) { i.removeClass(arg); }); },
			           replace  : function (arg) { return array.replace(this, arg); },
			           rest     : function (arg) { return array.rest(this, arg); },
			           rindex   : function (arg) { return array.rindex(this, arg); },
			           rotate   : function (arg) { return array.rotate(this, arg); },
			           series   : function (start, end, offset) { return array.series(start, end, offset); },
			           show     : function () { return this.each(function (i){ i.show(); }); },
			           size     : function () { var a = []; this.each(function (i) { a.push(i.size()); }); return a; },
			           split    : function (size) { return array.split(this, size); },
			           sum      : function () { return array.sum(this); },
			           take     : function (arg) { return array.take(this, arg); },
			           text     : function (arg) {
			           		return this.each(function (node) {
			           			if (typeof node !== "object") node = utility.object(node);
			           			if (typeof node.text === "function") node.text(arg);
			           		});
			           },
			           tpl      : function (arg) { return this.each(function (i) { i.tpl(arg); }); },
			           toggleClass: function (arg) { return this.each(function (i) { i.toggleClass(arg); }); },
			           total    : function () { return array.total(this); },
			           toObject : function () { return array.toObject(this); },
			           un       : function (event, id, state) { return this.each(function (i) { i.un(event, id, state); }); },
			           unique   : function () { return array.unique(this); },
			           update   : function (arg) { return this.each(function (i) { element.update(i, arg); }); },
			           val      : function (arg) {
			           		var a    = [],
			           		    type = null,
			           		    same = true;

			           		this.each(function (i) {
			           			if (type !== null) same = (type === i.type);
			           			type = i.type;
			           			if (typeof i.val === "function") a.push(i.val(arg));
			           		});
			           		return same ? a[0] : a;
			           	},
			           validate : function () { var a = []; this.each(function (i) { a.push(i.validate()); }); return a; },
			           zip      : function () { return array.zip(this, arguments); }},
			element : {addClass : function (arg) {
			           		utility.genId(this);
			           		return element.klass(this, arg, true);
			           },
			           after    : function (type, args) {
			           		utility.genId(this);
			           		return element.create(type, args, this, "after");
			           },
			           append   : function (type, args) {
			           		utility.genId(this);
			           		return element.create(type, args, this, "last");
			           },
			           attr     : function (key, value) {
			           		utility.genId(this);
			           		return element.attr(this, key, value);
			           },
			           before   : function (type, args) {
			           		utility.genId(this);
			           		return element.create(type, args, this, "before");
			           },
			           clear    : function () {
			           		utility.genId(this);
			           		return element.clear(this);
			           },
			           create   : function (type, args, position) {
			           		utility.genId(this);
			           		return element.create(type, args, this, position);
			           },
			           css       : function (key, value) {
			           		utility.genId(this);
			           		return element.css(this, key, value);
			           },
			           data      : function (key, value) {
			           		utility.genId(this);
			           		return element.data(this, key, value);
			           },
			           destroy   : function () { return element.destroy(this); },
			           disable   : function () { return element.disable(this); },
			           enable    : function () { return element.enable(this); },
			           find      : function (arg) {
			           		utility.genId(this);
			           		return element.find(this, arg);
			           },
			           fire     : function () {
			           		utility.genId(this);
			           		observer.fire.apply(observer, [this].concat(array.cast(arguments)));
			           		return this;
			           },
			           genId    : function () { return utility.genId(this); },
			           get      : function (uri, headers) {
			           		this.fire("beforeGet");
			           		var cached = cache.get(uri),
			           		    self   = this;

			           		!cached ? uri.get(function (arg) { self.html(arg).fire("afterGet"); }, function (arg) { self.fire("failedGet", {response: client.parse(arg), xhr: arg}); }, headers)
			           		        : this.html(cached.response).fire("afterGet");

			           		return this;
			           },
			           has      : function (arg) {
			           		utility.genId(this);
			           		return element.has(this, arg);
			           },
			           hasClass : function (arg) {
			           		utility.genId(this);
			           		return element.hasClass(this, arg);
			           },
			           hide     : function () {
			           		utility.genId(this);
			           		return element.hide(this);
			           },
			           html     : function (arg) {
			           		utility.genId(this);
			           		return typeof arg === "undefined" ? this.innerHTML : this.update({innerHTML: arg});
			           },
			           is       : function (arg) {
			           		utility.genId(this);
			           		return element.is(this, arg);
			           },
			           isAlphaNum: function () { return this.nodeName === "FORM" ? false : validate.test({alphanum: typeof this.value !== "undefined" ? this.value : element.text(this)}).pass; },
			           isBoolean: function () { return this.nodeName === "FORM" ? false : validate.test({"boolean": typeof this.value !== "undefined" ? this.value : element.text(this)}).pass; },
			           isChecked: function () { return this.nodeName !== "INPUT" ? false : this.attr("checked"); },
			           isDate   : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isDate()   : element.text(this).isDate(); },
			           isDisabled: function () { return this.nodeName !== "INPUT" ? false : this.attr("disabled"); },
			           isDomain : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isDomain() : element.text(this).isDomain(); },
			           isEmail  : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isEmail()  : element.text(this).isEmail(); },
			           isEmpty  : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isEmpty()  : element.text(this).isEmpty(); },
			           isHidden : function (arg) {
			           		utility.genId(this);
			           		return element.hidden(this);
			           },
			           isIP     : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isIP()     : element.text(this).isIP(); },
			           isInt    : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isInt()    : element.text(this).isInt(); },
			           isNumber : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isNumber() : element.text(this).isNumber(); },
			           isPhone  : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isPhone()  : element.text(this).isPhone(); },
			           isUrl    : function () { return this.nodeName === "FORM" ? false : typeof this.value !== "undefined" ? this.value.isUrl()    : element.text(this).isUrl(); },
			           jsonp    : function (uri, property, callback) {
			           		var target = this,
			           		    arg    = property, fn;

			           		fn = function (response) {
			           			var self = target,
			           			    node = response,
			           			    prop = arg,
			           			    i, nth, result;

			           			try {
			           				if (typeof prop !== "undefined") {
			           					prop = prop.replace(/]|'|"/g, "").replace(/\./g, "[").split("[");
			           					prop.each(function (i) {
			           						node = node[!!isNaN(i) ? i : parseInt(i)];
			           						if (typeof node === "undefined") throw Error(label.error.propertyNotFound);
			           					});
			           					result = node;
			           				}
			           				else result = response;
			           			}
			           			catch (e) {
			           				result = label.error.serverError;
			           				error(e, arguments, this);
			           			}

			           			self.text(result);
			           		};
			           		client.jsonp(uri, fn, function () { target.text(label.error.serverError); }, callback);
			           		return this;
			           },
			           listeners: function (event) {
			           		utility.genId(this);
			           		return $.listeners.call(this, event);
			           },
			           loading  : function () { return utility.loading(this); },
			           on       : function (event, listener, id, scope, state) {
			           		utility.genId(this);
			           		return $.on.call(this, event, listener, id, scope, state);
			           },
			           once     : function (event, listener, id, scope, state) {
			           		utility.genId(this);
			           		return $.once.call(this, event, listener, id, scope, state);
			           },
			           prepend  : function (type, args) {
			           		utility.genId(this);
			           		return element.create(type, args, this, "first");
			           },
			           prependChild: function (child) {
			           		utility.genId(this);
			           		return element.prependChild(this, child);
			           },
			           position : function () {
			           		utility.genId(this);
			           		return element.position(this);
			           },
			           removeClass : function (arg) {
			           		utility.genId(this);
			           		return element.klass(this, arg, false);
			           },
			           show     : function () {
			           		utility.genId(this);
			           		return element.show(this);
			           },
			           size     : function () {
			           		utility.genId(this);
			           		return element.size(this);
			           },
			           text     : function (arg) {
			           		utility.genId(this);
			           		return element.text(this, arg);
			           },
			           toggleClass: function (arg) {
			           		utility.genId(this);
			           		return element.toggleClass(this, arg);
			           },
			           tpl      : function (arg) { return utility.tpl(arg, this); },
			           un       : function (event, id, state) {
			           		utility.genId(this);
			           		return $.un.call(this, event, id, state);
			           },
			           update   : function (args) {
			           		utility.genId(this);
			           		return element.update(this, args);
			           },
			           val      : function (arg) {
			           		utility.genId(this);
			           		return element.val(this, arg);
			           },
			           validate : function () { return this.nodeName === "FORM" ? validate.test(this) : typeof this.value !== "undefined" ? !this.value.isEmpty() : !element.text(this).isEmpty(); }},
			"function": {reflect: function () { return utility.reflect(this); }},
			number  : {diff     : function (arg) { return number.diff (this, arg); },
			           fire     : function () { return observer.fire.apply(observer, [this.toString()].concat(array.cast(arguments))); return this; },
			           format   : function (delimiter, every) { return number.format(this, delimiter, every); },
			           half     : function (arg) { return number.half(this, arg); },
			           isEven   : function () { return number.even(this); },
			           isOdd    : function () { return number.odd(this); },
			           listeners: function (event) { return $.listeners.call(this.toString(), event); },
			           on       : function (event, listener, id, scope, state) { $.on.call(this.toString(), event, listener, id, scope || this, state); return this; },
			           once     : function (event, listener, id, scope, state) { $.once.call(this.toString(), event, listener, id, scope || this, state); return this; },
			           roundDown: function () { return number.round(this, "down"); },
			           roundUp  : function () { return number.round(this, "up"); },
			           un       : function (event, id, state) { $.un.call(this.toString(), event, id, state); return this; }},
			string  : {allows   : function (arg) { return client.allows(this, arg); },
			           capitalize: function () { return string.capitalize(this); },
			           del      : function (success, failure, headers) { return client.request(this, "DELETE", success, failure, null, headers); },
			           escape   : function () { return string.escape(this); },
			           expire   : function (silent) { return cache.expire(this, silent); },
			           explode  : function (arg) { return string.explode(this, arg); },
			           fire     : function () { return observer.fire.apply(observer, [this].concat(array.cast(arguments))); },
			           get      : function (success, failure, headers) { return client.request(this, "GET", success, failure, null, headers); },
			           headers  : function (success, failure) { return client.request(this, "HEAD", success, failure); },
			           hyphenate: function (camel) { return string.hyphenate(this, camel); },
			           isAlphaNum: function () { return validate.test({alphanum: this}).pass; },
			           isBoolean: function () { return validate.test({"boolean": this}).pass; },
			           isDate   : function () { return validate.test({date: this}).pass; },
			           isDomain : function () { return validate.test({domain: this}).pass; },
			           isEmail  : function () { return validate.test({email: this}).pass; },
			           isEmpty  : function () { return (string.trim(this) === ""); },
			           isIP     : function () { return validate.test({ip: this}).pass; },
			           isInt    : function () { return validate.test({integer: this}).pass; },
			           isNumber : function () { return validate.test({number: this}).pass; },
			           isPhone  : function () { return validate.test({phone: this}).pass; },
			           isUrl    : function () { return validate.test({url: this}).pass; },
			           jsonp    : function (success, failure, callback) { return client.jsonp(this, success, failure, callback); },
			           listeners: function (event) { return $.listeners.call(this, event); },
			           post     : function (success, failure, args, headers) { return client.request(this, "POST", success, failure, args, headers); },
			           put      : function (success, failure, args, headers) { return client.request(this, "PUT", success, failure, args, headers); },
			           on       : function (event, listener, id, scope, state) { return $.on.call(this, event, listener, id, scope, state); },
			           once     : function (event, listener, id, scope, state) { return $.once.call(this, event, listener, id, scope, state); },
			           options  : function (success, failure) { return client.request(this, "OPTIONS", success, failure); },
			           permissions: function () { return client.permissions(this); },
			           singular : function () { return string.singular(this); },
			           toCamelCase: function () { return string.toCamelCase(this); },
			           toNumber : function (base) { return number.parse(this, base); },
			           trim     : function () { return string.trim(this); },
			           un       : function (event, id, state) { return $.un.call(this, event, id, state); },
			           uncapitalize: function () { return string.uncapitalize(this); }}
		};

		utility.iterate(methods[type], function (v, k) { utility.property(obj.prototype, k, {value: v}); });
		return obj;
	},

	/**
	 * Returns an Object containing 1 or all key:value pairs from the querystring
	 *
	 * @method queryString
	 * @param  {String} arg [Optional] Key to find in the querystring
	 * @return {Object}     Object of 1 or all key:value pairs in the querystring
	 */
	queryString : function (arg) {
		var obj    = {},
		    result = location.search.isEmpty() ? null : location.search.replace("?", ""),
		    item;

		arg = arg || ".*";

		if (result !== null) {
			result = result.split("&");
			result.each(function (prop) {
				item = prop.split("=");

				if (item[0].isEmpty()) return;

				switch (true) {
					case typeof item[1] === "undefined":
					case item[1].isEmpty():
						item[1] = "";
						break;
					case item[1].isNumber():
						item[1] = Number(item[1]);
						break;
					case item[1].isBoolean():
						item[1] = (item[1] === "true");
						break;
				}

				switch (true) {
					case typeof obj[item[0]] === "undefined":
						obj[item[0]] = item[1];
						break;
					case !(obj[item[0]] instanceof Array):
						obj[item[0]] = [obj[item[0]]];
					default:
						obj[item[0]].push(item[1]);
				}
			});
		}
		return obj;
	},

	/**
	 * Returns an Array of parameters of a Function
	 * 
	 * @method reflect
	 * @param  {Function} arg Function to reflect
	 * @return {Array}        Array of parameters
	 */
	reflect : function (arg) {
		if (typeof arg === "undefined") arg = this;
		if (typeof arg === "undefined") arg = $;
		arg = arg.toString().match(/function\s+\w*\s*\((.*?)\)/)[1];
		return arg !== "" ? arg.explode() : [];
	},

	/**
	 * Creates a recursive function
	 * 
	 * Return false from the function to halt recursion
	 * 
	 * @method repeat
	 * @param  {Function} fn  Function to execute repeatedly
	 * @param  {Number}   ms  Milliseconds to stagger the execution
	 * @param  {String}   id  [Optional] Timeout ID
	 * @return {String}       Timeout ID
	 */
	repeat : function (fn, ms, id) {
		ms = ms || 10;
		id = id || utility.guid(true);

		utility.defer(function () {
			var recursive = function (fn, ms, id) {
				var recursive = this;

				if (fn() !== false) {
					$.repeating[id] = setTimeout(function () {
						recursive.call(recursive, fn, ms, id);
					}, ms);
				}
				else delete $.repeating[id];
			};

			recursive.call(recursive, fn, ms, id);
		}, ms, id);

		return id;
	},

	/**
	 * Creates a script Element to load an external script
	 * 
	 * @method script
	 * @param  {String} arg    URL to script
	 * @param  {Object} target [Optional] Element to receive the script
	 * @param  {String} pos    [Optional] Position to create the script at within the target
	 * @return {Object}        Script
	 */
	script : function (arg, target, pos) {
		return element.create("script", {type: "application/javascript", src: arg}, target || $("head")[0], pos);
	},

	/**
	 * Creates a link Element to load an external stylesheet
	 * 
	 * @method stylesheet
	 * @param  {String} arg   URL to stylesheet
	 * @param  {String} media [Optional] Medias the stylesheet applies to
	 * @return {Objecct}      Stylesheet
	 */
	stylesheet : function (arg, media) {
		return element.create("link", {rel: "stylesheet", type: "text/css", href: arg, media: media || "print, screen"}, $("head")[0]);
	},

	/**
	 * Stops an Event from bubbling
	 * 
	 * @param  {Object} e Event
	 * @return {Object}   Event
	 */
	stop : function (e) {
		if (typeof e.cancelBubble    !== "undefined") e.cancelBubble = true;
		if (typeof e.preventDefault  === "function")  e.preventDefault();
		if (typeof e.stopPropagation === "function")  e.stopPropagation();
		return e;
	},

	/**
	 * Transforms JSON to HTML and appends to Body or target Element
	 *
	 * @method tpl
	 * @param  {Object} data   JSON Object describing HTML
	 * @param  {Mixed}  target [Optional] Target Element or Element.id to receive the HTML
	 * @return {Object}        New Element created from the template
	 */
	tpl : function (arg, target) {
		var frag;

		switch (true) {
			case typeof arg !== "object":
			case !(/object|undefined/.test(typeof target)) && typeof (target = target.charAt(0) === "#" ? $(target) : $(target)[0]) === "undefined":
				throw Error(label.error.invalidArguments);
		}

		if (typeof target === "undefined") target = $("body")[0];

		frag  = document.createDocumentFragment();
		switch (true) {
			case arg instanceof Array:
				arg.each(function (i, idx) {
					element.create(array.cast(i, true)[0], frag).html(array.cast(i)[0]);
				});
				break;
			default:
				utility.iterate(arg, function (i, k) {
					switch (true) {
						case typeof i === "string":
							element.create(k, frag).html(i);
							break;
						case i instanceof Array:
						case i instanceof Object:
							utility.tpl(i, element.create(k, frag));
							break;
					}
				});
		}
		target.appendChild(frag);
		return array.last(target.childNodes);
	},

	/**
	 * Walks a structure and returns arg
	 * 
	 * @method  walk
	 * @param  {Mixed}  obj  Object or Array
	 * @param  {String} arg  String describing the property to return
	 * @return {Mixed}       arg
	 */
	walk : function (obj, arg) {
		arg.replace(/\]$/, "").replace(/\]/g, ".").split(/\.|\[/).each(function (i) {
			obj = obj[i];
		});
		return obj;
	}
};

/**
 * Validation methods and patterns
 *
 * pattern.url is authored by Diego Perini
 *
 * @class validate
 * @namespace abaaso
 */
var validate = {
	// Regular expression patterns to test against
	pattern : {
		alphanum : /^[a-zA-Z0-9]+$/,
		"boolean": /^(0|1|true|false)?$/,
		domain   : /^[\w\-_]+(\.[\w\-_]+)+([\w\-\.,@?^=%&amp;:\/~\+#]*[\w\-\@?^=%&amp;\/~\+#])?/,
		email    : /^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$/,
		ip       : /^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/,
		integer  : /(^-?\d\d*$)/,
		notEmpty : /\w{1,}/,
		number   : /(^-?\d\d*\.\d*$)|(^-?\d\d*$)|(^-?\.\d\d*$)/,
		phone    : /^([0-9\(\)\/\+ \-\.]+)$/,
		url      : /^(?:(?:https?|ftp):\/\/)(?:\S+(?::\S*)?@)?(?:(?!10(?:\.\d{1,3}){3})(?!127(?:\.\d{1,3}){3})(?!169\.254(?:\.\d{1,3}){2})(?!192\.168(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\u00a1-\uffff0-9]+-?)*[a-z\u00a1-\uffff0-9]+)(?:\.(?:[a-z\u00a1-\uffff0-9]+-?)*[a-z\u00a1-\uffff0-9]+)*(?:\.(?:[a-z\u00a1-\uffff]{2,})))(?::\d{2,5})?(?:\/[^\s]*)?$/i
	},

	/**
	 * Validates args based on the type or pattern specified
	 *
	 * @method test
	 * @param  {Object} args Object to test {(pattern[name] || /pattern/) : (value || #object.id)}
	 * @return {Object}      Results
	 */
	test : function (args) {
		var exception = false,
		    invalid   = [],
		    tracked   = {},
		    value     = null,
		    result    = [],
		    c         = [],
		    inputs    = [],
		    selects   = [],
		    i, p, o, x, nth;

		if (typeof args.nodeName !== "undefined" && args.nodeName === "FORM") {
			if (args.id.isEmpty()) args.genId();
			inputs  = $("#" + args.id + " input");
			selects = $("#" + args.id + " select");
			if (inputs.length > 0)  c = c.concat(inputs);
			if (selects.length > 0) c = c.concat(selects);
			c.each(function (i) {
				var z = {},
				    p, v, r;

				p = validate.pattern[i.nodeName.toLowerCase()] ? validate.pattern[i.nodeName.toLowerCase()] : ((!i.id.isEmpty() && validate.pattern[i.id.toLowerCase()]) ? validate.pattern[i.id.toLowerCase()] : "notEmpty");
				v = i.val();
				if (v === null) v = "";
				z[p] = v;
				r    = validate.test(z)
				if (!r.pass) {
					invalid.push({element: i, test: p, value: v});
					exception = true;
				}
			});
		}
		else {
			utility.iterate(args, function (i, k) {
				if (typeof k === "undefined" || typeof i === "undefined") {
					invalid.push({test: k, value: i});
					exception = true;
					return
				}
				value = String(i).charAt(0) === "#" ? (typeof $(i) !== "undefined" ? $(i).val() : "") : i;
				switch (k) {
					case "date":
						if (isNaN(new Date(value).getYear())) {
							invalid.push({test: k, value: value});
							exception = true;
						}
						break;
					case "domain":
						if (!validate.pattern.domain.test(value.replace(/.*\/\//, ""))) {
							invalid.push({test: k, value: value});
							exception = true;
						}
						break;
					case "domainip":
						if (!validate.pattern.domain.test(value.replace(/.*\/\//, "")) || !validate.pattern.ip.test(value)) {
							invalid.push({test: k, value: value});
							exception = true;
						}
						break;
					default:
						p = typeof validate.pattern[k] !== "undefined" ? validate.pattern[k] : k;
						if (!p.test(value)) {
							invalid.push({test: k, value: value});
							exception = true;
						}
				}
			});
		}
		return {pass: !exception, invalid: invalid};
	}
};

/**
 * XMLHttpRequest shim for node.js
 * 
 * @return {Object} Instance of xhr
 */
var xhr = function () {
	var UNSENT           = 0,
	    OPENED           = 1,
	    HEADERS_RECEIVED = 2,
	    LOADING          = 3,
	    DONE             = 4,
	    regex            = RegExp(HEADERS_RECEIVED + "|" + LOADING),
	    XMLHttpRequest, headers, handler, handlerError, state;

	headers = {
		"User-Agent"   : "abaaso/3.2.2 node.js/" + process.versions.node.replace(/^v/, "") + " (" + string.capitalize(process.platform) + " V8/" + process.versions.v8 + ")",
		"Content-Type" : "text/plain",
		"Accept"       : "*/*"
	};

	/**
	 * Changes the readyState of an XMLHttpRequest
	 * 
	 * @param  {String} arg  New readyState
	 * @return {Object}      XMLHttpRequest
	 */
	state = function (arg) {
		if (this.readyState !== arg) {
			this.readyState = arg;
			this.dispatchEvent("readystatechange");
			if (this.readyState === DONE && !this._error) {
				this.dispatchEvent("load");
				this.dispatchEvent("loadend");
			}
		}
		return this;
	};

	handler = function (res) {
		var self = this;

		state.call(this, HEADERS_RECEIVED);

		this.status      = res.statusCode;
		this._resheaders = res.headers;

		if (typeof this._resheaders["set-cookie"] !== "undefined" && this._resheaders["set-cookie"] instanceof Array) this._resheaders["set-cookie"] = this._resheaders["set-cookie"].join(";");

		res.on("data", function (arg) {
			res.setEncoding("utf8");
			if (self._send) {
				if (arg) self.responseText += arg;
				state.call(self, LOADING);
			}
		});

		res.on("end", function () {
			if (self._send) {
				state.call(self, DONE);
				self._send = false;
			}
		});

		res.on("close", function (e) {
			handlerError.call(self, e);
		});
	};

	handlerError = function (e) {
		this.status       = 503;
		this.statusText   = e;
		this.responseText = e.stack || e;
		this._error       = true;
		this.dispatchEvent("error");
		state.call(this, DONE);
	};

	XMLHttpRequest = function () {
		this.onabort            = null;
		this.onerror            = null;
		this.onload             = null;
		this.onloadend          = null;
		this.onloadstart        = null;
		this.onreadystatechange = null;
		this.readyState         = UNSENT;
		this.response           = null;
		this.responseText       = "";
		this.responseType       = "";
		this.responseXML        = null;
		this.status             = UNSENT;
		this.statusText         = "";

		// Psuedo private for prototype chain
		this._id                = utility.genId();
		this._error             = false;
		this._headers           = {};
		this._listeners         = {};
		this._params            = {};
		this._request           = null;
		this._resheaders        = {};
		this._send              = false;
	};

	/**
	 * Aborts a request
	 * 
	 * @return {Object} XMLHttpRequest
	 */
	XMLHttpRequest.prototype.abort = function () {
		if (this._request !== null) {
			this._request.abort();
			this._request = null;
		}

		this.responseText = "";
		this.responseXML  = "";
		this._error       = true;
		this._headers     = {};

		if (this._send === true || regex.test(this.readyState)) {
			this._send = false;
			state.call(this, DONE)
		}

		this.dispatchEvent("abort");
		this.readyState = UNSENT;
		return this;
	};

	/**
	 * Adds an event listener to an XMLHttpRequest instance
	 * 
	 * @param {String}   event Event to listen for
	 * @param {Function} fn    Event handler
	 * @return {Object}        XMLHttpRequest
	 */
	XMLHttpRequest.prototype.addEventListener = function (event, fn) {
		if (!this._listeners.hasOwnProperty(event)) this._listeners[event] = [];
		this._listeners[event].add(fn);
		return this;
	};

	/**
	 * Dispatches an event
	 * 
	 * @param  {String} event Name of event
	 * @return {Object}       XMLHttpRequest
	 */
	XMLHttpRequest.prototype.dispatchEvent = function (event) {
		var self = this;

		if (typeof this["on" + event] === "function") this["on" + event]();
		if (this._listeners.hasOwnProperty(event)) this._listeners[event].each(function (i) {
			if (typeof i === "function") i.call(self);
		});

		return this;
	};

	/**
	 * Gets all response headers
	 * 
	 * @return {Object} Response headers
	 */
	XMLHttpRequest.prototype.getAllResponseHeaders = function () {
		var result = "";

		if (this.readyState < HEADERS_RECEIVED || this._error) throw Error("INVALID_STATE_ERR: Headers have not been received");
		utility.iterate(this._resheaders, function (v, k) { result += k + ": " + v + "\n"; });
		return result;
	};

	/**
	 * Gets a specific response header
	 * 
	 * @param  {String} header Header to get
	 * @return {String}        Response header value
	 */
	XMLHttpRequest.prototype.getResponseHeader = function (header) {
		var result;

		if (this.readyState < HEADERS_RECEIVED || this._error) throw Error("INVALID_STATE_ERR: Headers have not been received");
		result = this._resheaders[header] || this._resheaders[header.toLowerCase()];
		return result;
	};

	/**
	 * Prepares an XMLHttpRequest instance to make a request
	 * 
	 * @param  {String}  method   HTTP method
	 * @param  {String}  url      URL to receive request
	 * @param  {Boolean} async    [Optional] Asynchronous request
	 * @param  {String}  user     [Optional] Basic auth username
	 * @param  {String}  password [Optional] Basic auth password
	 * @return {Object}           XMLHttpRequest
	 */
	XMLHttpRequest.prototype.open = function (method, url, async, user, password) {
		var self = this;

		if (typeof async !== "undefined" && async !== true) throw Error("Synchronous XMLHttpRequest requests are not supported");

		this.abort();
		this._error  = false;
		this._params = {
			method   : method,
			url      : url,
			async    : async    || true,
			user     : user     || null,
			password : password || null
		}

		utility.iterate(headers, function (v, k) { self._headers[k] = v; });
		this.readyState = OPENED;
		return this;
	};

	/**
	 * Overrides the Content-Type of the request
	 * 
	 * @param  {String} mime Mime type of the request (media type)
	 * @return {Object}      XMLHttpRequest
	 */
	XMLHttpRequest.prototype.overrideMimeType = function (mime) {
		this._headers["Content-Type"] = mime;
		return this;
	};

	/**
	 * Removes an event listener from an XMLHttpRequest instance
	 * 
	 * @param {String}   event Event to listen for
	 * @param {Function} fn    Event handler
	 * @return {Object}        XMLHttpRequest
	 */
	XMLHttpRequest.prototype.removeEventListener = function (event, fn) {
		if (!this._listeners.hasOwnProperty(event)) return;
		this._listeners[event].remove(fn);
		return this;
	};

	/**
	 * Sends an XMLHttpRequest request
	 * 
	 * @param  {Mixed} data [Optional] Payload to send with the request
	 * @return {Object}     XMLHttpRequest
	 * @todo  finish this method
	 */
	XMLHttpRequest.prototype.send = function (data) {
		data     = data || null;
		var self = this,
		    options, parsed, request, obj;

		switch (true) {
			case this.readyState < OPENED:
				throw Error("INVALID_STATE_ERR: Object is not open");
			case this._send:
				throw Error("INVALID_STATE_ERR: Object is sending");
		}

		parsed      = url.parse(this._params.url);
		parsed.port = parsed.port || (parsed.protocol === "https" ? 443 : 80);
		if (this._params.user !== null && this._params.password !== null) parsed.auth = this._params.user + ":" + this._params.password;

		// Specifying Content-Length accordingly
		if (/post|put/i.test(this._params.method)) this._headers["Content-Length"] = data !== null ? Buffer.byteLength(data) : 0;

		this._headers["Host"] = parsed.hostname + (!/80|443/.test(parsed.port) ? ":" + parsed.port : "");

		options = {
			hostname : parsed.hostname,
			path     : parsed.path,
			port     : parsed.port,
			method   : this._params.method,
			headers  : this._headers
		}

		if (typeof parsed.auth !== "undefined") options.auth = parsed.auth;

		self._send = true;
		self.dispatchEvent("readystatechange");

		obj           = parsed.protocol === "http:" ? http : https;
		request       = obj.request(options, function (arg) { handler.call(self, arg); }).on("error", function (e) { handlerError.call(self, e); });
		data === null ? request.setSocketKeepAlive(true, 10000) : request.write(data, "utf8");
		this._request = request;
		request.end();

		self.dispatchEvent("loadstart");

		return this;
	};

	/**
	 * Sets a request header of an XMLHttpRequest instance
	 * 
	 * @param {String} header HTTP header
	 * @param {String} value  Header value
	 * @return {Object}       XMLHttpRequest
	 */
	XMLHttpRequest.prototype.setRequestHeader = function (header, value) {
		switch (true) {
			case this.readyState !== OPENED:
				throw Error("INVALID_STATE_ERR: Object is not usable");
			case this._send:
				throw Error("INVALID_STATE_ERR: Object is sending");
		}
		this._headers[header] = value;
		return this;
	};

	return XMLHttpRequest;
};

/**
 * XML methods
 *
 * @class xml
 * @namespace abaaso
 */
var xml = {
	/**
	 * Returns XML (Document) Object from a String
	 *
	 * @method decode
	 * @param  {String} arg XML String
	 * @return {Object}     XML Object or undefined
	 */
	decode : function (arg) {
		try {
			var x;

			if (typeof arg !== "string" || arg.isEmpty()) throw Error(label.error.invalidArguments);

			if (client.ie) {
				x = new ActiveXObject("Microsoft.XMLDOM");
				x.async = "false";
				x.loadXML(arg);
			}
			else x = new DOMParser().parseFromString(arg, "text/xml");
			return x;
		}
		catch (e) {
			error(e, arguments, this);
			return undefined;
		}
	},

	/**
	 * Returns XML String from an Object or Array
	 *
	 * @method encode
	 * @param  {Mixed} arg Object or Array to cast to XML String
	 * @return {String}    XML String or undefined
	 */
	encode : function (arg, wrap) {
		try {
			if (typeof arg === "undefined") throw Error(label.error.invalidArguments);

			wrap    = !(wrap === false);
			var x   = wrap ? "<xml>" : "",
			    top = !(arguments[2] === false),
			    node, i;

			if (arg !== null && typeof arg.xml !== "undefined") arg = arg.xml;
			if (arg instanceof Document) arg = (new XMLSerializer()).serializeToString(arg);

			node = function (name, value) {
				var output = "<n>v</n>";
				if (/\&|\<|\>|\"|\'|\t|\r|\n|\@|\$/g.test(value)) output = output.replace(/v/, "<![CDATA[v]]>");
				return output.replace(/n/g, name).replace(/v/, value);
			}

			switch (true) {
				case typeof arg === "boolean":
				case typeof arg === "number":
				case typeof arg === "string":
					x += node("item", arg);
					break;
				case typeof arg === "object":
					utility.iterate(arg, function (v, k) { x += xml.encode(v, (typeof v === "object"), false).replace(/item|xml/g, isNaN(k) ? k : "item"); });
					break;
			}

			x += wrap ? "</xml>" : "";
			if (top) x = "<?xml version=\"1.0\" encoding=\"UTF8\"?>" + x;

			return x;
		}
		catch (e) {
			error(e, arguments, this);
			return undefined;
		}
	}
};

// concated before outro.js
error     = utility.error;
bootstrap = function () {
	var cleanup, fn;

	if (typeof abaaso.bootstrap === "function") delete abaaso.bootstrap;

	// Describing the Client
	abaaso.client.size = client.size();
	client.version();
	client.mobile();
	client.tablet();

	// IE7 and older is not supported
	if (client.ie && client.version < 8) return;
	
	cleanup = function (obj) {
		var nodes = [];

		observer.remove(obj);
		if (obj.childNodes.length > 0) nodes = array.cast(obj.childNodes);
		if (nodes.length > 0) nodes.each(function (i) { cleanup(i); });
	};

	fn = function (e) {
		if (/complete|loaded/.test(document.readyState)) {
			if (typeof abaaso.init === "function") abaaso.init();
			return false;
		}
	};

	if (typeof Array.prototype.filter === "undefined") {
		Array.prototype.filter = function (fn) {
			if (this === void 0 || this === null || typeof fn !== "function") throw Error(label.error.invalidArguments);

			var i      = null,
			    t      = Object(this),
			    nth    = t.length >>> 0,
			    result = [],
			    prop   = arguments[1],
			    val    = null;

			for (i = 0; i < nth; i++) {
				if (i in t) {
					val = t[i];
					if (fn.call(prop, val, i, t)) result.push(val);
				}
			}

			return result;
		};
	}

	if (typeof Array.prototype.forEach === "undefined") {
		Array.prototype.forEach = function (callback, thisArg) {
			if (this === null || typeof callback !== "function") throw Error(label.error.invalidArguments);

			var T,
			    k   = 0,
			    O   = Object(this),
			    len = O.length >>> 0;

			if (thisArg) T = thisArg;

			while (k < len) {
				var kValue;
				if (k in O) {
					kValue = O[k];
					callback.call(T, kValue, k, O);
				}
				k++;
			}
		};
	}

	if (typeof Array.prototype.indexOf === "undefined") {
		Array.prototype.indexOf = function(obj, start) {
			for (var i = (start || 0), j = this.length; i < j; i++) {
				if (this[i] === obj) return i;
			}

			return -1;
		}
	}

	if (typeof Array.prototype.map === "undefined") {
		Array.prototype.map = function (callback, thisArg) {
			var T, A, k;

			if (this == null) throw new TypeError(" this is null or not defined");

			var O = Object(this);
			var len = O.length >>> 0;

			if ({}.toString.call(callback) != "[object Function]") throw new TypeError(callback + " is not a function");
			if (thisArg) T = thisArg;

			A = new Array(len);
			k = 0;

			while(k < len) {
				var kValue, mappedValue;
				if (k in O) {
					kValue = O[k];
					mappedValue = callback.call(T, kValue, k, O);
					A[k] = mappedValue;
				}
				k++;
			}

			return A;
		}
	}

	if (typeof Array.prototype.reduce === "undefined") {
		Array.prototype.reduce = function (accumulator) {
			if (this === null || this === undefined) throw new TypeError("Object is null or undefined");

			var i = 0, l = this.length >> 0, curr;

			if (typeof accumulator !== "function") throw new TypeError("First argument is not callable");

			if (arguments.length < 2) {
				if (l === 0) throw new TypeError("Array length is 0 and no second argument");
				curr = this[0];
				i = 1; // start accumulating at the second element
			}
			else curr = arguments[1];

			while (i < l) {
				if (i in this) curr = accumulator.call(undefined, curr, this[i], i, this);
				++i;
			}

			return curr;
		};
	}

	if (!server && typeof document.documentElement.classList === "undefined") {
		(function (view) {
			var ClassList, getter, proto, target, descriptor;

			if (!("HTMLElement" in view) && !("Element" in view)) return;

			ClassList = function (obj) {
				var classes = !obj.className.isEmpty() ? obj.className.explode(" ") : [],
				    self    = this;

				classes.each(function (i) { self.push(i); });
				this.updateClassName = function () { obj.className = this.join(" "); };
			};

			getter = function () {
				return new ClassList(this);
			};

			proto  = ClassList["prototype"] = [];
			target = (view.HTMLElement || view.Element)["prototype"];

			proto.add = function (arg) {
				if (!array.contains(this, arg)) {
					this.push(arg);
					this.updateClassName();
				}
			};

			proto.contains = function (arg) {
				return array.contains(this, arg);
			};

			proto.remove = function (arg) {
				if (array.contains(this, arg)) {
					array.remove(this, arg);
					this.updateClassName();
				}
			};

			proto.toggle = function (arg) {
				array[array.contains(this, arg) ? "remove" : "add"](this, arg);
				this.updateClassName();
			};

			if (Object.defineProperty) {
				descriptor = {
					get          : getter,
					enumerable   : !client.ie || client.version > 8 ? true : false,
					configurable : true
				};

				Object.defineProperty(target, "classList", descriptor);
			}
			else if (Object.prototype.__defineGetter__) target.__defineGetter__("classList", getter);
			else throw Error("Could not create classList shim");
		})(global);
	}

	if (typeof Function.prototype.bind === "undefined") {
		Function.prototype.bind = function (arg) {
			var fn    = this,
			    slice = Array.prototype.slice,
			    args  = slice.call(arguments, 1);
			
			return function () {
				return fn.apply(arg, args.concat(slice.call(arguments)));
			};
		};
	}

	// Cookie class is not relevant for server environment
	if (server) {
		delete abaaso.cookie;
		XMLHttpRequest = xhr();
	}

	// Binding helper & namespace to $
	$ = abaaso.$.bind($);
	utility.alias($, abaaso);
	delete $.$;
	delete $.bootstrap;
	delete $.callback;
	delete $.init;
	delete $.loading;

	// Creating route.initial after alias() so it's not assumed
	abaaso.route.initial = null;

	// Short cut to loading.create
	$.loading = abaaso.loading.create.bind($.loading);

	// Unbinding observer methods to maintain scope
	$.fire      = abaaso.fire;
	$.on        = abaaso.on;
	$.once      = abaaso.once;
	$.un        = abaaso.un;
	$.listeners = abaaso.listeners;

	// Setting initial application state
	abaaso.state._current = abaaso.state.current = "initial";
	$.state._current      = $.state.current      = abaaso.state.current;

	// Setting sugar
	if (!server) switch (true) {
		case typeof global.$ === "undefined" || global.$ === null:
			global.$ = $;
			break;
		default:
			global.a$ = $;
			abaaso.aliased = "a$";
	}

	// Hooking abaaso into native Objects
	utility.proto(Array, "array");
	if (typeof Element      !== "undefined") utility.proto(Element,      "element");
	if (typeof HTMLDocument !== "undefined") utility.proto(HTMLDocument, "element");
	utility.proto(Function, "function");
	utility.proto(Number, "number");
	utility.proto(String, "string");

	// Creating a singleton
	abaaso.constructor = abaaso;

	// Creating error log
	$.error.log = abaaso.error.log = [];

	// Setting events & garbage collection
	$.on(global, "error", function (e) { $.fire("error", e); }, "error", global, "all");
	if (!server) {
		$.on(global, "hashchange", function ()  { $.fire("beforeHash, hash, afterHash", location.hash); }, "hash", global, "all");
		$.on(global, "resize",     function ()  { $.client.size = abaaso.client.size = client.size(); $.fire("resize", abaaso.client.size); }, "resize", global, "all");
		$.on(global, "load",       function ()  { $.fire("render").un("render"); });
		$.on(global, "DOMNodeInserted", function (e) {
			var obj = e.target;
			if (typeof obj.id !== "undefined" && obj.id.isEmpty()) {
				utility.genId(obj);
				if (obj.parentNode instanceof Element) obj.parentNode.fire("afterCreate", obj);
				$.fire("afterCreate", obj);
			}
		}, "mutation", global, "all");
		$.on(global, "DOMNodeRemoved", function (e) { cleanup(e.target); }, "mutation", global, "all");

		// Routing listener
		$.on("hash", function (arg) { if ($.route.enabled || abaaso.route.enabled) route.load(arg); }, "route", abaaso.route, "all");
	}

	// abaaso.state.current getter/setter
	var getter, setter;
	getter = function () { return this._current; };
	setter = function (arg) {
		if (arg === null || typeof arg !== "string" || this.current === arg || arg.isEmpty()) throw Error(label.error.invalidArguments);

		abaaso.state.previous = abaaso.state._current;
		abaaso.state._current = arg;
		return abaaso.fire(arg);
	};

	switch (true) {
		case (!client.ie || client.version > 8):
			utility.property(abaaso.state, "current", {enumerable: true, get: getter, set: setter});
			utility.property($.state,      "current", {enumerable: true, get: getter, set: setter});
			break;
		default: // Pure hackery, only exists when needed
			abaaso.state.change = function (arg) { setter.call(abaaso.state, arg); return abaaso.state.current = arg; };
			$.state.change      = function (arg) { setter.call(abaaso.state, arg); return abaaso.state.current = arg; };
	}

	$.ready = true;

	// Preparing init()
	switch (true) {
		case server:
			abaaso.init();
			break;
		case typeof global.define === "function":
			global.define("abaaso", function () { return abaaso.init(); });
			break;
		case (/complete|loaded/.test(document.readyState)):
			abaaso.init();
			break;
		case typeof document.addEventListener === "function":
			document.addEventListener("DOMContentLoaded", abaaso.init, false);
			break;
		case typeof document.attachEvent === "function":
			document.attachEvent("onreadystatechange", fn);
			break;
		default:
			utility.repeat(fn);
	}
};

return {
	// Classes
	array           : array,
	callback        : {},
	client          : {
		// Properties
		android : client.android,
		blackberry : client.blackberry,
		chrome  : client.chrome,
		firefox : client.firefox,
		ie      : client.ie,
		ios     : client.ios,
		linux   : client.linux,
		mobile  : client.mobile,
		opera   : client.opera,
		osx     : client.osx,
		playbook: client.playbook,
		safari  : client.safari,
		tablet  : client.tablet,
		size    : {height: 0, width: 0},
		version : 0,
		webos   : client.webos,
		windows : client.windows,

		// Methods
		del     : function (uri, success, failure, headers) { return client.request(uri, "DELETE", success, failure, null, headers); },
		get     : function (uri, success, failure, headers) { return client.request(uri, "GET", success, failure, null, headers); },
		headers : function (uri, success, failure) { return client.request(uri, "HEAD", success, failure); },
		post    : function (uri, success, failure, args, headers) { return client.request(uri, "POST", success, failure, args, headers); },
		put     : function (uri, success, failure, args, headers) { return client.request(uri, "PUT", success, failure, args, headers); },
		jsonp   : function (uri, success, failure, callback) { return client.jsonp(uri, success, failure, callback); },
		options : function (uri, success, failure) { return client.request(uri, "OPTIONS", success, failure); },
		permissions : client.permissions
	},
	cookie          : cookie,
	element         : element,
	json            : json,
	label           : label,
	loading         : {
		create  : utility.loading,
		url     : null
	},
	message         : message,
	mouse           : mouse,
	number          : number,
	observer        : {
		log     : observer.log,
		add     : observer.add,
		fire    : observer.fire,
		hook    : observer.hook,
		list    : observer.list,
		once    : observer.once,
		remove  : observer.remove
	},
	repeating       : {},
	route           : {
		enabled : false,
		del     : route.del,
		hash    : route.hash,
		init    : route.init,
		list    : route.list,
		load    : route.load,
		server  : route.server,
		set     : route.set
	},
	state           : {
		_current    : null,
		header      : null,
		previous    : null
	},
	string          : string,
	validate        : validate,
	xml             : xml,

	// Methods & Properties
	$               : utility.$,
	alias           : utility.alias,
	aliased         : "$",
	allows          : client.allows,
	append          : function (type, args, obj) {
		if (obj instanceof Element) obj.genId();
		return element.create(type, args, obj, "last");
	},
	bootstrap       : bootstrap,
	clear           : element.clear,
	clone           : utility.clone,
	coerce          : utility.coerce,
	compile         : utility.compile,
	create          : element.create,
	css             : utility.css,
	data            : data.factory,
	datalist        : datalist.factory,
	decode          : json.decode,
	defer           : utility.defer,
	define          : utility.define,
	del             : function (uri, success, failure, headers) { return client.request(uri, "DELETE", success, failure, null, headers); },
	destroy         : element.destroy,
	encode          : json.encode,
	error           : utility.error,
	expire          : cache.clean,
	expires         : 120000,
	extend          : utility.extend,
	filter          : filter.factory,
	fire            : function (arg) {
		var local = (typeof arg === "string"),
		    args  = array.cast(arguments),
		    obj   = this;

		if (local) {
			if (obj === $) obj = abaaso;
			args = [obj].concat(args);
		}

		observer.fire.apply(observer, args);
		return this;
	},
	genId           : utility.genId,
	get             : function (uri, success, failure, headers) { return client.request(uri, "GET", success, failure, null, headers); },
	guid            : utility.guid,
	headers         : function (uri, success, failure) { return client.request(uri, "HEAD", success, failure); },
	hidden          : element.hidden,
	id              : "abaaso",
	init            : function () {
		// Stopping multiple executions
		delete abaaso.init;

		// Firing events to setup
		return $.fire("init").un("init").fire("ready").un("ready");
	},
	iterate         : utility.iterate,
	jsonp           : function (uri, success, failure, callback) { return client.jsonp(uri, success, failure, callback); },
	listeners       : function (event) {
		var obj = this;

		if (typeof obj === "undefined" || obj === $) obj = abaaso;
		return observer.list(obj, event);
	},
	log             : utility.log,
	merge           : utility.merge,
	module          : utility.module,
	object          : utility.object,
	on              : function (obj, event, listener, id, scope, state) {
		var all = typeof listener === "function",
		    o, e, l, i, s, st;

		o  = all ? obj      : this;
		e  = all ? event    : obj;
		l  = all ? listener : event;
		i  = all ? id       : listener;
		s  = all ? scope    : id;
		st = all ? state    : scope;

		if (typeof o === "undefined" || o === $) o = abaaso;
		if (typeof s === "undefined") s = o;
		return observer.add(o, e, l, i, s, st);
	},
	once            : function (obj, event, listener, id, scope, state) {
		var all = typeof listener === "function",
		    o, e, l, i, s, st;

		o  = all ? obj      : this;
		e  = all ? event    : obj;
		l  = all ? listener : event;
		i  = all ? id       : listener;
		s  = all ? scope    : id;
		st = all ? state    : scope;

		if (typeof o === "undefined" || o === $) o = abaaso;
		if (typeof s === "undefined") s = o;
		return observer.once(o, e, l, i, s, st);
	},
	options         : function (uri, success, failure) { return client.request(uri, "OPTIONS", success, failure); },
	parse           : utility.parse,
	permissions     : client.permissions,
	position        : element.position,
	post            : function (uri, success, failure, args, headers) { return client.request(uri, "POST", success, failure, args, headers); },
	prepend         : function (type, args, obj) {
		if (obj instanceof Element) obj.genId();
		return element.create(type, args, obj, "first");
	},
	promise         : promise.factory,
	property        : utility.property,
	put             : function (uri, success, failure, args, headers) { return client.request(uri, "PUT", success, failure, args, headers); },
	queryString     : utility.queryString,
	ready           : false,
	reflect         : utility.reflect,
	repeat          : utility.repeat,
	stylesheet      : utility.stylesheet,
	script          : utility.script,
	stop            : utility.stop,
	store           : data.factory,
	tpl             : utility.tpl,
	un              : function (obj, event, id, state) {
		var all = typeof id !== "undefined",
		    o, e, i, s;

		o = all ? obj   : this;
		e = all ? event : obj;
		i = all ? id    : event;
		s = all ? state : id;

		if (typeof o === "undefined" || o === $) o = abaaso;
		return observer.remove(o, e, i, s);
	},
	update          : element.update,
	version         : "3.2.2",
	walk            : utility.walk
};

})();

// Conditional bootstrap incase of multiple loading
if (typeof abaaso.bootstrap === "function") abaaso.bootstrap();

// Node, AMD & window supported
switch (true) {
	case typeof exports !== "undefined":
		module.exports = abaaso;
		break;
	case typeof define === "function":
		define("abaaso", function () { return abaaso; });
		break;
	default:
		global.abaaso = abaaso;
}

})(this);
